#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include "SLAB_MEMORY.h"
#include "slab.h"
#include "./mpa/mpa.h"

static Value
clc_eval_max_expression(LocalEnvironment *env, 
                        Expression *left, Expression *right)
{
    Value	left_val;
    Value	right_val;
    Value	result;
    int         i, n;
    char        *first="input";
    left_val = eval_expression(env, left);
    right_val = eval_expression(env, right);
    if (left_val.type != right_val.type) 
    {
        clc_runtime_error(NOT_SAME_DIMENSION_MATRIX_ERR, "(%s)\n", first);
    }
    if (left_val.type == DOUBLE_VALUE)
    { 
        result.type = DOUBLE_VALUE;
        result.u.double_value = left_val.u.double_value;
        if (left_val.u.double_value < right_val.u.double_value) 
        {
            result.u.double_value = right_val.u.double_value;
        }
    } else if(left_val.type == MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        result.u.matrix_value.row = left_val.u.matrix_value.row;
        result.u.matrix_value.column = left_val.u.matrix_value.column;
        n = left_val.u.matrix_value.row*left_val.u.matrix_value.column;
        result.u.matrix_value.top = (double *)GC_malloc(n*sizeof(double));
        for(i=0; i<n; ++i) 
        {
            result.u.matrix_value.top[i] = left_val.u.matrix_value.top[i];
            if (left_val.u.matrix_value.top[i]<right_val.u.matrix_value.top[i]) 
            {
                result.u.matrix_value.top[i] = right_val.u.matrix_value.top[i];
            }
        } 
    } 
    return result;
}

static Value
clc_eval_con_expression(LocalEnvironment *env, 
                        Expression *exp, Expression *next)
{
    Value	exp_val;
    Value	next_val;
    Value	result;
    int i, j, k, l;
    exp_val = eval_expression(env, exp);
    next_val = eval_expression(env, next);
    if (exp_val.type == MATRIX_VALUE) 
    {
        result.type = MATRIX_VALUE;
        k = exp_val.u.matrix_value.row;
        l = exp_val.u.matrix_value.column;
        result.u.matrix_value.top = calloc(k+1,sizeof(double));
        for(i=0;i<k;i++) 
        {
            result.u.matrix_value.top[i] = 
                     exp_val.u.matrix_value.top[i];
        }
        result.u.matrix_value.top[k] = next_val.u.double_value;
        result.u.matrix_value.row = k+1;
        result.u.matrix_value.column = l;
    }
    if (exp_val.type == DIF_VALUE) 
    {
        result.type = DIF_MATRIX_VALUE;
        result.u.difmatrix_value.value = calloc(2,sizeof(double));
        result.u.difmatrix_value.value[0] = exp_val.u.dif_value.value;
        result.u.difmatrix_value.value[1] = next_val.u.dif_value.value;
        k = exp_val.u.dif_value.n;
        result.u.difmatrix_value.jacobi = calloc(2*k,sizeof(double));
        for(i=0;i<k;i++) 
        {
            result.u.difmatrix_value.jacobi[2*i] = 
                        exp_val.u.dif_value.jacobi[i];
        }
        for(i=0;i<k;i++) 
        {
            result.u.difmatrix_value.jacobi[2*i+1] = 
                        next_val.u.dif_value.jacobi[i];
        }
        result.u.difmatrix_value.n = 2;
        result.u.difmatrix_value.column = k;
    }
    if (exp_val.type == DIF_MATRIX_VALUE) 
    {
        result.type = DIF_MATRIX_VALUE;
        k = exp_val.u.difmatrix_value.n;
        l = exp_val.u.difmatrix_value.column;
        result.u.difmatrix_value.value = calloc(k+1,sizeof(double));
        for(i=0;i<k;i++) 
        {
            result.u.difmatrix_value.value[i] = 
                     exp_val.u.difmatrix_value.value[i];
        }
        result.u.difmatrix_value.value[k] = next_val.u.dif_value.value;
        result.u.difmatrix_value.jacobi = calloc(l*(k+1),sizeof(double));
        for(j=0;j<k;j++)
        {
            for(i=0;i<l;i++) 
            {
                result.u.difmatrix_value.jacobi[(k+1)*i+j] = 
                        exp_val.u.difmatrix_value.jacobi[k*i+j];
            }
        }
        for(i=0;i<l;i++) 
        {
            result.u.difmatrix_value.jacobi[(k+1)*i+k] = 
                        next_val.u.dif_value.jacobi[i];
        }
        result.u.difmatrix_value.n = k+1;
        result.u.difmatrix_value.column = l;
    }
    return result;
}

static Value
clc_eval_size_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == MATRIX_VALUE) 
    {
        result.type = MATRIX_VALUE;
        result.u.matrix_value.column = 2;
        result.u.matrix_value.row = 1;
        result.u.matrix_value.top = calloc(2,sizeof(double));
        result.u.matrix_value.top[0]=operand_val.u.matrix_value.row;
        result.u.matrix_value.top[1]=operand_val.u.matrix_value.column;
    }
    return result;
}

static Value
clc_eval_split_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == DIF_VALUE) {
        result.type = COUPLE_VALUE;
        result.u.couple_value.type_first  = DOUBLE_VALUE;
        result.u.couple_value.type_second = MATRIX_VALUE;
        result.u.couple_value.name_first = "val";
        result.u.couple_value.name_second = "jacobi";
        k = operand_val.u.dif_value.n;
        result.u.couple_value.u_second.matrix_value.column = operand_val.u.dif_value.n;
        result.u.couple_value.u_second.matrix_value.row = 1;
        result.u.couple_value.u_first.double_value = 
                operand_val.u.dif_value.value;
        result.u.couple_value.u_second.matrix_value.top = 
            (double *)calloc(k,sizeof(double));
        for(i=0;i<k;i++) 
        {
            result.u.couple_value.u_second.matrix_value.top[i]
                =operand_val.u.dif_value.jacobi[i];
        } 
    } else if (operand_val.type == DIF_MATRIX_VALUE) {
        result.type = COUPLE_VALUE;
        result.u.couple_value.type_first  = MATRIX_VALUE;
        result.u.couple_value.type_second = MATRIX_VALUE;
        result.u.couple_value.name_first = "val";
        result.u.couple_value.name_second = "jacobi";
        k = operand_val.u.difmatrix_value.n;
        l = operand_val.u.difmatrix_value.column;
        result.u.couple_value.u_first.matrix_value.row = k;
        result.u.couple_value.u_first.matrix_value.column = 1;
        result.u.couple_value.u_second.matrix_value.column = l;
        result.u.couple_value.u_second.matrix_value.row = k;
        result.u.couple_value.u_first.matrix_value.top = 
            (double *)calloc(k,sizeof(double));
        result.u.couple_value.u_second.matrix_value.top = 
            (double *)calloc(k*l,sizeof(double));
        for(i=0;i<k;i++) 
        {
            result.u.couple_value.u_first.matrix_value.top[i]
                =operand_val.u.difmatrix_value.value[i];
        } 
        for(i=0;i<k*l;i++) 
        {
            result.u.couple_value.u_second.matrix_value.top[i]
                =operand_val.u.difmatrix_value.jacobi[i];
        } 
    } 
    return result;
}

Value
clc_eval_difmatrix_value_expression(LocalEnvironment *env, double *value, 
                                         double *jacobi, long int n, long int column)
{
    Value	result;

    result.u.difmatrix_value.value = value;
    result.u.difmatrix_value.jacobi = jacobi;
    result.u.difmatrix_value.n = n;
    result.u.difmatrix_value.column = column;
    result.type = DIF_MATRIX_VALUE; 

    return result;
}

static Value
eval_eval_expression(LocalEnvironment *env,
			      Expression *name, Expression *argument)
{
    Value	result, v_i;
    Expression		*arg_p;
    ParameterList	*param_p;
    LocalEnvironment	*local_env;
    FunctionDefinition	*func;
    
    v_i = eval_expression(env, name);
    func = clc_search_function(v_i.u.name_value.name);
    if (func == NULL) {
	clc_runtime_error(FUNCTION_NOT_FOUND_ERR, "name..%s\n", v_i.u.name_value.name);
    }
    local_env = alloc_local_environment();

    for (arg_p = argument, param_p = func->parameter;
	 arg_p;
	 arg_p = arg_p->u.expression_list.next,
	 param_p = param_p->next) {
	Value arg_val;

	if (param_p == NULL) {
	    clc_runtime_error(ARGUMENT_TOO_MANY_ERR, NULL);
	}
	arg_val = eval_expression(env,
				  arg_p->u.expression_list.expression);
	clc_add_local_variable(local_env, param_p->name, &arg_val);
    }
    if (param_p) {
	clc_runtime_error(ARGUMENT_TOO_FEW_ERR, NULL);
    }
    result = eval_expression(local_env, func->expression_list);

    dispose_local_environment(local_env);

    return result;
}

static void
eval_binary_dif_double(ExpressionType operator, dif left, double right,
		   Value *result)
{
    int i, j;
    long int m,n,k;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION
        || operator == EXACT_SOLVE_EXPRESSION) {
	result->type = DIF_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result->u.dif_value.value = 
               left.value + right;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    left.jacobi[i];
        }
        (*result).u.dif_value.n = left.n;
	break; 
      case SUB_EXPRESSION:
	result->u.dif_value.value = 
               left.value - right;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    left.jacobi[i];
        }
        (*result).u.dif_value.n = left.n;
	break; 
      case MUL_EXPRESSION:
	result->u.dif_value.value = 
               left.value * right;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    left.jacobi[i]*right;
        }
        (*result).u.dif_value.n = left.n;
	break; 
      case DIV_EXPRESSION:
	result->u.dif_value.value = 
               left.value / right;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    left.jacobi[i]/right;
        }
        (*result).u.dif_value.n = left.n;
	break; 
      case SOLVE_EXPRESSION:
      case EXACT_SOLVE_EXPRESSION:
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
} 

static void
eval_binary_double_dif(ExpressionType operator, double left, dif right, Value *result)
{
    int i, j;
    long int m,n,k;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION
        || operator == EXACT_SOLVE_EXPRESSION) {
	result->type = DIF_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result->u.dif_value.value = 
               left + right.value;
        for(i=0; i<right.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    right.jacobi[i];
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case SUB_EXPRESSION:
	result->u.dif_value.value = 
               left - right.value;
        for(i=0; i<right.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    -right.jacobi[i];
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case MUL_EXPRESSION:
	result->u.dif_value.value = 
               left * right.value;
        for(i=0; i<right.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    right.jacobi[i]*left;
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case DIV_EXPRESSION:
	result->u.dif_value.value = 
               left / right.value;
        for(i=0; i<right.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    ((-right.jacobi[i])*left)/(right.value * right.value);
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case SOLVE_EXPRESSION:
      case EXACT_SOLVE_EXPRESSION:
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
} 

static void
eval_binary_dif(ExpressionType operator, dif left, dif right,
		   Value *result)
{
    int i, j;
    long int m,n,k;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION
        || operator == EXACT_SOLVE_EXPRESSION) {
	result->type = DIF_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result->u.dif_value.value = 
               left.value + right.value;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    left.jacobi[i]+right.jacobi[i];
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case SUB_EXPRESSION:
	result->u.dif_value.value = 
               left.value - right.value;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    left.jacobi[i]-right.jacobi[i];
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case MUL_EXPRESSION:
	result->u.dif_value.value = 
               left.value * right.value;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    left.jacobi[i]*right.value+right.jacobi[i]*left.value;
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case DIV_EXPRESSION:
	result->u.dif_value.value = 
               left.value / right.value;
        for(i=0; i<left.n; ++i) {
	     result->u.dif_value.jacobi[i] = 
                    (left.jacobi[i]*right.value-right.jacobi[i]*left.value)/(right.value * right.value);
        }
        (*result).u.dif_value.n = right.n;
	break; 
      case SOLVE_EXPRESSION:
      case EXACT_SOLVE_EXPRESSION:
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
} 

static Value
clc_eval_dif_expression(LocalEnvironment *env, Expression *operand, Expression *dif_n, Expression *dif_i)
{
    Value	operand_val, v_n, v_i;
    Value	result;
    MPA         a, b;
    int i, n;

    operand_val = eval_expression(env, operand);
    v_n = eval_expression(env, dif_n);
    v_i = eval_expression(env, dif_i);
    n = (int)v_n.u.double_value; 
    i = (int)v_i.u.double_value; 
    if(operand_val.type == DOUBLE_VALUE) 
    {
        result.type = DIF_VALUE;
        result.u.dif_value.value = operand_val.u.double_value;
        result.u.dif_value.jacobi = (double *)calloc(n,sizeof(double));
        result.u.dif_value.jacobi[i] = 1.0;
        result.u.dif_value.n = n;
    }  
    return result;
}

void
eval_linpro_expression(LocalEnvironment *env, char *identifier, char *name_second, char *name_third)
{
    Variable    *pos;
    Variable    *pos1;
    Variable    *pos2;
    FILE *fp;
    int m1, n1, m2, n2, m3, n3;
    int i, j, i2;
    Value v;

    fp = fopen("./linpro/_temp", "w");
    if (fp == NULL) 
    {
         clc_runtime_error(CANNOT_OPEN_FILE_ERR, "(%s)\n", NULL);
    } 
    pos = NULL;
    pos1 = NULL;
    pos2 = NULL;
    if (env != NULL) {
       for(pos = env->variable; pos; pos = pos->next) {
          if (!strcmp(pos->name, identifier))
             break;
       }
    }
    if (pos == NULL) 
    {
        for(pos = clc_current_interpreter->variable; pos; pos = pos->next) 
        {
            if (!strcmp(pos->name, identifier))
                break;
        }
    }
    if (pos == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);
    }    
    if (env != NULL) 
    {
       for(pos1 = env->variable; pos1; pos1 = pos1->next) 
       {
          if (!strcmp(pos->name, name_second))
             break;
       }
    }
    if (pos1 == NULL) 
    {
        for(pos1 = clc_current_interpreter->variable; pos1; pos1 = pos1->next) 
        {
            if (!strcmp(pos1->name, name_second))
                break;
        }
    }
    if (pos1 == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_second);
    }    
    if (env != NULL) {
       for(pos2 = env->variable; pos2; pos2 = pos2->next) {
          if (!strcmp(pos2->name, name_third))
             break;
       }
    }
    if (pos2 == NULL) 
    {
        for(pos2 = clc_current_interpreter->variable; pos2; pos2 = pos2->next) 
        {
            if (!strcmp(pos2->name, name_third))
                break;
        }
    }
    if (pos2 == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_third);
    }    
    if(pos->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", identifier);
    }
    if(pos1->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", name_second);
    } 
    if(pos2->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", name_third);
    }
    m1=pos->value.u.matrix_value.column;
    n1=pos->value.u.matrix_value.row;
    m2=pos1->value.u.matrix_value.column;
    n2=pos1->value.u.matrix_value.row;
    m3=pos2->value.u.matrix_value.column;
    n3=pos2->value.u.matrix_value.row;
    fprintf(fp,"max: ");
    fprintf(fp,"%10.5f x%d", pos->value.u.matrix_value.top[0], 0);
    for(i = 1; i < m1*n1; i++) 
    {
        if(pos->value.u.matrix_value.top[i] >= 0)
        {
            fprintf(fp," + %10.5f x%d", pos->value.u.matrix_value.top[i], i);
        }
        else
        {
            fprintf(fp," %10.5f x%d", pos->value.u.matrix_value.top[i], i);
        
        }
    }
    fprintf(fp,";\nC1: ");
    for(i=0; i<n2; i++)
    {
        fprintf(fp,"%10.5f x%d", pos1->value.u.matrix_value.top[i], 0);
        for(j=1; j < m2; j++)
        {
            if(pos1->value.u.matrix_value.top[i+j*n2] >= 0)
            {
                fprintf(fp," + %10.5f x%d", pos1->value.u.matrix_value.top[i+j*n2], j);
            }
            else
            {
                fprintf(fp," %10.5f x%d", pos1->value.u.matrix_value.top[i+j*n2], j);
            }
        }
        fprintf(fp," < %10.5f;\n", pos2->value.u.matrix_value.top[i]); 
    } 
    fclose(fp);
    fp = popen("lp_solve -p < ./linpro/_temp > ./linpro/_out", "r");
    pclose(fp);
    clc_linpro_out();
}

static Value
clc_eval_matrix(LocalEnvironment *env, Expression *matrix_element)
{
    Expression *expr;
    Value v;

    expr =  clc_is_matrix_value_expression(env, matrix_element);
    if(expr->type == MATRIX_VALUE_EXPRESSION) 
    {
        v = clc_eval_matrix_value_expression(env,
                                       expr->u.matrix_value.top,
                                       expr->u.matrix_value.column,
                                       expr->u.matrix_value.row);
    }
    if(expr->type ==COMPLEX_MATRIX_VALUE_EXPRESSION)
    {
        v = clc_eval_complex_matrix_value_expression(env,
                                       expr->u.complex_matrix_value.top,
                                       expr->u.complex_matrix_value.column,
                                       expr->u.complex_matrix_value.row);
    } 
    if(expr->type == DIF_MATRIX_EXPRESSION) 
    {
        v = clc_eval_difmatrix_value_expression(env,
                                       expr->u.difmatrix_value.value,
                                       expr->u.difmatrix_value.jacobi,
                                       expr->u.difmatrix_value.n,
                                       expr->u.difmatrix_value.column);
    }
    return v;
}

static Value
clc_eval_pow_expression(LocalEnvironment *env, 
                             Expression *left, Expression *right)
{
    Value	operand_left;
    Value	operand_right;
    Value	result;
    double      ar, ai, br, bi, lr, li, z, x, y;
    int it;

    operand_left = eval_expression(env, left);
    operand_right = eval_expression(env, right);
    if(operand_left.type == DOUBLE_VALUE && 
       operand_right.type == DOUBLE_VALUE) 
    {
        it = (int)operand_right.u.double_value;
        z = operand_right.u.double_value - (double)it;
        if(operand_left.u.double_value < 0 && z != 0.0) 
        {
            ar = operand_left.u.double_value; 
            br = operand_right.u.double_value; 

            z = -ar;
            lr = log(z);
            li = atan2(0.0, ar);
            x = exp(lr * br);
            y = li * br;        
        
            result.type = COMPLEX_VALUE;
            result.u.complex_value.r = x * cos(y); 
            result.u.complex_value.i = x * sin(y);
        }
        else 
        {
            result.type = DOUBLE_VALUE;
            result.u.double_value = pow(operand_left.u.double_value,
                                    operand_right.u.double_value);
        }
    } 
    else if (operand_left.type == COMPLEX_VALUE ||
             operand_right.type == COMPLEX_VALUE) 
    {
        ar = operand_left.u.complex_value.r; 
        ai = operand_left.u.complex_value.i;
        br = operand_right.u.complex_value.r; 
        bi = operand_right.u.complex_value.i;

        z = sqrt(ar*ar+ai*ai);
        lr = log(z);
        li = atan2(ai, ar);
        x = exp(lr * br - li * bi);
        y = lr *bi + li * br;        
        
        result.type = COMPLEX_VALUE;
        result.u.complex_value.r = x * cos(y); 
        result.u.complex_value.i = x * sin(y);
    }
    return result;
}

static Value 
eval_triple_expression(LocalEnvironment *env, char *name_first, char *name_second,
                                              char *name_third)
{
    Value	v;
    Value	*vp;
    vp = clc_search_local_variable(env, name_first);
    v.type = TRIPLE_VALUE;
    v.u.triple_value.name_first = (char *)clc_global_malloc(strlen(name_first)+1);
    strcpy(v.u.triple_value.name_first, name_first);
    v.u.triple_value.name_second = (char *)clc_global_malloc(strlen(name_second)+1);
    strcpy(v.u.triple_value.name_second, name_second);
    v.u.triple_value.name_third = (char *)clc_global_malloc(strlen(name_third)+1);
    strcpy(v.u.triple_value.name_third, name_third);

    if (vp != NULL) {
        switch (vp->type) {
            case INT_VALUE:
            case COUPLE_VALUE:
            case TRIPLE_VALUE:
            case NULL_VALUE:
            case DOUBLE_VALUE:
                v.u.triple_value.type_first = DOUBLE_VALUE;
                v.u.triple_value.u_first.double_value 
                      = vp->u.double_value;
                break;
            case COMPLEX_VALUE:
                v.u.triple_value.type_first = COMPLEX_VALUE;
                v.u.triple_value.u_first.complex_value 
                      = vp->u.complex_value;
                break;
            case INTERVAL_VALUE:
                v.u.triple_value.type_first = INTERVAL_VALUE;
                v.u.triple_value.u_first.interval_value 
                      = vp->u.interval_value;
                break;
            case INTERVAL_MATRIX_VALUE:
                v.u.triple_value.type_first = INTERVAL_MATRIX_VALUE;
                v.u.triple_value.u_first.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                break;
            case MATRIX_VALUE:
                v.u.triple_value.type_first = MATRIX_VALUE;
                v.u.triple_value.u_first.matrix_value 
                      = vp->u.matrix_value;
                break;
            case COMPLEX_MATRIX_VALUE:
                v.u.triple_value.type_first = COMPLEX_MATRIX_VALUE;
                v.u.triple_value.u_first.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                break;
        } 
    } else {
        vp = clc_search_global_variable(name_first);
        if (vp != NULL) {
            switch (vp->type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.triple_value.type_first = DOUBLE_VALUE;
                    v.u.triple_value.u_first.double_value 
                      = vp->u.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.triple_value.type_first = COMPLEX_VALUE;
                    v.u.triple_value.u_first.complex_value 
                      = vp->u.complex_value;
                    break;
                case INTERVAL_VALUE:
                    v.u.triple_value.type_first = INTERVAL_VALUE;
                    v.u.triple_value.u_first.interval_value 
                      = vp->u.interval_value;
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.triple_value.type_first = INTERVAL_MATRIX_VALUE;
                    v.u.triple_value.u_first.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                    break;
                case MATRIX_VALUE:
                    v.u.triple_value.type_first = MATRIX_VALUE;
                    v.u.triple_value.u_first.matrix_value 
                      = vp->u.matrix_value;
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.triple_value.type_first = COMPLEX_MATRIX_VALUE;
                    v.u.triple_value.u_first.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                    break;
            } 
        } else {   
            clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_first);
        }
    }
    vp = clc_search_local_variable(env, name_second);
    if (vp != NULL) {
        switch (vp->type) {
            case INT_VALUE:
            case COUPLE_VALUE:
            case TRIPLE_VALUE:
            case NULL_VALUE:
            case DOUBLE_VALUE:
                v.u.triple_value.type_second = DOUBLE_VALUE;
                v.u.triple_value.u_second.double_value 
                      = vp->u.double_value;
                break;
            case COMPLEX_VALUE:
                v.u.triple_value.type_second = COMPLEX_VALUE;
                v.u.triple_value.u_second.complex_value 
                      = vp->u.complex_value;
                break;
            case INTERVAL_VALUE:
                v.u.triple_value.type_second = INTERVAL_VALUE;
                v.u.triple_value.u_second.interval_value 
                      = vp->u.interval_value;
                break;
            case INTERVAL_MATRIX_VALUE:
                v.u.triple_value.type_second = INTERVAL_MATRIX_VALUE;
                v.u.triple_value.u_second.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                break;
            case MATRIX_VALUE:
                v.u.triple_value.type_second = MATRIX_VALUE;
                v.u.triple_value.u_second.matrix_value 
                      = vp->u.matrix_value;
                break;
            case COMPLEX_MATRIX_VALUE:
                v.u.triple_value.type_second = COMPLEX_MATRIX_VALUE;
                v.u.triple_value.u_second.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                break;
        } 
    } else {
        vp = clc_search_global_variable(name_second);
        if (vp != NULL) {
            switch (vp->type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.triple_value.type_second = DOUBLE_VALUE;
                    v.u.triple_value.u_second.double_value 
                      = vp->u.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.triple_value.type_second = COMPLEX_VALUE;
                    v.u.triple_value.u_second.complex_value 
                      = vp->u.complex_value;
                    break;
                case INTERVAL_VALUE:
                    v.u.triple_value.type_second = INTERVAL_VALUE;
                    v.u.triple_value.u_second.interval_value 
                      = vp->u.interval_value;
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.triple_value.type_second = INTERVAL_MATRIX_VALUE;
                    v.u.triple_value.u_second.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                    break;
                case MATRIX_VALUE:
                    v.u.triple_value.type_second = MATRIX_VALUE;
                    v.u.triple_value.u_second.matrix_value 
                      = vp->u.matrix_value;
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.triple_value.type_second = COMPLEX_MATRIX_VALUE;
                    v.u.triple_value.u_second.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                    break;
            } 
        } else {   
            clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_second);
        }
    }
    vp = clc_search_local_variable(env, name_third);
    if (vp != NULL) {
        switch (vp->type) {
            case INT_VALUE:
            case COUPLE_VALUE:
            case TRIPLE_VALUE:
            case NULL_VALUE:
            case DOUBLE_VALUE:
                v.u.triple_value.type_third = DOUBLE_VALUE;
                v.u.triple_value.u_third.double_value 
                      = vp->u.double_value;
                break;
            case COMPLEX_VALUE:
                v.u.triple_value.type_third = COMPLEX_VALUE;
                v.u.triple_value.u_third.complex_value 
                      = vp->u.complex_value;
                break;
            case INTERVAL_VALUE:
                v.u.triple_value.type_third = INTERVAL_VALUE;
                v.u.triple_value.u_third.interval_value 
                      = vp->u.interval_value;
                break;
            case INTERVAL_MATRIX_VALUE:
                v.u.triple_value.type_third = INTERVAL_MATRIX_VALUE;
                v.u.triple_value.u_third.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                break;
            case MATRIX_VALUE:
                v.u.triple_value.type_third = MATRIX_VALUE;
                v.u.triple_value.u_third.matrix_value 
                      = vp->u.matrix_value;
                break;
            case COMPLEX_MATRIX_VALUE:
                v.u.triple_value.type_third = COMPLEX_MATRIX_VALUE;
                v.u.triple_value.u_third.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                break;
        } 
        return v;
    } else {
        vp = clc_search_global_variable(name_third);
        if (vp != NULL) {
            switch (vp->type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.triple_value.type_third = DOUBLE_VALUE;
                    v.u.triple_value.u_third.double_value 
                      = vp->u.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.triple_value.type_third = COMPLEX_VALUE;
                    v.u.triple_value.u_third.complex_value 
                      = vp->u.complex_value;
                    break;
                case INTERVAL_VALUE:
                    v.u.triple_value.type_third = INTERVAL_VALUE;
                    v.u.triple_value.u_third.interval_value 
                      = vp->u.interval_value;
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.triple_value.type_third = INTERVAL_MATRIX_VALUE;
                    v.u.triple_value.u_third.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                    break;
                case MATRIX_VALUE:
                    v.u.triple_value.type_third = MATRIX_VALUE;
                    v.u.triple_value.u_third.matrix_value 
                      = vp->u.matrix_value;
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.triple_value.type_third = COMPLEX_MATRIX_VALUE;
                    v.u.triple_value.u_third.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                    break;
            } 
            return v;
        } else {   
            clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_third);
        }
        return v; /* dummy */
    }
}

static Value
eval_int_expression(int int_value)
{
    Value	v;

    v.type = INT_VALUE;
    v.u.int_value = int_value;

    return v;
}

static Value
eval_double_expression(double double_value)
{
    Value	v;

    v.type = DOUBLE_VALUE;
    v.u.double_value = double_value;

    return v;
}

static Value
eval_complex_expression(double r,double i)
{
    Value	v;

    v.type = COMPLEX_VALUE;
    v.u.complex_value.r = r;
    v.u.complex_value.i = i;

    return v;
}

static Value
eval_identifier_expression(LocalEnvironment *env, char *identifier)
{
    Value	v;
    Value	*vp;

    vp = clc_search_local_variable(env, identifier);
    if (vp != NULL) {
        v = *vp;
        return v;
    }
    vp = clc_search_global_variable(identifier);
    if (vp != NULL) {
        v = *vp;
        return v;
    }    
    clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);
    return v; /* dummy. make compiler happy. */
}

static Value
eval_couple_identifier_expression(LocalEnvironment *env, char *identifier, char *name)
{
    Value	v;
    Value	*vp;

    vp = clc_search_local_variable(env, identifier);
    if (vp != NULL) {
    if (vp->type == COUPLE_VALUE) {
        if (!strcmp(vp->u.couple_value.name_first, name)) {
            v.type = vp->u.couple_value.type_first; 
            switch (v.type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.double_value = 
                       vp->u.couple_value.u_first.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.complex_value =
                       vp->u.couple_value.u_first.complex_value; 
                    break;
                case INTERVAL_VALUE:
                    v.u.interval_value =
                       vp->u.couple_value.u_first.interval_value; 
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.interval_matrix_value =
                       vp->u.couple_value.u_first.interval_matrix_value; 
                    break;
                case MATRIX_VALUE:
                    v.u.matrix_value =
                    vp->u.couple_value.u_first.matrix_value; 
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.complex_matrix_value =
                    vp->u.couple_value.u_first.complex_matrix_value; 
                    break;
            } 
        return v;
        } else {
            if (!strcmp(vp->u.couple_value.name_second, name)) {
                v.type = vp->u.couple_value.type_second; 
                switch (v.type) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        v.u.double_value = 
                           vp->u.couple_value.u_second.double_value;
                        break;
                    case COMPLEX_VALUE:
                        v.u.complex_value =
                           vp->u.couple_value.u_second.complex_value; 
                        break;
                    case INTERVAL_VALUE:
                        v.u.interval_value =
                           vp->u.couple_value.u_second.interval_value; 
                        break;
                    case INTERVAL_MATRIX_VALUE:
                        v.u.interval_matrix_value =
                           vp->u.couple_value.u_second.interval_matrix_value; 
                        break;
                    case MATRIX_VALUE:
                        v.u.matrix_value =
                           vp->u.couple_value.u_second.matrix_value; 
                        break;
                    case COMPLEX_MATRIX_VALUE:
                        v.u.complex_matrix_value =
                           vp->u.couple_value.u_second.complex_matrix_value; 
                        break;
                } 
                return v;
            }
        }
        } else if (vp->type == TRIPLE_VALUE) {
        if (!strcmp(vp->u.triple_value.name_first, name)) {
            v.type = vp->u.triple_value.type_first; 
            switch (v.type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.double_value = 
                       vp->u.triple_value.u_first.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.complex_value =
                       vp->u.triple_value.u_first.complex_value; 
                    break;
                case INTERVAL_VALUE:
                    v.u.interval_value =
                       vp->u.triple_value.u_first.interval_value; 
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.interval_matrix_value =
                       vp->u.triple_value.u_first.interval_matrix_value; 
                    break;
                case MATRIX_VALUE:
                    v.u.matrix_value =
                    vp->u.triple_value.u_first.matrix_value; 
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.complex_matrix_value =
                    vp->u.triple_value.u_first.complex_matrix_value; 
                    break;
            } 
        return v;
        } else if (!strcmp(vp->u.triple_value.name_second, name)) {
                v.type = vp->u.triple_value.type_second; 
                switch (v.type) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        v.u.double_value = 
                           vp->u.triple_value.u_second.double_value;
                        break;
                    case COMPLEX_VALUE:
                        v.u.complex_value =
                           vp->u.triple_value.u_second.complex_value; 
                        break;
                    case INTERVAL_VALUE:
                        v.u.interval_value =
                           vp->u.triple_value.u_second.interval_value; 
                        break;
                    case INTERVAL_MATRIX_VALUE:
                        v.u.interval_matrix_value =
                           vp->u.triple_value.u_second.interval_matrix_value; 
                        break;
                    case MATRIX_VALUE:
                        v.u.matrix_value =
                           vp->u.triple_value.u_second.matrix_value; 
                        break;
                    case COMPLEX_MATRIX_VALUE:
                        v.u.complex_matrix_value =
                           vp->u.triple_value.u_second.complex_matrix_value; 
                        break;
                } 
                return v;
        } else if (!strcmp(vp->u.triple_value.name_third, name)) {
                v.type = vp->u.triple_value.type_third; 
                switch (v.type) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        v.u.double_value = 
                           vp->u.triple_value.u_third.double_value;
                        break;
                    case COMPLEX_VALUE:
                        v.u.complex_value =
                           vp->u.triple_value.u_third.complex_value; 
                        break;
                    case INTERVAL_VALUE:
                        v.u.interval_value =
                           vp->u.triple_value.u_third.interval_value; 
                        break;
                    case INTERVAL_MATRIX_VALUE:
                        v.u.interval_matrix_value =
                           vp->u.triple_value.u_third.interval_matrix_value; 
                        break;
                    case MATRIX_VALUE:
                        v.u.matrix_value =
                           vp->u.triple_value.u_third.matrix_value; 
                        break;
                    case COMPLEX_MATRIX_VALUE:
                        v.u.complex_matrix_value =
                           vp->u.triple_value.u_third.complex_matrix_value; 
                        break;
                } 
                return v;
            }
        }
    }
    vp = clc_search_global_variable(identifier);
    if (vp != NULL) {
    if (vp->type == COUPLE_VALUE) {
        if (!strcmp(vp->u.couple_value.name_first, name)) {
            v.type = vp->u.couple_value.type_first; 
            switch (v.type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.double_value = 
                       vp->u.couple_value.u_first.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.complex_value =
                       vp->u.couple_value.u_first.complex_value; 
                    break;
                case INTERVAL_VALUE:
                    v.u.interval_value =
                       vp->u.couple_value.u_first.interval_value; 
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.interval_matrix_value =
                       vp->u.couple_value.u_first.interval_matrix_value; 
                    break;
                case MATRIX_VALUE:
                    v.u.matrix_value =
                    vp->u.couple_value.u_first.matrix_value; 
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.complex_matrix_value =
                    vp->u.couple_value.u_first.complex_matrix_value; 
                    break;
            } 
        return v;
        } else {
            if (!strcmp(vp->u.couple_value.name_second, name)) {
                v.type = vp->u.couple_value.type_second; 
                switch (v.type) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        v.u.double_value = 
                           vp->u.couple_value.u_second.double_value;
                        break;
                    case COMPLEX_VALUE:
                        v.u.complex_value =
                           vp->u.couple_value.u_second.complex_value; 
                        break;
                    case INTERVAL_VALUE:
                        v.u.interval_value =
                           vp->u.couple_value.u_second.interval_value; 
                        break;
                    case INTERVAL_MATRIX_VALUE:
                        v.u.interval_matrix_value =
                           vp->u.couple_value.u_second.interval_matrix_value; 
                        break;
                    case MATRIX_VALUE:
                        v.u.matrix_value =
                           vp->u.couple_value.u_second.matrix_value; 
                        break;
                    case COMPLEX_MATRIX_VALUE:
                        v.u.complex_matrix_value =
                           vp->u.couple_value.u_second.complex_matrix_value; 
                        break;
                } 
                return v;
            }
        }
        } else if (vp->type == TRIPLE_VALUE) {
        if (!strcmp(vp->u.triple_value.name_first, name)) {
            v.type = vp->u.triple_value.type_first; 
            switch (v.type) {
                case DOUBLE_VALUE:
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    v.u.double_value = 
                       vp->u.triple_value.u_first.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.complex_value =
                       vp->u.triple_value.u_first.complex_value; 
                    break;
                case INTERVAL_VALUE:
                    v.u.interval_value =
                       vp->u.triple_value.u_first.interval_value; 
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.interval_matrix_value =
                       vp->u.triple_value.u_first.interval_matrix_value; 
                    break;
                case MATRIX_VALUE:
                    v.u.matrix_value =
                    vp->u.triple_value.u_first.matrix_value; 
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.complex_matrix_value =
                    vp->u.triple_value.u_first.complex_matrix_value; 
                    break;
            } 
        return v;
        } else if (!strcmp(vp->u.triple_value.name_second, name)) {
                v.type = vp->u.triple_value.type_second; 
                switch (v.type) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        v.u.double_value = 
                           vp->u.triple_value.u_second.double_value;
                        break;
                    case COMPLEX_VALUE:
                        v.u.complex_value =
                           vp->u.triple_value.u_second.complex_value; 
                        break;
                    case INTERVAL_VALUE:
                        v.u.interval_value =
                           vp->u.triple_value.u_second.interval_value; 
                        break;
                    case INTERVAL_MATRIX_VALUE:
                        v.u.interval_matrix_value =
                           vp->u.triple_value.u_second.interval_matrix_value; 
                        break;
                    case MATRIX_VALUE:
                        v.u.matrix_value =
                           vp->u.triple_value.u_second.matrix_value; 
                        break;
                    case COMPLEX_MATRIX_VALUE:
                        v.u.complex_matrix_value =
                           vp->u.triple_value.u_second.complex_matrix_value; 
                        break;
                } 
                return v;
        } else if (!strcmp(vp->u.triple_value.name_third, name)) {
                v.type = vp->u.triple_value.type_third; 
                switch (v.type) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        v.u.double_value = 
                           vp->u.triple_value.u_third.double_value;
                        break;
                    case COMPLEX_VALUE:
                        v.u.complex_value =
                           vp->u.triple_value.u_third.complex_value; 
                        break;
                    case INTERVAL_VALUE:
                        v.u.interval_value =
                           vp->u.triple_value.u_third.interval_value; 
                        break;
                    case INTERVAL_MATRIX_VALUE:
                        v.u.interval_matrix_value =
                           vp->u.triple_value.u_third.interval_matrix_value; 
                        break;
                    case MATRIX_VALUE:
                        v.u.matrix_value =
                           vp->u.triple_value.u_third.matrix_value; 
                        break;
                    case COMPLEX_MATRIX_VALUE:
                        v.u.complex_matrix_value =
                           vp->u.triple_value.u_third.complex_matrix_value; 
                        break;
                } 
                return v;
            } else {
            clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name);
            }
        }
    }
    clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);
    return v; 
}

static Value
eval_couple_expression(LocalEnvironment *env, char *name_first, char *name_second)
{
    Value	v;
    Value	*vp;
    vp = clc_search_local_variable(env, name_first);
    v.type = COUPLE_VALUE;
    v.u.couple_value.name_first = (char *)clc_global_malloc(strlen(name_first)+1);
    strcpy(v.u.couple_value.name_first, name_first);
    v.u.couple_value.name_second = (char *)clc_global_malloc(strlen(name_second)+1);
    strcpy(v.u.couple_value.name_second, name_second);
    if (vp != NULL) {
        switch (vp->type) {
            case INT_VALUE:
            case COUPLE_VALUE:
            case TRIPLE_VALUE:
            case NULL_VALUE:
            case DOUBLE_VALUE:
                v.u.couple_value.type_first = DOUBLE_VALUE;
                v.u.couple_value.u_first.double_value 
                      = vp->u.double_value;
                break;
            case COMPLEX_VALUE:
                v.u.couple_value.type_first = COMPLEX_VALUE;
                v.u.couple_value.u_first.complex_value 
                      = vp->u.complex_value;
                break;
            case INTERVAL_VALUE:
                v.u.couple_value.type_first = INTERVAL_VALUE;
                v.u.couple_value.u_first.interval_value 
                      = vp->u.interval_value;
                break;
            case INTERVAL_MATRIX_VALUE:
                v.u.couple_value.u_first.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                break;
            case MATRIX_VALUE:
                v.u.couple_value.u_first.matrix_value 
                      = vp->u.matrix_value;
                break;
            case COMPLEX_MATRIX_VALUE:
                v.u.couple_value.u_first.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                break;
        } 
    } else {
        vp = clc_search_global_variable(name_first);
        if (vp != NULL) {
            switch (vp->type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.couple_value.type_first = DOUBLE_VALUE;
                    v.u.couple_value.u_first.double_value 
                      = vp->u.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.couple_value.type_first = COMPLEX_VALUE;
                    v.u.couple_value.u_first.complex_value 
                      = vp->u.complex_value;
                    break;
                case INTERVAL_VALUE:
                    v.u.couple_value.type_first = INTERVAL_VALUE;
                    v.u.couple_value.u_first.interval_value 
                      = vp->u.interval_value;
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.couple_value.type_first = INTERVAL_MATRIX_VALUE;
                    v.u.couple_value.u_first.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                    break;
                case MATRIX_VALUE:
                    v.u.couple_value.type_first = MATRIX_VALUE;
                    v.u.couple_value.u_first.matrix_value 
                      = vp->u.matrix_value;
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.couple_value.type_first = COMPLEX_MATRIX_VALUE;
                    v.u.couple_value.u_first.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                    break;
            } 
        } else {   
            clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_first);
        }
    }
    vp = clc_search_local_variable(env, name_second);
    if (vp != NULL) {
        switch (vp->type) {
            case INT_VALUE:
            case COUPLE_VALUE:
            case TRIPLE_VALUE:
            case NULL_VALUE:
            case DOUBLE_VALUE:
                v.u.couple_value.type_second = DOUBLE_VALUE;
                v.u.couple_value.u_second.double_value 
                      = vp->u.double_value;
                break;
            case COMPLEX_VALUE:
                v.u.couple_value.type_second = COMPLEX_VALUE;
                v.u.couple_value.u_second.complex_value 
                      = vp->u.complex_value;
                break;
            case INTERVAL_VALUE:
                v.u.couple_value.type_second = INTERVAL_VALUE;
                v.u.couple_value.u_second.interval_value 
                      = vp->u.interval_value;
                break;
            case INTERVAL_MATRIX_VALUE:
                v.u.couple_value.type_second = INTERVAL_MATRIX_VALUE;
                v.u.couple_value.u_second.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                break;
            case MATRIX_VALUE:
                v.u.couple_value.type_second = MATRIX_VALUE;
                v.u.couple_value.u_second.matrix_value 
                      = vp->u.matrix_value;
                break;
            case COMPLEX_MATRIX_VALUE:
                v.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                v.u.couple_value.u_second.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                break;
        } 
        return v;
    } else {
        vp = clc_search_global_variable(name_second);
        if (vp != NULL) {
            switch (vp->type) {
                case INT_VALUE:
                case COUPLE_VALUE:
                case TRIPLE_VALUE:
                case NULL_VALUE:
                case DOUBLE_VALUE:
                    v.u.couple_value.type_second = DOUBLE_VALUE;
                    v.u.couple_value.u_second.double_value 
                      = vp->u.double_value;
                    break;
                case COMPLEX_VALUE:
                    v.u.couple_value.type_second = COMPLEX_VALUE;
                    v.u.couple_value.u_second.complex_value 
                      = vp->u.complex_value;
                    break;
                case INTERVAL_VALUE:
                    v.u.couple_value.type_second = INTERVAL_VALUE;
                    v.u.couple_value.u_second.interval_value 
                      = vp->u.interval_value;
                    break;
                case INTERVAL_MATRIX_VALUE:
                    v.u.couple_value.type_second = INTERVAL_MATRIX_VALUE;
                    v.u.couple_value.u_second.interval_matrix_value 
                      = vp->u.interval_matrix_value;
                    break;
                case MATRIX_VALUE:
                    v.u.couple_value.type_second = MATRIX_VALUE;
                    v.u.couple_value.u_second.matrix_value 
                      = vp->u.matrix_value;
                    break;
                case COMPLEX_MATRIX_VALUE:
                    v.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    v.u.couple_value.u_second.complex_matrix_value 
                      = vp->u.complex_matrix_value;
                    break;
            } 
            return v;
        } else {   
            clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_second);
        }
        return v; /* dummy */
    }
}

/* static void
eval_save_expression(LocalEnvironment *env, char *identifier,
                          char *data_file)
{
    Variable    *pos;
    FILE *fp;
    printf("data=%s\n",data_file);
    fp = fopen(data_file, "w");
    if (fp == NULL) {
         printf("   Error:canonot open file!\n");
    } else {
        if (env != NULL) {
            for(pos = env->variable; pos; pos = pos->next) {
                if (!strcmp(pos->name, identifier))
                     break;
            }
        }
        if (pos != NULL) {
            clc_fprint_value(pos->value, fp);
        }    
        for(pos = clc_current_interpreter->variable; pos; pos = pos->next) {
           if (!strcmp(pos->name, identifier))
              break;
        }
        if (pos != NULL) {
            clc_fprint_value(pos->value, fp);
        }    
        fclose(fp);
    }
}*/ 

static void
eval_who_expression(LocalEnvironment *env)
{
    Variable    *pos;
    FunctionDefinition *fpos;

    if (env != NULL) {
        printf("Local Variables: \n");
        for (pos = env->variable; pos; pos = pos->next) {
             if(pos->name != NULL) {
                 printf("   %s", pos->name);
             }
        }
    }

    printf("Global Variables: \n");
    for (pos = clc_current_interpreter->variable;
                                pos; pos = pos->next) { 
         if(pos->name != NULL) {
             printf("   %s", pos->name);
         }
    }

    printf("\nUser Defined Functions: \n");
    for (fpos = clc_current_interpreter->function_list;
                                fpos; fpos = fpos->next) {
        if(fpos->name != NULL) {
            printf("   %s", fpos->name);
        }
    } 
    printf("\n");
}

static Value
eval_matrix_ones_expression(LocalEnvironment *env,
                            Expression *row, Expression *column)
{
    Value v_row, v_column;
    Value  v;
    int i, j; 
    int n, m;

    v_row = eval_expression(env, row);
    v_column = eval_expression(env, column);
    
    n = (int)v_row.u.double_value;
    m = (int)v_column.u.double_value;
    
    v.u.matrix_value.top = (double*)calloc(n*m, sizeof(double));
    for(i=0; i<n; ++i) {
         for(j=0; j<m; ++j) {
             v.u.matrix_value.top[i+j*n] = 1.; 
         }
    }
    v.u.matrix_value.row=n;
    v.u.matrix_value.column=m;
    v.type = MATRIX_VALUE;
    return v;
}

static Value
eval_matrix_eye_expression(LocalEnvironment *env,
                            Expression *row, Expression *column)
{
    Value v_row, v_column;
    Value  v;
    int i, j; 
    int n, m;

    v_row = eval_expression(env, row);
    v_column = eval_expression(env, column);
    
    n = (int)v_row.u.double_value;
    m = (int)v_column.u.double_value;
    
    v.u.matrix_value.top = (double*)calloc(n*m, sizeof(double));
    for(i=0; i<n; ++i) {
         for(j=0; j<m; ++j) {
             v.u.matrix_value.top[i+j*n] = 0.; 
         }
         v.u.matrix_value.top[i+i*n] = 1.;
    }
    v.u.matrix_value.row=n;
    v.u.matrix_value.column=m;
    v.type = MATRIX_VALUE;
    return v;
}

static Value
eval_lread_expression(void)
{
    Value  v;
    int i, j; 
    int n, m;
    FILE *fp;
    char file_name[128] = "./linpro/_out"; 
    char buf[1024];
    double a;

    if((fp = fopen(file_name, "r")) == NULL) {
        fprintf(stderr, "_out cannot be open\n");
    }
    
    m = -2;
    while(fgets(buf,1024,fp)!=0) {
        m++;
    }
    m = m/2;
    v.type = TRIPLE_VALUE;
    v.u.triple_value.type_first  = DOUBLE_VALUE;
    v.u.triple_value.type_second = MATRIX_VALUE;
    v.u.triple_value.type_third = MATRIX_VALUE;
    v.u.triple_value.name_first = "OBJ";
    v.u.triple_value.name_second = "PRIM";
    v.u.triple_value.name_third = "DUAL";
    v.u.triple_value.u_second.matrix_value.column = 1;
    v.u.triple_value.u_second.matrix_value.row = m;
    v.u.triple_value.u_third.matrix_value.column = 1;
    v.u.triple_value.u_third.matrix_value.row = m;

    v.u.triple_value.u_second.matrix_value.top  = 
        (double *)calloc(m,sizeof(double));
    v.u.triple_value.u_third.matrix_value.top  = 
        (double *)calloc(m,sizeof(double));
    fclose(fp);
    if((fp = fopen(file_name, "r")) == NULL) {
        fprintf(stderr, "_out cannot be open\n");
    }
    i = 0; 
    fgets(buf,1024,fp);
    j = 0;
    while(isdigit(buf[j]) == 0) {
        ++j;
        if(j>1024) break;
    }
    a = atof(buf + j);
    v.u.triple_value.u_first.double_value = a; 

    for(i=0;i<m;++i) {
        fgets(buf,1024,fp);
        j = 0;
        while(buf[j] != ' ') {
            ++j;
            if(j>1024) break;
        }
        while(buf[j] == ' ') {
            ++j;
            if(j>1024) break;
        }
        a = atof(buf + j);
        v.u.triple_value.u_second.matrix_value.top[i] = a; 
    }
    fgets(buf,1024,fp);
    fgets(buf,1024,fp);
    for(i=0;i<m;++i) {
        fgets(buf,1024,fp);
        j = 0;
        while(buf[j] != ' ') {
            ++j;
            if(j>1024) break;
        }
        while(buf[j] == ' ') {
            ++j;
            if(j>1024) break;
        }
        a = atof(buf + j);
        v.u.triple_value.u_third.matrix_value.top[i] = a; 
    }
    fclose(fp);
    return v;
}

static Value
eval_fft_expression(LocalEnvironment *env,
                            Expression *data, Expression *number)
{
    Value v_data, v_number;
    Value  v;
    int i, j, M; 
    int n, m;
    int num;
    int *ip;
    double *w;

    v_data = eval_expression(env, data);
    if(v_data.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", "data");
    } 
    n = (int)v_data.u.matrix_value.row;
    m = (int)v_data.u.matrix_value.column;
    if (number == NULL) 
    {
        num = n*m;
    }
    else
    {
        v_number = eval_expression(env, number);
        if(v_number.type != DOUBLE_VALUE) 
        {
            clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", "number");
        } 
        num = (int)v_number.u.double_value;
    }
    ip = (int *)GC_malloc_atomic((2+num)*sizeof(int));
    w = (double *)GC_malloc_atomic((num/2)*sizeof(double)); 
    v.u.matrix_value.top = (double*)GC_malloc_atomic(num* sizeof(double));
    for(i=0; i<num; i++) {
        v.u.matrix_value.top[i] = 0.;
    }

    M = n*m;
    if(num < M) M=num; 
    for(i=0; i<M; ++i) 
    {
       v.u.matrix_value.top[i] = v_data.u.matrix_value.top[i];
    }
    ip[0] = 0;
    rdft(num, 1, v.u.matrix_value.top, ip, w);

    v.u.matrix_value.row=1;
    v.u.matrix_value.column=num;
    v.type = MATRIX_VALUE;
    GC_free(ip);
    GC_free(w);
    return v;
}

static Value
eval_ifft_expression(LocalEnvironment *env,
                            Expression *data, Expression *number)
{
    Value v_data, v_number;
    Value  v;
    int i, j, M; 
    int n, m;
    int num;
    int *ip;
    double *w;

    v_data = eval_expression(env, data);
    if(v_data.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", "data");
    } 
    n = (int)v_data.u.matrix_value.row;
    m = (int)v_data.u.matrix_value.column;
    if (number == NULL) 
    {
        num = n*m;
    }
    else
    {
        v_number = eval_expression(env, number);
        if(v_number.type != DOUBLE_VALUE) 
        {
            clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", "number");
        } 
        num = (int)v_number.u.double_value;
    }
    ip = (int *)GC_malloc_atomic((2+num)*sizeof(int));
    w = (double *)GC_malloc_atomic((num/2)*sizeof(double)); 
    v.u.matrix_value.top = (double*)GC_malloc_atomic(num* sizeof(double));
    for(i=0; i<num; i++) {
        v.u.matrix_value.top[i] = 0.;
    }

    M = n*m;
    if(num < M) M=num; 
    for(i=0; i<M; ++i) 
    {
       v.u.matrix_value.top[i] = 2.0 * v_data.u.matrix_value.top[i]/num;
    }
    ip[0] = 0;
    rdft(num, -1, v.u.matrix_value.top, ip, w);

    v.u.matrix_value.row=1;
    v.u.matrix_value.column=num;
    v.type = MATRIX_VALUE;
    GC_free(ip);
    GC_free(w);
    return v;
}

static Value
eval_matrix_rand_expression(LocalEnvironment *env,
                            Expression *row, Expression *column)
{
    Value v_row, v_column;
    Value  v;
    int i, j; 
    int n, m;

    srand((unsigned)time(NULL));
    v_row = eval_expression(env, row);
    v_column = eval_expression(env, column);
    
    n = (int)v_row.u.double_value;
    m = (int)v_column.u.double_value;
    
    v.u.matrix_value.top = (double*)calloc(n*m, sizeof(double));
    for(i=0; i<n; ++i) {
         for(j=0; j<m; ++j) {
             v.u.matrix_value.top[i+j*n] = rand()/(RAND_MAX+1.0); 
         }
    }
    v.u.matrix_value.row=n;
    v.u.matrix_value.column=m;
    v.type = MATRIX_VALUE;
    return v;
}

static Value
eval_linspace_expression(LocalEnvironment *env,
                     Expression *start, Expression *end, Expression *number)
{
    Value v_start, v_end, v_number;
    Value  v;
    int j, m;
    double s,e,h;

    v_number = eval_expression(env, number);
    v_start  = eval_expression(env, start);
    v_end    = eval_expression(env, end);
    
    m = (int)v_number.u.double_value;
    s = v_start.u.double_value;
    e = v_end.u.double_value;
    h = (e-s)/(v_number.u.double_value-1.);
    
    v.u.matrix_value.top = (double*)calloc(m, sizeof(double));
   
    for(j=0; j<m; ++j) {
        v.u.matrix_value.top[j] = s+j*h; 
    }

    v.u.matrix_value.row=1;
    v.u.matrix_value.column=m;
    v.type = MATRIX_VALUE;
    return v;
}

static Value
eval_matrix_zeros_expression(LocalEnvironment *env,
                            Expression *row, Expression *column)
{
    Value v_row, v_column;
    Value  v;
    int i, j; 
    int n, m;

    v_row = eval_expression(env, row);
    v_column = eval_expression(env, column);
    
    n = (int)v_row.u.double_value;
    m = (int)v_column.u.double_value;
    
    v.u.matrix_value.top = (double*)calloc(n*m, sizeof(double));
    for(i=0; i<n; ++i) {
         for(j=0; j<m; ++j) {
             v.u.matrix_value.top[i+j*n] = 0.; 
         }
    }
    v.u.matrix_value.row=n;
    v.u.matrix_value.column=m;
    v.type = MATRIX_VALUE;
    return v;
}


static Value
eval_matrix_element_assign_expression2(LocalEnvironment *env,
		       char *identifier, Expression *row,
                       Expression *operand)
{
    Value	v;
    Value	right;
    Value	v_row;
    Value	*vp;
    int i,j;

    right = eval_expression(env, operand);
    if (right.type != DOUBLE_VALUE) {
        clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", identifier);
    }

    v_row = eval_expression(env, row);
    
    if (v_row.type != DOUBLE_VALUE ) {
        clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", identifier);
    }
    i = (int)v_row.u.double_value;
    vp = clc_search_local_variable(env, identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE && i <= v.u.matrix_value.row * v.u.matrix_value.column && i >=0 ) {
            vp->u.matrix_value.top[i]=right.u.double_value;
            return v;
        } else {
            clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
        }
    }
    vp = clc_search_global_variable(identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE && i < v.u.matrix_value.row * v.u.matrix_value.column && i >=0 ) {
            vp->u.matrix_value.top[i]=right.u.double_value;
            return v;
        } else {
            clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
        }
    }
    clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);

    return v; /* dummy. make compiler happy. */
}


static Value
eval_matrix_element_assign_expression(LocalEnvironment *env,
		       char *identifier, Expression *row, Expression *column,
                       Expression *operand)
{
    Value	v;
    Value	right;
    Value	v_row, v_column;
    Value	*vp;
    int i,j;

    right = eval_expression(env, operand);
    if (right.type != DOUBLE_VALUE) {
        clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", identifier);
    }

    v_row = eval_expression(env, row);
    v_column = eval_expression(env, column);
    
    if (v_row.type != DOUBLE_VALUE || v_column.type != DOUBLE_VALUE) {
        clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", identifier);
    }
    i = (int)v_row.u.double_value;
    j = (int)v_column.u.double_value;
    vp = clc_search_local_variable(env, identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE && i <= v.u.matrix_value.row && j<= v.u.matrix_value.column && i >=0 && j >= 0) {
            vp->u.matrix_value.top[i+j*(v.u.matrix_value.row)]=right.u.double_value;
            return v;
        } else {
            clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
        }
    }
    vp = clc_search_global_variable(identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE && i < v.u.matrix_value.row && j< v.u.matrix_value.column && i >=0 && j >= 0) {
            vp->u.matrix_value.top[i+j*(v.u.matrix_value.row)]=right.u.double_value;
            return v;
        } else {
            clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
        }
    }
    clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);

    return v; /* dummy. make compiler happy. */
}


static Value
eval_matrix_element_call_expression(LocalEnvironment *env, char *identifier,
         Expression *row, Expression *column)
{
    Value	result, v, v_row, v_column;
    Value	*vp;
    int i,j;
    v_row = eval_expression(env, row);
    v_column = eval_expression(env, column);
    
    if (v_row.type != DOUBLE_VALUE || v_column.type != DOUBLE_VALUE) {
        clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", identifier);
    }
    i = (int)v_row.u.double_value;
    j = (int)v_column.u.double_value;
    vp = clc_search_local_variable(env, identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE && i <= v.u.matrix_value.row && j<= v.u.matrix_value.column && i >=0 && j >= 0) {
            result.type = DOUBLE_VALUE;
            result.u.double_value=v.u.matrix_value.top[i+j*(v.u.matrix_value.row)];    
	    return result;
        } else {
            clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
        }
    }
    vp = clc_search_global_variable(identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE && i < v.u.matrix_value.row && j< v.u.matrix_value.column && i >=0 && j >= 0) {
            result.type = DOUBLE_VALUE;
            result.u.double_value=v.u.matrix_value.top[i+j*(v.u.matrix_value.row)];    
	    return result;
        } 
        else 
        {
            if (v.type == COMPLEX_MATRIX_VALUE && i < v.u.complex_matrix_value.row && j< v.u.complex_matrix_value.column && i >=0 && j >= 0) 
            {
                result.type = COMPLEX_VALUE;
                result.u.complex_value=v.u.complex_matrix_value.top[i+j*(v.u.matrix_value.row)];    
	        return result;
            }
            else
            {
                if (v.type == INTERVAL_MATRIX_VALUE && i < v.u.interval_matrix_value.upper.row && j< v.u.interval_matrix_value.upper.column && i >=0 && j >= 0) 
                {
                    result.type = INTERVAL_VALUE;
                    result.u.interval_value.upper=v.u.interval_matrix_value.upper.top[i+j*(v.u.matrix_value.row)];    
                    result.u.interval_value.lower=v.u.interval_matrix_value.lower.top[i+j*(v.u.matrix_value.row)];    
	            return result;
                } 
                else
                {
                    clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
                }
            }
        }
    }
    clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);

    return v; /* dummy. make compiler happy. */
}

static Value
eval_matrix_element_call_expression2(LocalEnvironment *env, char *identifier,
         Expression *row)
{
    Value	result, v, v_row;
    Value	*vp;
    int i,j,k;
    v_row = eval_expression(env, row);
    
    if (v_row.type != DOUBLE_VALUE ) {
        clc_runtime_error(DOUBLE_EXPECTED_ERR, "(%s)\n", identifier);
    }
    i = (int)v_row.u.double_value;
    vp = clc_search_local_variable(env, identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE 
            && i <= v.u.matrix_value.row * v.u.matrix_value.column && i >=0 ) 
        {
            result.type = DOUBLE_VALUE;
            result.u.double_value=v.u.matrix_value.top[i];    
	    return result;
        } 
        else if (v.type == DIF_MATRIX_VALUE 
            && i < v.u.difmatrix_value.n && i >=0 ) 
        {
            result.type = DIF_VALUE;
            result.u.dif_value.value=v.u.difmatrix_value.value[i];    
            result.u.dif_value.jacobi=calloc(v.u.difmatrix_value.column,sizeof(double));
            for(k=0; k<v.u.difmatrix_value.column; k++)
            { 
                result.u.dif_value.jacobi[k]
                    =v.u.difmatrix_value.jacobi[i+k*v.u.difmatrix_value.n];    
            }
            result.u.dif_value.n=v.u.difmatrix_value.column;
	    return result;
        } 
        else 
        {
            clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
        }
    }
    vp = clc_search_global_variable(identifier);
    if (vp != NULL) {
	v = *vp;
        if (v.type == MATRIX_VALUE && i < v.u.matrix_value.row * v.u.matrix_value.column && i >=0 ) {
            result.type = DOUBLE_VALUE;
            result.u.double_value=v.u.matrix_value.top[i];    
	    return result;
        } 
        else if (v.type == DIF_MATRIX_VALUE 
            && i < v.u.difmatrix_value.n && i >=0 ) 
        {
            result.type = DIF_VALUE;
            result.u.dif_value.value=v.u.difmatrix_value.value[i];    
            result.u.dif_value.jacobi=calloc(v.u.difmatrix_value.column,sizeof(double));
            for(k=0; k<v.u.difmatrix_value.column; k++)
            { 
                result.u.dif_value.jacobi[k]
                    =v.u.difmatrix_value.jacobi[i+k*v.u.difmatrix_value.n];    
            }
            result.u.dif_value.n=v.u.difmatrix_value.column;
	    return result;
        } 
        else 
        {
            clc_runtime_error(ARGUMENT_OUT_OF_RANGE_ERR, "(%s)\n", identifier);
        }
    }
    clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);

    return v; /* dummy. make compiler happy. */
}

Value eval_expression(LocalEnvironment *env, Expression *expr);
Value eval_function_expression(LocalEnvironment *env, Expression *expr);

static Value
eval_expression_list_expression(LocalEnvironment *env,
				Expression *expression, Expression *next)
{
    Value	v;

    v = eval_expression(env, expression);
    if (next) {
	v = eval_expression(env, next);
    }

    return v;
}

static Value
eval_function_expression_list_expression(LocalEnvironment *env,
				Expression *expression, Expression *next)
{
    Value	v;
/*
    v = eval_function_expression(env, expression);
    if (next) {
	v = eval_function_expression(env, next);
    }
*/
    v = eval_expression(env, expression);
    if (next) {
	v = eval_expression(env, next);
    }
    return v;
}

/* static Value
eval_copy_value(Value v)
{
    Value result;
    int m,n;
    int i;
    
    result.type = v.type;
    switch(v.type) 
    {
        case MATRIX_VALUE:
             n = v.u.matrix_value.column; 
             m = v.u.matrix_value.row; 
             result.u.matrix_value.column = n; 
             result.u.matrix_value.row = m; 
             result.u.matrix_value.top = GC_malloc_atomic(m*n*sizeof(double));
             for(i=0; i<m*n; ++i) 
             {
                 result.u.matrix_value.top[i] = v.u.matrix_value.top[i];
             } 
             break;
        case INT_VALUE:
        case DOUBLE_VALUE:
        case COMPLEX_VALUE:
        case INTERVAL_VALUE:
        case INTERVAL_MATRIX_VALUE:
        case COMPLEX_MATRIX_VALUE:
        case COUPLE_VALUE:
        case TRIPLE_VALUE:
        case NULL_VALUE:
        result = v;
    }
    return result;
} */

static Value
eval_assign_expression(LocalEnvironment *env,
		       char *identifier, Expression *expression)
{
    Value	v;
    Value	*left;
    v = eval_expression(env, expression);
    left = clc_search_local_variable(env, identifier);
    if (left == NULL) {
	left = clc_search_global_variable(identifier);
    }
    if (left != NULL) {
	/* Notice! Overwriting variable by pointer. */
	*left = v;
    } else {
	if (env != NULL) {
	    clc_add_local_variable(env, identifier, &v);
	} else {
	    clc_add_global_variable(identifier, &v);
	}
    }
    return v;
}

static Value
eval_function_assign_expression(LocalEnvironment *env,
		       char *identifier, Expression *expression)
{
    Value	v;
    Value	*left;

    v = eval_function_expression(env, expression);

    left = clc_search_local_variable(env, identifier);
    if (left != NULL) {
	/* Notice! Overwriting variable by pointer. */
	*left = v;
    } else {
        clc_add_local_variable(env, identifier, &v);
    }

    return v;
}

static void
eval_binary_interval(ExpressionType operator, 
                   double left_lower,  double left_upper,
                   double right_lower, double right_upper,Value *result)
{
    double t;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = INTERVAL_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        down();
	result->u.interval_value.lower = left_lower + right_lower;
        up();
	result->u.interval_value.upper = left_upper + right_upper;
	break;
      case SUB_EXPRESSION:
        down();
	result->u.interval_value.lower = left_lower - right_upper;
        up();
	result->u.interval_value.upper = left_upper - right_lower;
	break;
      case MUL_EXPRESSION:
        if (left_lower >= 0.) {
            if(right_lower >= 0.) {
                down();
                result->u.interval_value.lower = left_lower * right_lower;
                up();
                result->u.interval_value.upper = left_upper * right_upper;
            } else {
                if (right_upper >= 0.) {
                    down();
                    result->u.interval_value.lower = left_upper * right_lower;
                    up();
                    result->u.interval_value.upper = left_upper * right_upper;
                } else {
                    down();
                    result->u.interval_value.lower = left_upper * right_lower;
                    up();
                    result->u.interval_value.upper = left_lower * right_upper;
                }
            }
        } else {
            if (left_upper >= 0.) {
                if(right_lower >= 0.) {
                    down();
                    result->u.interval_value.lower = left_lower * right_upper;
                    up();
                    result->u.interval_value.upper = left_upper * right_upper;
                } else {
                    if(right_upper >= 0.) {
                        down();
                        result->u.interval_value.lower = left_lower * right_upper;
                        t = left_upper * right_lower;
                        if (result->u.interval_value.lower > t) {
                            result->u.interval_value.lower = t;
                        }
                        up();
                        result->u.interval_value.upper = left_lower * right_lower;
                        t = left_upper * right_upper;
                        if (result->u.interval_value.upper < t) {
                            result->u.interval_value.upper = t;
                        }
                    } else {
                        down();
                        result->u.interval_value.lower = left_upper * right_lower;
                        up();
                        result->u.interval_value.upper = left_lower * right_lower;
                    }
                }
            } else {
                if (right_lower >= 0.) {
                    down();
                    result->u.interval_value.lower = left_lower * right_upper;
                    up();
                    result->u.interval_value.upper = left_upper * right_lower;
                } else {
                    if (right_upper >= 0.) {
                        down();
                        result->u.interval_value.lower = left_lower * right_upper;
                        up();
                        result->u.interval_value.upper = left_lower * right_lower;
                    } else {
                        down();
                        result->u.interval_value.lower = left_upper * right_upper;
                        up();
                        result->u.interval_value.upper = left_lower * right_lower;
                    }
                 }
              }
           }
	break;
      case DIV_EXPRESSION:
        if (left_lower >= 0.) {
           if (right_lower > 0.) {
                down();
                result->u.interval_value.lower = left_lower / right_upper;
                up();
                result->u.interval_value.upper = left_upper / right_lower;
            } else {
                if (right_upper >= 0.) 
                    printf("Error:interval zero divid\n");
                else {
                    down();
                    result->u.interval_value.lower = left_upper / right_upper;
                    up();
                    result->u.interval_value.upper = left_lower / right_lower;
                }
            }
        } else {
            if(left_upper > 0.) {
                if (right_lower > 0) {
                    down();
                    result->u.interval_value.lower = left_lower / right_lower;
                    up();
                    result->u.interval_value.upper = left_upper / right_lower;
                } else {
                    if (right_upper >= 0.) 
                        printf("Error:interval zero divided\n");
                    else {
                        down();
                        result->u.interval_value.lower = left_upper / right_upper;
                        up();
                        result->u.interval_value.upper = left_lower / right_upper;
                    }
                }
            } else {
                if (right_lower > 0) {
                    down();
                    result->u.interval_value.lower = left_lower / right_lower;
                    up();
                    result->u.interval_value.upper = left_upper / right_upper;
                } else {
                    if (right_upper >= 0.) 
                        printf("Error:interval zero divided\n");
                    else {
                        down();
                        result->u.interval_value.upper = left_upper / right_lower;
                        up();
                        result->u.interval_value.lower = left_lower / right_upper;
                    }   
                 }
             }
          }
	break;
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static int
eval_binary_int(ExpressionType operator, int left, int right)
{
    int	result;

    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result = left + right;
	break;
      case SUB_EXPRESSION:
	result = left - right;
	break;
      case MUL_EXPRESSION:
	result = left * right;
	break;
      case DIV_EXPRESSION:
	result = left / right;
	break;
      case MOD_EXPRESSION:
	result = left % right;
	break;
      case EQ_EXPRESSION:
	result = left == right;
	break;
      case NE_EXPRESSION:
	result = left != right;
	break;
      case OR_EXPRESSION:
	result = left || right;
	break;
      case AND_EXPRESSION:
	result = left && right;
	break;
      case GT_EXPRESSION:
	result = left > right;
	break;
      case GE_EXPRESSION:
	result = left >= right;
	break;
      case LT_EXPRESSION:
	result = left < right;
	break;
      case LE_EXPRESSION:
	result = left <= right;
	break;
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      default:
    }
    return result;
}

static void
eval_binary_double(ExpressionType operator, double left, double right,
		   Value *result)
{
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = DOUBLE_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result->u.double_value = left + right;
	break;
      case SUB_EXPRESSION:
	result->u.double_value = left - right;
	break;
      case MUL_EXPRESSION:
	result->u.double_value = left * right;
	break;
      case DIV_EXPRESSION:
	result->u.double_value = left / right;
	break;
      case MOD_EXPRESSION:
	result->u.double_value = fmod(left, right);
	break;
      case EQ_EXPRESSION:
	result->u.int_value = left == right;
	break;
      case NE_EXPRESSION:
	result->u.int_value = left != right;
	break;
      case OR_EXPRESSION:
	result->u.int_value = left || right;
	break;
      case AND_EXPRESSION:
	result->u.int_value = left && right;
	break;
      case GT_EXPRESSION:
	result->u.int_value = left > right;
	break;
      case GE_EXPRESSION:
	result->u.int_value = left >= right;
	break;
      case LT_EXPRESSION:
	result->u.int_value = left < right;
	break;
      case LE_EXPRESSION:
	result->u.int_value = left <= right;
	break;
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_complex_double(ExpressionType operator, 
                   double left_r, double left_i,
		   double right, Value *result)
{
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = COMPLEX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result->u.complex_value.r = left_r + right;
	result->u.complex_value.i = left_i + right;
	break;
      case SUB_EXPRESSION:
	result->u.complex_value.r = left_r - right;
	result->u.complex_value.i = left_i - right;
	break;
      case MUL_EXPRESSION:
	result->u.complex_value.r = left_r * right;
	result->u.complex_value.i = left_i * right;
	break;
      case DIV_EXPRESSION:
	result->u.complex_value.r = left_r/right;
	result->u.complex_value.i = left_i/right;
	break;
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_double_complex(ExpressionType operator, double left, double right_r,
		   double right_i,Value *result)
{
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = COMPLEX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result->u.complex_value.r = left + right_r;
	result->u.complex_value.i = right_i;
	break;
      case SUB_EXPRESSION:
	result->u.complex_value.r = left - right_r;
	result->u.complex_value.i = - right_i;
	break;
      case MUL_EXPRESSION:
	result->u.complex_value.r = left * right_r;
	result->u.complex_value.i = left * right_i;
	break;
      case DIV_EXPRESSION:
	result->u.complex_value.r = 
                  (left*right_r) / (right_r*right_r+right_i*right_i);
	result->u.complex_value.i = 
                  -(left*right_i) / (right_r*right_r+right_i*right_i);
	break;
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_complex_complex(ExpressionType operator, 
                   double left_r,  double left_i,
                   double right_r, double right_i,Value *result)
{
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = COMPLEX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
	result->u.complex_value.r = left_r + right_r;
	result->u.complex_value.i = left_i + right_i;
	break;
      case SUB_EXPRESSION:
	result->u.complex_value.r = left_r - right_r;
	result->u.complex_value.i = left_i - right_i;
	break;
      case MUL_EXPRESSION:
	result->u.complex_value.r = left_r * right_r - left_i * right_i;
	result->u.complex_value.i = left_r * right_i + left_i * right_r;
	break;
      case DIV_EXPRESSION:
	result->u.complex_value.r = 
                  (left_r * right_r + left_i * right_i) / (right_r*right_r+right_i*right_i);
	result->u.complex_value.i = 
                  (-left_r * right_i + left_i * right_r) / (right_r*right_r+right_i*right_i);
	break;
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_intervalmatrix(ExpressionType operator, Intervalmatr left, Intervalmatr right,
		   Value *result)
{
    int i;
    long int m,n,k;
    double *center_left, *radius_left;
    double *center_right, *radius_right;
    double *temp_1, *temp_2, *temp_3;
    double *abs_left, *abs_right;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION) {
	result->type = INTERVAL_MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        down();
        for(i=0; i<left.upper.row*left.upper.column; ++i) {
	     result->u.interval_matrix_value.lower.top[i] = 
                        left.lower.top[i]+right.lower.top[i];
        }
        up();
        for(i=0; i<left.upper.row*left.upper.column; ++i) {
	     result->u.interval_matrix_value.upper.top[i] = 
                        left.upper.top[i]+right.upper.top[i];
        }
        (*result).u.interval_matrix_value.lower.column = left.upper.column;
        (*result).u.interval_matrix_value.upper.column = left.upper.column;
        (*result).u.interval_matrix_value.lower.row    = left.upper.row;
        (*result).u.interval_matrix_value.upper.row    = left.upper.row;
	break; 
      case SUB_EXPRESSION:
        down();
        for(i=0; i<left.upper.row*left.upper.column; ++i) {
	     result->u.interval_matrix_value.lower.top[i] = 
                        left.lower.top[i]-right.upper.top[i];
        }
        up();
        for(i=0; i<left.upper.row*left.upper.column; ++i) {
	     result->u.interval_matrix_value.upper.top[i] = 
                        left.upper.top[i]-right.lower.top[i];
        }
        (*result).u.interval_matrix_value.lower.column = left.upper.column;
        (*result).u.interval_matrix_value.upper.column = left.upper.column;
        (*result).u.interval_matrix_value.lower.row    = left.upper.row;
        (*result).u.interval_matrix_value.upper.row    = left.upper.row;
	break; 
      case MUL_EXPRESSION:

        m = (long int) left.upper.row;
        n = (long int) right.upper.column;
        k = (long int) left.upper.column;
        
        center_left     = (double*)GC_malloc_atomic(m*k*sizeof(double));
        center_right    = (double*)GC_malloc_atomic(k*n*sizeof(double));
        radius_left     = (double*)GC_malloc_atomic(m*k*sizeof(double));
        radius_right    = (double*)GC_malloc_atomic(k*n*sizeof(double));
        temp_1          = (double*)GC_malloc_atomic(m*n*sizeof(double));
        temp_2          = (double*)GC_malloc_atomic(m*n*sizeof(double));
        temp_3          = (double*)GC_malloc_atomic(m*n*sizeof(double));
        abs_left        = (double*)GC_malloc_atomic(m*k*sizeof(double));
        abs_right       = (double*)GC_malloc_atomic(k*n*sizeof(double));

        up();
        for(i=0; i<k*n; ++i) {
            center_right[i] = (right.lower.top[i]+right.upper.top[i])*0.5;
            radius_right[i] = center_right[i]-right.lower.top[i];
        }
        for(i=0; i<m*k; ++i) {
            center_left[i]  = (left.lower.top[i]+left.upper.top[i])*0.5;
            radius_left[i]  = center_left[i]-left.lower.top[i];
        }
        slb_matrix_multyply (center_left,center_right,
                             result->u.interval_matrix_value.upper.top,m,n,k);
        for(i=0; i<m*k; ++i) {
           abs_left[i] = fabs(center_left[i]);
        }

        for(i=0; i<k*n; ++i) {
           abs_right[i] = fabs(center_right[i]);
        }

        slb_matrix_multyply (abs_left,radius_right,temp_1,m,n,k);

        slb_matrix_multyply (radius_left,abs_right,temp_2,m,n,k);

        slb_matrix_multyply (radius_left,radius_right,temp_3,m,n,k);

        for(i=0;i<m*n;++i) {
            result->u.interval_matrix_value.upper.top[i] 
                        += (temp_1[i]+temp_2[i]+temp_3[i]); 
        }

        down();
        slb_matrix_multyply (center_left,center_right,
                             result->u.interval_matrix_value.lower.top,m,n,k);
        
        for(i=0; i<m*k; ++i) {
           abs_left[i] *= -1.0;
        }
                
        for(i=0; i<k*n; ++i) {
           abs_right[i] *= -1.0;
        }

        slb_matrix_multyply (abs_left,radius_right,temp_1,m,n,k);

        slb_matrix_multyply (radius_left,abs_right,temp_2,m,n,k);

        for(i=0; i<k*n; ++i) {
           radius_right[i] *= -1.0;
        }

        slb_matrix_multyply (radius_left,radius_right,temp_3,m,n,k);

        for(i=0;i<m*n;++i) {
            result->u.interval_matrix_value.lower.top[i] 
                        += (temp_1[i]+temp_2[i]+temp_3[i]); 
        }

        (*result).u.interval_matrix_value.lower.column = n;
        (*result).u.interval_matrix_value.upper.column = n;
        (*result).u.interval_matrix_value.lower.row = m;
        (*result).u.interval_matrix_value.upper.row = m;

        GC_free(center_left);
        GC_free(center_right);
        GC_free(radius_left);
        GC_free(radius_right);
        GC_free(temp_1);
        GC_free(temp_2);
        GC_free(temp_3);
        GC_free(abs_left);
        GC_free(abs_right);

	break; 
      case SOLVE_EXPRESSION:
/*        slab_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; */
      case DIV_EXPRESSION:
/*        slab_app_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break;  */
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_matrix_intervalmatrix(ExpressionType operator, matr left, Intervalmatr right,
		   Value *result)
{
    int i;
    long int m,n,k;
    double *center, *radius, *temp, *abs_left;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION) {
	result->type = INTERVAL_MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        down();
        for(i=0; i<left.row*left.column; ++i) {
	     result->u.interval_matrix_value.lower.top[i] = 
                        left.top[i]+right.lower.top[i];
        }
        up();
        for(i=0; i<left.row*left.column; ++i) {
	     result->u.interval_matrix_value.upper.top[i] = 
                        left.top[i]+right.upper.top[i];
        }
        (*result).u.interval_matrix_value.lower.column = left.column;
        (*result).u.interval_matrix_value.upper.column = left.column;
        (*result).u.interval_matrix_value.lower.row = left.row;
        (*result).u.interval_matrix_value.upper.row = left.row;
	break; 
      case SUB_EXPRESSION:
        down();
        for(i=0; i<left.row*left.column; ++i) {
	     result->u.interval_matrix_value.lower.top[i] = 
                        left.top[i]-right.upper.top[i];
        }
        up();
        for(i=0; i<left.row*left.column; ++i) {
	     result->u.interval_matrix_value.upper.top[i] = 
                        left.top[i]-right.lower.top[i];
        }
        (*result).u.interval_matrix_value.lower.column = left.column;
        (*result).u.interval_matrix_value.upper.column = left.column;
        (*result).u.interval_matrix_value.lower.row = left.row;
        (*result).u.interval_matrix_value.upper.row = left.row;
	break; 
      case MUL_EXPRESSION:
        m = (long int) left.row;
        n = (long int) right.upper.column;
        k = (long int) left.column;
        
        center    = (double*)GC_malloc_atomic(n*k*sizeof(double));
        radius    = (double*)GC_malloc_atomic(n*k*sizeof(double));
        temp      = (double*)GC_malloc_atomic(m*n*sizeof(double));
        abs_left  = (double*)GC_malloc_atomic(m*k*sizeof(double));

        up();
        for(i=0; i<k*n; ++i) {
            center[i] = (right.lower.top[i]+right.upper.top[i])*0.5;
            radius[i] = center[i]-right.lower.top[i];
        }
        slb_matrix_multyply (left.top,center,
                             result->u.interval_matrix_value.upper.top,m,n,k);
        for(i=0; i<m*k; ++i) {
           abs_left[i] = fabs(left.top[i]);
        }

        slb_matrix_multyply (abs_left,radius,temp,m,n,k);

        for(i=0;i<m*n;++i) {
            result->u.interval_matrix_value.upper.top[i] += temp[i]; 
        }

        down();
        slb_matrix_multyply (left.top,center,
                             result->u.interval_matrix_value.lower.top,m,n,k);
        
        for(i=0; i<m*k; ++i) {
           abs_left[i] *= -1.0;
        }
                
        slb_matrix_multyply (abs_left,radius,temp,m,n,k);
        
        for(i=0;i<m*n;++i) {
            result->u.interval_matrix_value.lower.top[i] += temp[i]; 
        }

        (*result).u.interval_matrix_value.lower.column = n;
        (*result).u.interval_matrix_value.upper.column = n;
        (*result).u.interval_matrix_value.lower.row = m;
        (*result).u.interval_matrix_value.upper.row = m;

        GC_free(center);
        GC_free(radius);
        GC_free(temp);
        GC_free(abs_left);

	break; 
      case SOLVE_EXPRESSION:
/*        slab_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; */
      case DIV_EXPRESSION:
/*        slab_app_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break;  */
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_intervalmatrix_matrix(ExpressionType operator, Intervalmatr left, matr right,
		   Value *result)
{
    int i, j;
    long int m,n,k;
    double *center, *radius, *temp, *abs_right;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION) {
	result->type = INTERVAL_MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        down();
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.interval_matrix_value.lower.top[i+j*right.row] = 
                        left.lower.top[i+j*right.row]+right.top[i+j*right.row];
             }
        }
        up();
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.interval_matrix_value.upper.top[i+j*right.row] = 
                        left.upper.top[i+j*right.row]+right.top[i+j*right.row];
             }
        }
        (*result).u.interval_matrix_value.lower.column = right.column;
        (*result).u.interval_matrix_value.upper.column = right.column;
        (*result).u.interval_matrix_value.lower.row = right.row;
        (*result).u.interval_matrix_value.upper.row = right.row;
	break; 
      case SUB_EXPRESSION:
        down();
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.interval_matrix_value.lower.top[i+j*right.row] = 
                        left.lower.top[i+j*right.row]-right.top[i+j*right.row];
             }
        }
        up();
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.interval_matrix_value.upper.top[i+j*right.row] = 
                        left.upper.top[i+j*right.row]-right.top[i+j*right.row];
             }
        }
        (*result).u.interval_matrix_value.lower.column = right.column;
        (*result).u.interval_matrix_value.upper.column = right.column;
        (*result).u.interval_matrix_value.lower.row = right.row;
        (*result).u.interval_matrix_value.upper.row = right.row;
	break; 
      case MUL_EXPRESSION:

        m = (long int) left.upper.row;
        n = (long int) right.column;
        k = (long int) left.upper.column;
        
        center    = (double*)GC_malloc_atomic(m*k*sizeof(double));
        radius    = (double*)GC_malloc_atomic(m*k*sizeof(double));
        temp      = (double*)GC_malloc_atomic(m*n*sizeof(double));
        abs_right = (double*)GC_malloc_atomic(k*n*sizeof(double));

        up();
        for(i=0; i<m*k; ++i) {
            center[i] = (left.lower.top[i]+left.upper.top[i])*0.5;
            radius[i] = center[i]-left.lower.top[i];
        }
        slb_matrix_multyply (center,right.top,
                             result->u.interval_matrix_value.upper.top,m,n,k);
        for(i=0; i<n*k; ++i) {
           abs_right[i] = fabs(right.top[i]);
        }

        slb_matrix_multyply (radius,abs_right,temp,m,n,k);

        for(i=0;i<m*n;++i) {
            result->u.interval_matrix_value.upper.top[i] += temp[i]; 
        }

        down();
        slb_matrix_multyply (center,right.top,
                             result->u.interval_matrix_value.lower.top,m,n,k);
        
        for(i=0; i<n*k; ++i) {
           abs_right[i] *= -1.0;
        }
                
        slb_matrix_multyply (radius,abs_right,temp,m,n,k);
        
        for(i=0;i<m*n;++i) {
            result->u.interval_matrix_value.lower.top[i] += temp[i]; 
        }

        (*result).u.interval_matrix_value.lower.column = n;
        (*result).u.interval_matrix_value.upper.column = n;
        (*result).u.interval_matrix_value.lower.row = m;
        (*result).u.interval_matrix_value.upper.row = m;

        GC_free(center);
        GC_free(radius);
        GC_free(temp);
        GC_free(abs_right);

	break; 
      case SOLVE_EXPRESSION:
/*        slab_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; */
      case DIV_EXPRESSION:
/*        slab_app_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break;  */
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_matrix(ExpressionType operator, matr left, matr right,
		   Value *result)
{
    int i, j;
    long int m,n,k;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION
        || operator == EXACT_SOLVE_EXPRESSION) {
	result->type = MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.matrix_value.top[i+j*left.row] = 
                        left.top[i+j*left.row]+right.top[i+j*left.row];
             }
        }
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case SUB_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
	         result->u.matrix_value.top[i+j*left.row] = 
                        left.top[i+j*left.row]-right.top[i+j*left.row];
             }
        }
        (*result).u.matrix_value.column = left.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case MUL_EXPRESSION:
/*
        for(i=0; i<left.row; ++i) {
             for(j=0; j<right.column; ++j) {
                 result->u.matrix_value.top[i+j*left.row] = 0.0;
                 for(k=0; k<left.column; ++k) {
	             result->u.matrix_value.top[i+j*left.row] += 
                        left.top[i+k*left.row]*right.top[k+j*right.row];
                 }
             }
        }
*/
        m = (long int) left.row;
        n = (long int) right.column;
        k = (long int) left.column;
        slb_matrix_multyply (left.top,right.top,result->u.matrix_value.top,m,n,k);
    
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;

	break; 
      case SOLVE_EXPRESSION:
        slab_app_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row,right.column);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case EXACT_SOLVE_EXPRESSION:
        slab_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row, right.column);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case DIV_EXPRESSION:
        slab_app_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row, right.column);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_complex_matrix(ExpressionType operator, Complexmatr left, Complexmatr right,
		   Value *result)
{
    int i, j;
    long int m,n,k;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION || operator == SOLVE_EXPRESSION
        || operator == EXACT_SOLVE_EXPRESSION) {
	result->type = COMPLEX_MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.complex_matrix_value.top[i+j*left.row].r = 
                        left.top[i+j*left.row].r+right.top[i+j*left.row].r;
	         result->u.complex_matrix_value.top[i+j*left.row].i = 
                        left.top[i+j*left.row].i+right.top[i+j*left.row].i;
             }
        }
        (*result).u.complex_matrix_value.column = right.column;
        (*result).u.complex_matrix_value.row = left.row;
	break; 
      case SUB_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
	         result->u.complex_matrix_value.top[i+j*left.row].r = 
                        left.top[i+j*left.row].r-right.top[i+j*left.row].r;
	         result->u.complex_matrix_value.top[i+j*left.row].i = 
                        left.top[i+j*left.row].i-right.top[i+j*left.row].i;
             }
        }
        (*result).u.complex_matrix_value.column = left.column;
        (*result).u.complex_matrix_value.row = left.row;
	break; 
      case MUL_EXPRESSION:
/*
        for(i=0; i<left.row; ++i) {
             for(j=0; j<right.column; ++j) {
                 result->u.matrix_value.top[i+j*left.row] = 0.0;
                 for(k=0; k<left.column; ++k) {
	             result->u.matrix_value.top[i+j*left.row] += 
                        left.top[i+k*left.row]*right.top[k+j*right.row];
                 }
             }
        } */
        m = (long int) left.row;
        n = (long int) right.column;
        k = (long int) left.column;
        slb_complex_matrix_multyply (left.top,right.top,result->u.complex_matrix_value.top,m,n,k);
    
        (*result).u.complex_matrix_value.column = n;
        (*result).u.complex_matrix_value.row = m;

	break; 
      case SOLVE_EXPRESSION:
        slab_app_solve_complex_matrix_equation(left.top, right.top, 
                                   result->u.complex_matrix_value.top, left.row, right.column);
        (*result).u.complex_matrix_value.column = right.column;
        (*result).u.complex_matrix_value.row = left.row;
	break; 
/*      case EXACT_SOLVE_EXPRESSION:
        slab_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case DIV_EXPRESSION:
        slab_app_solve_matrix_equation(left.top, right.top, 
                                   result->u.matrix_value.top, left.row);
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = left.row;
	break; */ 
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_double_matrix(ExpressionType operator, double left, matr right,
		   Value *result)
{
    int i, j;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.matrix_value.top[i+j*right.row] = 
                        left+right.top[i+j*right.row];
             }
        }
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = right.row;
	break; 
      case SUB_EXPRESSION:
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.matrix_value.top[i+j*right.row] = 
                        left-right.top[i+j*right.row];
             }
        }
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = right.row;
	break; 
      case MUL_EXPRESSION:
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
                 result->u.matrix_value.top[i+j*right.row] =
                        left*right.top[i+j*right.row];
             }
        }
        (*result).u.matrix_value.column = right.column;
        (*result).u.matrix_value.row = right.row;
	break; 
      case DIV_EXPRESSION:
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_complex_complexmatrix(ExpressionType operator, Complex left, Complexmatr right,
		   Value *result)
{
    int i, j;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = COMPLEX_MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.complex_matrix_value.top[i+j*right.row].r = 
                        left.r+right.top[i+j*right.row].r;
	         result->u.complex_matrix_value.top[i+j*right.row].i = 
                        left.i+right.top[i+j*right.row].i;
             }
        }
        (*result).u.complex_matrix_value.column = right.column;
        (*result).u.complex_matrix_value.row = right.row;
	break; 
      case SUB_EXPRESSION:
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
	         result->u.complex_matrix_value.top[i+j*right.row].r = 
                        left.r-right.top[i+j*right.row].r;
	         result->u.complex_matrix_value.top[i+j*right.row].i = 
                        left.i-right.top[i+j*right.row].i;
             }
        }
        (*result).u.complex_matrix_value.column = right.column;
        (*result).u.complex_matrix_value.row = right.row;
	break; 
      case MUL_EXPRESSION:
        for(i=0; i<right.row; ++i) {
             for(j=0; j<right.column; ++j) {
                 result->u.complex_matrix_value.top[i+j*right.row].r =
                  left.r*right.top[i+j*right.row].r-left.i*right.top[i+j*right.row].i;
                 result->u.complex_matrix_value.top[i+j*right.row].i =
                  left.i*right.top[i+j*right.row].r+left.r*right.top[i+j*right.row].i;
             }
        }
        (*result).u.complex_matrix_value.column = right.column;
        (*result).u.complex_matrix_value.row = right.row;
	break; 
      case DIV_EXPRESSION:
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_matrix_double(ExpressionType operator, matr left, double right,
		   Value *result)
{
    int i, j;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
	         result->u.matrix_value.top[i+j*left.row] = 
                        left.top[i+j*left.row]+right;
             }
        }
        (*result).u.matrix_value.column = left.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case SUB_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
	         result->u.matrix_value.top[i+j*left.row] = 
                        left.top[i+j*left.row]-right;
             }
        }
        (*result).u.matrix_value.column = left.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case MUL_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
                 result->u.matrix_value.top[i+j*left.row] =
                        left.top[i+j*left.row]*right;
             }
        }
        (*result).u.matrix_value.column = left.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case DIV_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
                 result->u.matrix_value.top[i+j*left.row] =
                        left.top[i+j*left.row]/right;
             }
        }
        (*result).u.matrix_value.column = left.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

static void
eval_binary_complexmatrix_complex(ExpressionType operator, Complexmatr left, Complex right,
		   Value *result)
{
    int i, j;
    if (operator == ADD_EXPRESSION || operator == SUB_EXPRESSION
	|| operator == MUL_EXPRESSION || operator == DIV_EXPRESSION
	|| operator == MOD_EXPRESSION) {
	result->type = COMPLEX_MATRIX_VALUE;
    } else {
	result->type = INT_VALUE;
    }
    switch (operator) {
      case INT_EXPRESSION:	/* FALLTHRU */
      case DOUBLE_EXPRESSION:	/* FALLTHRU */
      case IDENTIFIER_EXPRESSION:	/* FALLTHRU */
      case EXPRESSION_LIST_EXPRESSION:	/* FALLTHRU */
      case ASSIGN_EXPRESSION:
	break;
      case ADD_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
	         result->u.complex_matrix_value.top[i+j*left.row].r = 
                        left.top[i+j*left.row].r+right.r;
	         result->u.complex_matrix_value.top[i+j*left.row].i = 
                        left.top[i+j*left.row].i+right.i;
             }
        }
        (*result).u.complex_matrix_value.column = left.column;
        (*result).u.complex_matrix_value.row = left.row;
	break; 
      case SUB_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
	         result->u.complex_matrix_value.top[i+j*left.row].r = 
                        left.top[i+j*left.row].r-right.r;
	         result->u.complex_matrix_value.top[i+j*left.row].i = 
                        left.top[i+j*left.row].i-right.i;
             }
        }
        (*result).u.complex_matrix_value.column = left.column;
        (*result).u.complex_matrix_value.row = left.row;
	break; 
      case MUL_EXPRESSION:
        for(i=0; i<left.row; ++i) {
             for(j=0; j<left.column; ++j) {
                 result->u.complex_matrix_value.top[i+j*left.row].r =
                   left.top[i+j*left.row].r*right.r-left.top[i+j*left.row].i*right.i;
                 result->u.complex_matrix_value.top[i+j*left.row].i =
                   left.top[i+j*left.row].i*right.r+left.top[i+j*left.row].r*right.i;
             }
        }
        (*result).u.matrix_value.column = left.column;
        (*result).u.matrix_value.row = left.row;
	break; 
      case DIV_EXPRESSION:
      case MOD_EXPRESSION:
      case EQ_EXPRESSION:
      case NE_EXPRESSION:
      case GT_EXPRESSION:
      case GE_EXPRESSION:
      case LT_EXPRESSION:
      case LE_EXPRESSION:
      case MINUS_EXPRESSION:	/* FALLTHRU */
      case IF_EXPRESSION:	/* FALLTHRU */
      case WHILE_EXPRESSION:	/* FALLTHRU */
      case FUNCTION_CALL_EXPRESSION:	/* FALLTHRU */
      case MATRIX_EXPRESSION:
      case MATRIX_ELEMENT_EXPRESSION:
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
}

Value
clc_eval_binary_expression(LocalEnvironment *env,
			   ExpressionType operator,
			   Expression *left, Expression *right)
{
    Value	left_val;
    Value	right_val;
    Value	result;
    Complex     *temp_left, *temp_right;
    int i;
    left_val = eval_expression(env, left);
    right_val = eval_expression(env, right);

    if (left_val.type == DIF_VALUE && right_val.type == DOUBLE_VALUE) {
        result.type = DIF_VALUE;
        result.u.dif_value.jacobi =
                 (double*)calloc(left_val.u.dif_value.n,sizeof(double));
        eval_binary_dif_double(operator, left_val.u.dif_value, right_val.u.double_value, &result); 
    } else if (left_val.type == DOUBLE_VALUE && right_val.type == DIF_VALUE) {
        result.type = DIF_VALUE;
        result.u.dif_value.jacobi =
                 (double*)calloc(right_val.u.dif_value.n,sizeof(double));
        eval_binary_double_dif(operator, left_val.u.double_value, right_val.u.dif_value, &result); 
    } else if (left_val.type == DIF_VALUE && right_val.type == DIF_VALUE) {
        result.type = DIF_VALUE;
        result.u.dif_value.jacobi =
                 (double*)calloc(left_val.u.dif_value.n,sizeof(double));
        eval_binary_dif(operator, left_val.u.dif_value, right_val.u.dif_value, &result); 
    } else if (left_val.type == MATRIX_VALUE && right_val.type == COMPLEX_VALUE) {
        result.u.complex_matrix_value.top =
                 (Complex*)GC_malloc(left_val.u.matrix_value.column*left_val.u.matrix_value.row*sizeof(Complex));
        temp_left =
                 (Complex*)GC_malloc(left_val.u.matrix_value.column*left_val.u.matrix_value.row*sizeof(Complex));
        for (i=0; i < left_val.u.matrix_value.column*left_val.u.matrix_value.row;++i) {
            temp_left[i].r = left_val.u.matrix_value.top[i];
            temp_left[i].i = 0.0;    
        }
        left_val.type = COMPLEX_MATRIX_VALUE;
        left_val.u.complex_matrix_value.top = temp_left;
        eval_binary_complexmatrix_complex(operator,
                         left_val.u.complex_matrix_value, right_val.u.complex_value,
                         &result); 
        GC_free(temp_left);
    } else if (left_val.type == COMPLEX_VALUE && right_val.type == MATRIX_VALUE) {
        result.u.complex_matrix_value.top =
                 (Complex *)calloc(right_val.u.matrix_value.column*right_val.u.matrix_value.row,sizeof(Complex));
        temp_right = 
                 (Complex *)GC_malloc(right_val.u.matrix_value.column*right_val.u.matrix_value.row*sizeof(Complex));
        for (i=0; i < right_val.u.matrix_value.column*right_val.u.matrix_value.row;++i) {
            temp_right[i].r = right_val.u.matrix_value.top[i];
            temp_right[i].i = 0.0;    
        }
        right_val.type = COMPLEX_MATRIX_VALUE;
        right_val.u.complex_matrix_value.top = temp_right;
        eval_binary_complex_complexmatrix(operator,
                         left_val.u.complex_value, right_val.u.complex_matrix_value,
                         &result);
        GC_free(temp_right);
    } else if (left_val.type == COMPLEX_MATRIX_VALUE && right_val.type == DOUBLE_VALUE) {
        result.u.complex_matrix_value.top =
                 (Complex*)calloc(left_val.u.complex_matrix_value.column*left_val.u.complex_matrix_value.row,sizeof(Complex));
        right_val.type = COMPLEX_VALUE;
        right_val.u.complex_value.r = right_val.u.double_value;
        right_val.u.complex_value.i = 0.0;
        eval_binary_complexmatrix_complex(operator,
                         left_val.u.complex_matrix_value, right_val.u.complex_value,
                         &result); 
    } else if (left_val.type == DOUBLE_VALUE && right_val.type == COMPLEX_MATRIX_VALUE) {
        result.u.complex_matrix_value.top =
                 (Complex *)calloc(right_val.u.complex_matrix_value.column*right_val.u.complex_matrix_value.row,sizeof(Complex));
        left_val.type = COMPLEX_VALUE;
        left_val.u.complex_value.r = left_val.u.double_value;
        left_val.u.complex_value.i = 0.0;

        eval_binary_complex_complexmatrix(operator,
                         left_val.u.complex_value, right_val.u.complex_matrix_value,
                         &result);
    } else if (left_val.type == COMPLEX_VALUE && right_val.type == COMPLEX_MATRIX_VALUE) {
        result.u.complex_matrix_value.top =
                 (Complex *)calloc(right_val.u.complex_matrix_value.column*right_val.u.complex_matrix_value.row,sizeof(Complex));
        eval_binary_complex_complexmatrix(operator,
                         left_val.u.complex_value, right_val.u.complex_matrix_value,
                         &result);
    } else if (left_val.type == COMPLEX_MATRIX_VALUE && right_val.type == COMPLEX_VALUE) {
        result.u.complex_matrix_value.top =
                 (Complex*)calloc(left_val.u.complex_matrix_value.column*left_val.u.complex_matrix_value.row,sizeof(Complex));
        eval_binary_complexmatrix_complex(operator,
                         left_val.u.complex_matrix_value, right_val.u.complex_value,
                         &result); 
    } else if (left_val.type == COMPLEX_MATRIX_VALUE && right_val.type == MATRIX_VALUE) {
        result.u.complex_matrix_value.top = 
                 (Complex *)GC_malloc(right_val.u.matrix_value.column*left_val.u.complex_matrix_value.row*sizeof(Complex));
        temp_right=(Complex *)GC_malloc(right_val.u.matrix_value.row*right_val.u.matrix_value.column*sizeof(Complex));
        for(i=0; i<right_val.u.matrix_value.row*right_val.u.matrix_value.column;++i) {
            temp_right[i].r = right_val.u.matrix_value.top[i];
            temp_right[i].i = 0.0;
        }
        right_val.type = COMPLEX_MATRIX_VALUE;
        right_val.u.complex_matrix_value.top = temp_right;
        eval_binary_complex_matrix(operator,
                         left_val.u.complex_matrix_value, right_val.u.complex_matrix_value,
                         &result);
        GC_free(temp_right);
    } else if (left_val.type == MATRIX_VALUE && right_val.type == COMPLEX_MATRIX_VALUE) {
        result.u.complex_matrix_value.top = 
                 (Complex *)GC_malloc(right_val.u.complex_matrix_value.column*left_val.u.matrix_value.row*sizeof(Complex));
        temp_left=(Complex *)GC_malloc(left_val.u.matrix_value.row*left_val.u.matrix_value.column*sizeof(Complex));
        for(i=0; i<left_val.u.matrix_value.row*left_val.u.matrix_value.column;++i) {
            temp_left[i].r = left_val.u.matrix_value.top[i];
            temp_left[i].i = 0.0;
        }
        left_val.type = COMPLEX_MATRIX_VALUE;
        left_val.u.complex_matrix_value.top = temp_left;
        eval_binary_complex_matrix(operator,
                         left_val.u.complex_matrix_value, right_val.u.complex_matrix_value,
                         &result);
        GC_free(temp_left);
    } else if (left_val.type == COMPLEX_MATRIX_VALUE && right_val.type == COMPLEX_MATRIX_VALUE) {
        result.u.complex_matrix_value.top = 
                 (Complex *)GC_malloc(right_val.u.complex_matrix_value.column*left_val.u.complex_matrix_value.row*sizeof(Complex));
        eval_binary_complex_matrix(operator,
                         left_val.u.complex_matrix_value, right_val.u.complex_matrix_value,
                         &result);
    } else if (left_val.type == INTERVAL_MATRIX_VALUE && right_val.type == INTERVAL_MATRIX_VALUE) {
        result.u.interval_matrix_value.lower.top = 
                 (double*)calloc(right_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        result.u.interval_matrix_value.upper.top = 
                 (double*)calloc(right_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        eval_binary_intervalmatrix(operator,
                         left_val.u.interval_matrix_value,
                         right_val.u.interval_matrix_value,
                         &result);
    } else if (left_val.type == MATRIX_VALUE && right_val.type == INTERVAL_MATRIX_VALUE) {
        result.u.interval_matrix_value.lower.top = 
                 (double*)calloc(right_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        result.u.interval_matrix_value.upper.top = 
                 (double*)calloc(right_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        eval_binary_matrix_intervalmatrix(operator,
                         left_val.u.matrix_value,
                         right_val.u.interval_matrix_value,
                         &result);
    } else if (left_val.type == INTERVAL_MATRIX_VALUE && right_val.type == MATRIX_VALUE) {
        result.u.interval_matrix_value.lower.top = 
                 (double*)calloc(right_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        result.u.interval_matrix_value.upper.top = 
                 (double*)calloc(right_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        eval_binary_intervalmatrix_matrix(operator,
                         left_val.u.interval_matrix_value,
                         right_val.u.matrix_value,
                         &result);
    } else if (left_val.type == INTERVAL_VALUE && right_val.type == INTERVAL_VALUE) {
        eval_binary_interval(operator,
                         left_val.u.interval_value.lower,
                         left_val.u.interval_value.upper,
                         right_val.u.interval_value.lower,
                         right_val.u.interval_value.upper,
                         &result);
    } else if (left_val.type == MATRIX_VALUE && right_val.type == MATRIX_VALUE) {
        result.u.matrix_value.top = 
                 (double*)calloc(right_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        eval_binary_matrix(operator,
                         left_val.u.matrix_value, right_val.u.matrix_value,
                         &result);
    } else {if (left_val.type == DOUBLE_VALUE 
                     && right_val.type == MATRIX_VALUE) {
       result.u.matrix_value.top =
                 (double*)calloc(right_val.u.matrix_value.column*right_val.u.matrix_value.row,sizeof(double));
        eval_binary_double_matrix(operator,
                         left_val.u.double_value, right_val.u.matrix_value,
                         &result);
    } else {if (left_val.type == MATRIX_VALUE
                     && right_val.type == DOUBLE_VALUE) {
       result.u.matrix_value.top =
                 (double*)calloc(left_val.u.matrix_value.column*left_val.u.matrix_value.row,sizeof(double));
        eval_binary_matrix_double(operator,
                         left_val.u.matrix_value, right_val.u.double_value,
                         &result);
    } else {
        if (left_val.type == INT_VALUE
	    && right_val.type == INT_VALUE) {
	    result.type = INT_VALUE;
	    result.u.int_value
	        = eval_binary_int(operator,
			      left_val.u.int_value, right_val.u.int_value);
        } else if (left_val.type == DOUBLE_VALUE
	           && right_val.type == DOUBLE_VALUE) {
	    eval_binary_double(operator,
			   left_val.u.double_value, right_val.u.double_value,
			   &result);
        } else if (left_val.type == DOUBLE_VALUE
                   && right_val.type == COMPLEX_VALUE) {
            eval_binary_double_complex(operator,
                           left_val.u.double_value,
                           right_val.u.complex_value.r,
                           right_val.u.complex_value.i,&result);
        } else if (left_val.type == COMPLEX_VALUE
                   && right_val.type == DOUBLE_VALUE) {
            eval_binary_complex_double(operator,
                           left_val.u.complex_value.r,
                           left_val.u.complex_value.i,
                           right_val.u.double_value,&result);
        } else if (left_val.type == COMPLEX_VALUE
                   && right_val.type == COMPLEX_VALUE) {
            eval_binary_complex_complex(operator,
                           left_val.u.complex_value.r,
                           left_val.u.complex_value.i,
                           right_val.u.complex_value.r,
                           right_val.u.complex_value.i,&result);
        } else {
	    /* cast int to double */
	    if (left_val.type == INT_VALUE) {
	        left_val.u.double_value = left_val.u.int_value;
	    } else {
	        right_val.u.double_value = right_val.u.int_value;
	    }
	    eval_binary_double(operator,
			   left_val.u.double_value, right_val.u.double_value,
			   &result);
        }
    }
    }
    }
    return result;
}

static clock_t tictime;

Value
clc_eval_tic_expression(LocalEnvironment *env, Expression *operand)
{
    Value	result;
    double      t;

    tictime = clock();

    result.type = INT_VALUE;
    result.u.int_value = 1;
    return result;
}

Value
clc_eval_toc_expression(LocalEnvironment *env, Expression *operand)
{
    Value	result;
    clock_t toctime;
    double elapsed, t;
   
    toctime = clock();
    elapsed = (double) (toctime-tictime)/(double)CLOCKS_PER_SEC;
    result.type = DOUBLE_VALUE;
    result.u.double_value = elapsed;
    return result;
}

Value
clc_eval_down_expression(LocalEnvironment *env, Expression *operand)
{
    Value	result;
    down();
    result.type = INT_VALUE;
    result.u.int_value = 1;
    return result;
}

Value
clc_eval_up_expression(LocalEnvironment *env, Expression *operand)
{
    Value	result;

    up();
    result.type = INT_VALUE;
    result.u.int_value = 1;
    return result;
}

Value
clc_eval_near_expression(LocalEnvironment *env, Expression *operand)
{
    Value	result;

    near();
    result.type = INT_VALUE;
    result.u.int_value = 1;
    return result;
}

Value
clc_eval_upper_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int         i, k, l; 

    operand_val = eval_expression(env, operand);
    if(operand_val.type == DOUBLE_VALUE) 
    {
        result.type = DOUBLE_VALUE;
        result.u.double_value = operand_val.u.double_value;
    } else if (operand_val.type == INTERVAL_VALUE) {
        result.type = DOUBLE_VALUE;
        result.u.double_value = operand_val.u.interval_value.upper;
    } else if (operand_val.type == MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = operand_val.u.matrix_value.top[i];
        }
        result.u.matrix_value.column = k;
        result.u.matrix_value.row = l;
    } else if (operand_val.type == INTERVAL_MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.interval_matrix_value.upper.column;
        l=operand_val.u.interval_matrix_value.upper.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = operand_val.u.interval_matrix_value.upper.top[i];
        }
        result.u.complex_matrix_value.column = k;
        result.u.complex_matrix_value.row = l;
    }
    return result;
}

Value
clc_eval_lower_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int         i, k, l; 

    operand_val = eval_expression(env, operand);
    if(operand_val.type == DOUBLE_VALUE) 
    {
        result.type = DOUBLE_VALUE;
        result.u.double_value = operand_val.u.double_value;
    } else if (operand_val.type == INTERVAL_VALUE) {
        result.type = DOUBLE_VALUE;
        result.u.double_value = operand_val.u.interval_value.lower;
    } else if (operand_val.type == MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = operand_val.u.matrix_value.top[i];
        }
        result.u.matrix_value.column = k;
        result.u.matrix_value.row = l;
    } else if (operand_val.type == INTERVAL_MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.interval_matrix_value.lower.column;
        l=operand_val.u.interval_matrix_value.lower.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = operand_val.u.interval_matrix_value.lower.top[i];
        }
        result.u.complex_matrix_value.column = k;
        result.u.complex_matrix_value.row = l;
    }
    return result;
}

Value
clc_eval_real_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int         i, k, l; 

    operand_val = eval_expression(env, operand);
    if(operand_val.type == DOUBLE_VALUE) 
    {
        result.type = DOUBLE_VALUE;
        result.u.double_value = operand_val.u.double_value;
    } else if (operand_val.type == COMPLEX_VALUE) {
	result.type = DOUBLE_VALUE;
	result.u.double_value = operand_val.u.complex_value.r; 
    } else if (operand_val.type == MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = operand_val.u.matrix_value.top[i];
        }
        result.u.matrix_value.column = k;
        result.u.matrix_value.row = l;
    } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.complex_matrix_value.column;
        l=operand_val.u.complex_matrix_value.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = operand_val.u.complex_matrix_value.top[i].r;
        }
        result.u.complex_matrix_value.column = k;
        result.u.complex_matrix_value.row = l;
    }
    return result;
}

Value
clc_eval_imag_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int         i, k, l; 

    operand_val = eval_expression(env, operand);
    if(operand_val.type == DOUBLE_VALUE) 
    {
        result.type = DOUBLE_VALUE;
        result.u.double_value = 0.;
    } else if (operand_val.type == COMPLEX_VALUE) {
	result.type = DOUBLE_VALUE;
	result.u.double_value = operand_val.u.complex_value.i; 
    } else if (operand_val.type == MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = 0.;
        }
        result.u.matrix_value.column = k;
        result.u.matrix_value.row = l;
    } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.complex_matrix_value.column;
        l=operand_val.u.complex_matrix_value.row;
        result.u.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
        for(i=0; i<k*l; ++i)
        {
            result.u.matrix_value.top[i] = operand_val.u.complex_matrix_value.top[i].i;
        }
        result.u.complex_matrix_value.column = k;
        result.u.complex_matrix_value.row = l;
    }
    return result;
}


Value
clc_eval_inv_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    char        uplo, diag;
    int k, l;
    int i, j;
    int upper, lower;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = 1/operand_val.u.int_value;
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = 1/operand_val.u.double_value;
            return result;
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = operand_val.u.complex_value.r/ 
               (operand_val.u.complex_value.r*operand_val.u.complex_value.r+ 
                operand_val.u.complex_value.i*operand_val.u.complex_value.i);
	    result.u.complex_value.i = -operand_val.u.complex_value.i/ 
               (operand_val.u.complex_value.r*operand_val.u.complex_value.r+ 
                operand_val.u.complex_value.i*operand_val.u.complex_value.i);
            return result;
        } else if (operand_val.type == MATRIX_VALUE) {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
                upper = 1;
                for (i=1; i<l; ++i) {
                   for (j=0; j<i; ++j) {
                       if (operand_val.u.matrix_value.top[i+j*l] != 0.0) {
                           upper = 2;
                           break;
                       }
                    }
                    if (upper == 2) break;
                }
                lower = 1;
                for (i=0; i<l; ++i) {
                    for (j=i+1; j<k; ++j) {
                       if (operand_val.u.matrix_value.top[i+j*l] != 0.0) {
                           lower = 2;
                           break;
                       }
                    }
                    if (lower == 2) break;
                    }
                }
                if (upper == 1) {
                    uplo = 'U';
                    diag = 'N';
                    slb_triangular_matrix_inverse(
                        operand_val.u.matrix_value.top, 
                        result.u.matrix_value.top, uplo, diag, k);
                } else if (lower == 1) {
                    uplo = 'L';
                    diag = 'N';
                    slb_triangular_matrix_inverse(
                        operand_val.u.matrix_value.top, 
                        result.u.matrix_value.top, uplo, diag, k);
                } else if (upper != 1 && lower != 1) {
                    slb_matrix_inverse(operand_val.u.matrix_value.top, 
                                       result.u.matrix_value.top,k);
                }
                result.u.matrix_value.column = k;
                result.u.matrix_value.row = l;
                return result;
        } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
            result.type = COMPLEX_MATRIX_VALUE;
            k=operand_val.u.complex_matrix_value.column;
            l=operand_val.u.complex_matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                result.u.complex_matrix_value.top = (Complex *)calloc(k*l,sizeof(Complex));
                upper = 1;
                for (i=1; i<l; ++i) {
                   for (j=0; j<i; ++j) {
                       if ((operand_val.u.complex_matrix_value.top[i+j*l].r != 0.0) ||
                           (operand_val.u.complex_matrix_value.top[i+j*l].i != 0.0)) {
                           upper = 2;
                           break;
                       }
                    }
                    if (upper == 2) break;
                }
                lower = 1;
                for (i=0; i<l; ++i) {
                    for (j=i+1; j<k; ++j) {
                       if ((operand_val.u.complex_matrix_value.top[i+j*l].r != 0.0) ||
                           (operand_val.u.complex_matrix_value.top[i+j*l].i != 0.0)) {
                           lower = 2;
                           break;
                       }
                    }
                    if (lower == 2) break;
                    }
                }
                if (upper == 1) {
                    uplo = 'U';
                    diag = 'N';
                    slb_complex_triangular_matrix_inverse(
                        operand_val.u.complex_matrix_value.top, 
                        result.u.complex_matrix_value.top, uplo, diag, k);
                } else if (lower == 1) {
                    uplo = 'L';
                    diag = 'N';
                    slb_complex_triangular_matrix_inverse(
                        operand_val.u.complex_matrix_value.top, 
                        result.u.complex_matrix_value.top, uplo, diag, k);
                } else if (upper != 1 && lower != 1) {
                    slb_complex_matrix_inverse(operand_val.u.complex_matrix_value.top, 
                                   result.u.complex_matrix_value.top,k);
                }
                result.u.complex_matrix_value.column = k;
                result.u.complex_matrix_value.row = l;
                return result;
        }
    }
    return result; /* dummy */
}

Value
clc_eval_lu_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *vec;
    Complex     *vector;
    long int k, l;
    int i, j;
    operand_val = eval_expression(env, operand);
    if (operand_val.type == MATRIX_VALUE) {
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        if(k != l) {
            clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
        } else {
        result.type = TRIPLE_VALUE;
        result.u.triple_value.type_first  = MATRIX_VALUE;
        result.u.triple_value.type_second = MATRIX_VALUE;
        result.u.triple_value.type_third = MATRIX_VALUE;
        result.u.triple_value.name_first = "L";
        result.u.triple_value.name_second = "U";
        result.u.triple_value.name_third = "P";
        result.u.triple_value.u_first.matrix_value.column = k;
        result.u.triple_value.u_first.matrix_value.row = l;
        result.u.triple_value.u_second.matrix_value.column = k;
        result.u.triple_value.u_second.matrix_value.row = l;
        result.u.triple_value.u_third.matrix_value.column = k;
        result.u.triple_value.u_third.matrix_value.row = l;
        result.u.triple_value.u_first.matrix_value.top  = 
            (double *)GC_malloc_atomic(k*l*sizeof(double));
        for(i=0; i<k*l; i++) {
            result.u.triple_value.u_first.matrix_value.top[i] = 0.;
        }
        result.u.triple_value.u_second.matrix_value.top  = 
            (double *)GC_malloc_atomic(k*l*sizeof(double));
        for(i=0; i<k*l; i++) {
            result.u.triple_value.u_second.matrix_value.top[i] = 0.;
        }
        result.u.triple_value.u_third.matrix_value.top  = 
            (double *)GC_malloc_atomic(k*l*sizeof(double));
        for(i=0; i<k*l; i++) {
            result.u.triple_value.u_third.matrix_value.top[i] = 0.;
        }
        vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
        for (i=0; i<k*l; ++i) {
            vec[i] = operand_val.u.matrix_value.top[i];
        }
        slb_lu(vec, result.u.triple_value.u_third.matrix_value.top,k,l);
        for (i=0; i<l; ++i) {
            for(j=i; j<k; ++j) {
               result.u.triple_value.u_second.matrix_value.top[i+j*l]
                   =vec[i+j*l];
            }
        }                   
        for (i=0; i<l; ++i) {
            for(j=0; j<i; ++j) {
               result.u.triple_value.u_first.matrix_value.top[i+j*l]
                   =vec[i+j*l];
            }
            result.u.triple_value.u_first.matrix_value.top[i+i*l]
               = 1.0;
        }                   
        GC_free(vec);    
        return result;
        }
    } else if(operand_val.type == COMPLEX_MATRIX_VALUE) {
        k=operand_val.u.complex_matrix_value.column;
        l=operand_val.u.complex_matrix_value.row;
        if(k != l) {
            clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
        } else {
        result.type = TRIPLE_VALUE;
        result.u.triple_value.type_first  = COMPLEX_MATRIX_VALUE;
        result.u.triple_value.type_second = COMPLEX_MATRIX_VALUE;
        result.u.triple_value.type_third = MATRIX_VALUE;
        result.u.triple_value.name_first = "L";
        result.u.triple_value.name_second = "U";
        result.u.triple_value.name_third = "P";
        result.u.triple_value.u_first.complex_matrix_value.column = k;
        result.u.triple_value.u_first.complex_matrix_value.row = l;
        result.u.triple_value.u_second.complex_matrix_value.column = k;
        result.u.triple_value.u_second.complex_matrix_value.row = l;
        result.u.triple_value.u_third.matrix_value.column = k;
        result.u.triple_value.u_third.matrix_value.row = l;
        result.u.triple_value.u_first.complex_matrix_value.top  = 
            (Complex *)calloc(k*l,sizeof(Complex));
        result.u.triple_value.u_second.complex_matrix_value.top  = 
            (Complex *)calloc(k*l,sizeof(Complex));
        result.u.triple_value.u_third.matrix_value.top  = 
            (double *)calloc(k*l,sizeof(double));
        vector = (Complex *)GC_malloc(k*l*sizeof(Complex));
        for (i=0; i<k*l; ++i) {
            vector[i] = operand_val.u.complex_matrix_value.top[i];
        }
        slb_complex_lu(vector, result.u.triple_value.u_third.matrix_value.top,k,l);
        for (i=0; i<l; ++i) {
            for(j=i; j<k; ++j) {
               result.u.triple_value.u_second.complex_matrix_value.top[i+j*l]
                   =vector[i+j*l];
            }
        }                   
        for (i=0; i<l; ++i) {
            for(j=0; j<i; ++j) {
               result.u.triple_value.u_first.complex_matrix_value.top[i+j*l]
                   =vector[i+j*l];
            }
            result.u.triple_value.u_first.complex_matrix_value.top[i+i*l].r
               = 1.0;
        }                   
        GC_free(vector);    
        return result;
        }
    }
    return result; /* dummy */
}

Value
clc_eval_svd_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *vec;
    Complex     *vector;
    long int k, l, min, max;
    int i;
    operand_val = eval_expression(env, operand);
    if (operand_val.type == MATRIX_VALUE) 
    {
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        min = k;
        max = k;
        if (l < k)
        {
            min = l;
        }
        if (k < l)
        {
            max = l;
        }
                
        result.type = TRIPLE_VALUE;
        result.u.triple_value.type_first  = MATRIX_VALUE;
        result.u.triple_value.type_second = MATRIX_VALUE;
        result.u.triple_value.type_third = MATRIX_VALUE;
        result.u.triple_value.name_first = "U";
        result.u.triple_value.name_second = "S";
        result.u.triple_value.name_third = "V";
        result.u.triple_value.u_first.matrix_value.column = l;
        result.u.triple_value.u_first.matrix_value.row = l;
        result.u.triple_value.u_second.matrix_value.column = k;
        result.u.triple_value.u_second.matrix_value.row = l;
        result.u.triple_value.u_third.matrix_value.column = k;
        result.u.triple_value.u_third.matrix_value.row = k;
        result.u.triple_value.u_first.matrix_value.top  = 
            (double *)calloc(l*l,sizeof(double));
        result.u.triple_value.u_second.matrix_value.top  = 
            (double *)calloc(k*l,sizeof(double));
        result.u.triple_value.u_third.matrix_value.top  = 
            (double *)calloc(k*k,sizeof(double));
        vec = (double *)GC_malloc(k*l*sizeof(double));
        for (i=0; i<k*l; ++i) {
            vec[i] = operand_val.u.matrix_value.top[i];
        }
        slb_svd(vec, 
                    result.u.triple_value.u_first.matrix_value.top,
                    result.u.triple_value.u_second.matrix_value.top,
                    result.u.triple_value.u_third.matrix_value.top,
                    k,l);
        GC_free(vec);    
        return result;
    } 
    if(operand_val.type == COMPLEX_MATRIX_VALUE) 
    {
        k=operand_val.u.complex_matrix_value.column;
        l=operand_val.u.complex_matrix_value.row;
        min = k;
        max = k;
        if (l < k)
        {
            min = l;
        }
        if (k < l)
        {
            max = l;
        }
                
        result.type = TRIPLE_VALUE;
        result.u.triple_value.type_first  = COMPLEX_MATRIX_VALUE;
        result.u.triple_value.type_second = MATRIX_VALUE;
        result.u.triple_value.type_third = COMPLEX_MATRIX_VALUE;
        result.u.triple_value.name_first = "U";
        result.u.triple_value.name_second = "S";
        result.u.triple_value.name_third = "V";
        result.u.triple_value.u_first.complex_matrix_value.column = l;
        result.u.triple_value.u_first.complex_matrix_value.row = l;
        result.u.triple_value.u_second.matrix_value.column = k;
        result.u.triple_value.u_second.matrix_value.row = l;
        result.u.triple_value.u_third.complex_matrix_value.column = k;
        result.u.triple_value.u_third.complex_matrix_value.row = k;
        result.u.triple_value.u_first.complex_matrix_value.top  = 
            (Complex *)calloc(l*l,sizeof(Complex));
        result.u.triple_value.u_second.matrix_value.top  = 
            (double *)calloc(k*l,sizeof(double));
        result.u.triple_value.u_third.complex_matrix_value.top  = 
            (Complex *)calloc(k*k,sizeof(Complex));
        vector = (Complex *)GC_malloc(k*l*sizeof(Complex));
        for (i=0; i<k*l; ++i) {
            vector[i] = operand_val.u.complex_matrix_value.top[i];
        }
        slb_complex_svd(vector, 
                    result.u.triple_value.u_first.complex_matrix_value.top,
                    result.u.triple_value.u_second.matrix_value.top,
                    result.u.triple_value.u_third.complex_matrix_value.top,
                    k,l);
        GC_free(vector);    
        return result;
    }
    return result; /* dummy */
}

Value
clc_eval_chol_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *vec;
    Complex     *vector;
    long int k, l;
    int i, j;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == MATRIX_VALUE) 
    {
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        if(k != l) 
        {
            clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
        } 
        else
        {
            result.type = MATRIX_VALUE;
            result.u.matrix_value.column = l;
            result.u.matrix_value.row = l;
            result.u.matrix_value.top  = 
                (double *)calloc(l*l,sizeof(double));
            vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
            for (i=0; i<k*l; ++i) 
            {
                vec[i] = operand_val.u.matrix_value.top[i];
            }
            slb_chol(vec, l);
            for (i=0; i<l; ++i) {
                for(j=i; j<l; ++j) {
                   result.u.matrix_value.top[i+j*l]
                       =vec[i+j*l];
                }
            }                   
            GC_free(vec);
            return result;
        }
    }
    if (operand_val.type == COMPLEX_MATRIX_VALUE) 
    {
        k=operand_val.u.complex_matrix_value.column;
        l=operand_val.u.complex_matrix_value.row;
        if(k != l) 
        {
            clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
        } 
        else
        {
            result.type = COMPLEX_MATRIX_VALUE;
            result.u.complex_matrix_value.column = l;
            result.u.complex_matrix_value.row = l;
            result.u.complex_matrix_value.top  = 
                (Complex *)calloc(l*l,sizeof(Complex));
            vector = (Complex *)calloc(k*l,sizeof(Complex));
            for (i=0; i<k*l; ++i) 
            {
                vector[i] = operand_val.u.complex_matrix_value.top[i];
            }
            slb_complex_chol(vector, l);
            for (i=0; i<l; ++i) {
                for(j=i; j<l; ++j) {
                   result.u.complex_matrix_value.top[i+j*l]
                       =vector[i+j*l];
                }
            }                   
            GC_free(vector);
            return result;
        }
    }
    return result; /* dummy */
}



Value
clc_eval_schur_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *vec;
    Complex     *vector;
    long int k, l;
    int i, j;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == MATRIX_VALUE) 
    {
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        if(k != l) 
        {
            clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
        } 
        else 
        {
            result.type = COUPLE_VALUE;
            result.u.couple_value.type_first  = MATRIX_VALUE;
            result.u.couple_value.type_second = MATRIX_VALUE;
            result.u.couple_value.name_first = "t";
            result.u.couple_value.name_second = "z";
            result.u.couple_value.u_first.matrix_value.column = l;
            result.u.couple_value.u_first.matrix_value.row = l;
            result.u.couple_value.u_second.matrix_value.column = l;
            result.u.couple_value.u_second.matrix_value.row = l;
            result.u.couple_value.u_first.matrix_value.top  = 
            (double *)calloc(l*l,sizeof(double));
            result.u.couple_value.u_second.matrix_value.top  = 
            (double *)calloc(l*l,sizeof(double));
            vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
            for (i=0; i<k*l; ++i) 
            {
                vec[i] = operand_val.u.matrix_value.top[i];
            }
            slb_schur(vec, result.u.couple_value.u_second.matrix_value.top,l);
            for (i=0; i<l; ++i) 
            {
                for(j=i; j<k; ++j) 
                {
                     result.u.couple_value.u_first.matrix_value.top[i+j*l]
                       =vec[i+j*l];
                }
            }                   
            GC_free(vec);    
            return result;
        } 
    }
    if (operand_val.type == COMPLEX_MATRIX_VALUE) 
    {
        k=operand_val.u.complex_matrix_value.column;
        l=operand_val.u.complex_matrix_value.row;
        if(k != l) 
        {
            clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
        } 
        else 
        {
            result.type = COUPLE_VALUE;
            result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
            result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
            result.u.couple_value.name_first = "t";
            result.u.couple_value.name_second = "z";
            result.u.couple_value.u_first.complex_matrix_value.column = l;
            result.u.couple_value.u_first.complex_matrix_value.row = l;
            result.u.couple_value.u_second.complex_matrix_value.column = l;
            result.u.couple_value.u_second.complex_matrix_value.row = l;
            result.u.couple_value.u_first.complex_matrix_value.top  = 
            (Complex *)calloc(l*l,sizeof(Complex));
            result.u.couple_value.u_second.complex_matrix_value.top  = 
            (Complex *)calloc(l*l,sizeof(Complex));
            vector = (Complex *)GC_malloc(k*l*sizeof(Complex));
            for (i=0; i<k*l; ++i) 
            {
                vector[i] = operand_val.u.complex_matrix_value.top[i];
            }
            slb_complex_schur(vector, result.u.couple_value.u_second.complex_matrix_value.top,l);
            for (i=0; i<l; ++i) 
            {
                 for(j=i; j<k; ++j) 
                 {
                     result.u.couple_value.u_first.complex_matrix_value.top[i+j*l]
                       =vector[i+j*l];
                 }
            }                   
            GC_free(vector);    
            return result;
        } 
    }
    return result; /* dummy */
}

Value
clc_eval_qr_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *vec;
    Complex     *vector;
    long int k, l, min, max;
    int i, j;
    operand_val = eval_expression(env, operand);
    if (operand_val.type == MATRIX_VALUE) {
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        min = k;
        max = k;
        if (k < l) {
            max = l;
        } else {
            min = l;
        }
        result.type = COUPLE_VALUE;
        result.u.couple_value.type_first  = MATRIX_VALUE;
        result.u.couple_value.type_second = MATRIX_VALUE;
        result.u.couple_value.name_first = "q";
        result.u.couple_value.name_second = "r";
        result.u.couple_value.u_first.matrix_value.column = l;
        result.u.couple_value.u_first.matrix_value.row = l;
        result.u.couple_value.u_second.matrix_value.column = k;
        result.u.couple_value.u_second.matrix_value.row = l;
        result.u.couple_value.u_first.matrix_value.top  = 
            (double *)calloc(l*l,sizeof(double));
        result.u.couple_value.u_second.matrix_value.top  = 
            (double *)calloc(k*l,sizeof(double));
        vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
        for (i=0; i<k*l; ++i) {
            vec[i] = operand_val.u.matrix_value.top[i];
        }
        slb_qr(vec, result.u.couple_value.u_first.matrix_value.top,k,l);
        for (i=0; i<l; ++i) {
            for(j=i; j<k; ++j) {
               result.u.couple_value.u_second.matrix_value.top[i+j*l]
                   =vec[i+j*l];
            }
        }                   
        GC_free(vec);    
        return result;
    } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        min = k;
        max = k;
        if (k < l) {
            max = l;
        } else {
            min = l;
        }
        result.type = COUPLE_VALUE;
        result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
        result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
        result.u.couple_value.name_first = "q";
        result.u.couple_value.name_second = "r";
        result.u.couple_value.u_first.complex_matrix_value.column = l;
        result.u.couple_value.u_first.complex_matrix_value.row = l;
        result.u.couple_value.u_second.complex_matrix_value.column = k;
        result.u.couple_value.u_second.complex_matrix_value.row = l;
        result.u.couple_value.u_first.complex_matrix_value.top  = 
            (Complex *)calloc(l*l,sizeof(Complex));
        result.u.couple_value.u_second.complex_matrix_value.top  = 
            (Complex *)calloc(k*l,sizeof(Complex));
        vector = (Complex *)GC_malloc(k*l*sizeof(Complex));
        for (i=0; i<k*l; ++i) {
            vector[i] = operand_val.u.complex_matrix_value.top[i];
        }
        slb_complex_qr(vector, result.u.couple_value.u_first.complex_matrix_value.top,k,l);
        for (i=0; i<l; ++i) {
            for(j=i; j<k; ++j) {
               result.u.couple_value.u_second.complex_matrix_value.top[i+j*l]
                   =vector[i+j*l];
            }
        }                   
        GC_free(vector);    
        return result;
    }
    return result; /* dummy */
}

Value
clc_eval_veig_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *w, /* eigenvalues for real symmetric matrix */
                *w_real, *w_imag, /* eigenvalues for general matrix */
                *vec; /* eigenvectors for real symmetric matrix */ 
    int k, l;
    int i, j;
    int matrix_type; /* case 1: symmetric real matrix or Hermite matrix
                        case 2: general real matrix or general matrix */
    Complex     *w_complex; /* eigenvalues for general complex matrix */
    Complex     *vec_c, *oper_c;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = operand_val.u.int_value;
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = operand_val.u.double_value;
            return result;
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = operand_val.u.complex_value.r; 
	    result.u.complex_value.i = operand_val.u.complex_value.i; 
            return result;
        } else if (operand_val.type == MATRIX_VALUE) {
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                matrix_type = 1;
                for (i=0; i<l; ++i) {
                    for(j=i+1; j<k; ++j) {
                        if (operand_val.u.matrix_value.top[j+i*l] !=
                            operand_val.u.matrix_value.top[i+j*k]) {
                            matrix_type = 2;
                            break;
                        }
                    }
                    if (matrix_type == 2) break;
                } 
                if (matrix_type == 1) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = MATRIX_VALUE;
                    result.u.couple_value.type_second = MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w = (double *)GC_malloc_atomic(k*sizeof(double));
                    vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
                    for (i=0; i<k*l; ++i) {
                        vec[i] = operand_val.u.matrix_value.top[i];
                    }
                    result.u.couple_value.u_first.matrix_value.top  = 
                        (double *)GC_malloc_atomic(k*l*sizeof(double));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_first.matrix_value.top[i]=0.; 
                    }
                    slb_matrix_eigenvalue(vec, w,k);
                    for (i=0; i<k; ++i) {
                       result.u.couple_value.u_first.matrix_value.top[i*k+i] = w[i];
                    }
                    result.u.couple_value.u_second.matrix_value.top = vec;
                    result.u.couple_value.u_first.matrix_value.column = k;
                    result.u.couple_value.u_first.matrix_value.row = l;
                    result.u.couple_value.u_second.matrix_value.column = k;
                    result.u.couple_value.u_second.matrix_value.row = l;
                    GC_free(w);
                    return result;
                } else if (matrix_type == 2) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    vec_c = (Complex *)GC_malloc_atomic(k*l*sizeof(Complex));
                    oper_c = (Complex *)GC_malloc_atomic(k*l*sizeof(Complex));
                    for(i=0; i<k*l; ++i) {
                        oper_c[i].r=operand_val.u.matrix_value.top[i];
                        oper_c[i].i=0.;
                    }
                    w_complex = (Complex *)GC_malloc_atomic(k*sizeof(Complex));
                    result.u.couple_value.u_first.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    slb_gec_matrix_veigenvalue2(oper_c, w_complex, vec_c, k);
                    for (i=0; i<k; ++i) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i].r = w_complex[i].r;
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i].i = w_complex[i].i;
                    }
                    for (j=0; j<k; ++j) {
                        for (i=0; i<l; ++i) {
                            result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].r = vec_c[i+j*l].r;
                            result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].i = vec_c[i+j*l].i;
                        }
                    }
                    result.u.couple_value.u_first.complex_matrix_value.column = k;
                    result.u.couple_value.u_first.complex_matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w_complex);
                    GC_free(vec_c);
                    GC_free(oper_c);
                    return result;
                }
            }
        } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
            k=operand_val.u.complex_matrix_value.column;
            l=operand_val.u.complex_matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                matrix_type = 1;
                for (i=0; i<l; ++i) {
                    for(j=i+1; j<k; ++j) {
                        if (operand_val.u.complex_matrix_value.top[j+i*l].r !=
                            operand_val.u.complex_matrix_value.top[i+j*k].r) {
                            matrix_type = 2;
                            break;
                        }
                        if (operand_val.u.complex_matrix_value.top[j+i*l].i !=
                            -operand_val.u.complex_matrix_value.top[i+j*k].i) {
                            matrix_type = 2;
                            break;
                        }
                    }
                    if (matrix_type == 2) break;
                } 
                if (matrix_type == 1) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w = (double *)GC_malloc_atomic(k*sizeof(double));
                    result.u.couple_value.u_first.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for (i=0; i<k*l; ++i) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i] =
                            operand_val.u.complex_matrix_value.top[i];
                    }
                    slb_he_matrix_eigenvalue(result.u.couple_value.u_second.complex_matrix_value.top, w,k);
                    for (i=0; i<k; ++i) {
                       result.u.couple_value.u_first.matrix_value.top[i*k+i] = w[i];
                    }
                    result.u.couple_value.u_first.matrix_value.column = k;
                    result.u.couple_value.u_first.matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w);
                    return result;
                } else if (matrix_type == 2) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w_complex = (Complex *)GC_malloc(k*sizeof(Complex));
                    result.u.couple_value.u_first.complex_matrix_value.top  = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i].r  = 0.; 
                        result.u.couple_value.u_first.complex_matrix_value.top[i].i  = 0.; 
                    }
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i].r  = 0.; 
                        result.u.couple_value.u_second.complex_matrix_value.top[i].i  = 0.; 
                    }
                    slb_gec_matrix_veigenvalue2(operand_val.u.complex_matrix_value.top, w_complex, 
                        result.u.couple_value.u_second.complex_matrix_value.top, k);
                    for (i=0; i<k; ++i) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i] = w_complex[i];
                    }
                    result.u.couple_value.u_first.complex_matrix_value.column = k;
                    result.u.couple_value.u_first.complex_matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w_complex);
                    return result;
                }
            }
        }
    }
    return result; /* dummy */
}

Value
clc_eval_tri_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *w, /* eigenvalues for real symmetric matrix */
                *w_real, *w_imag, /* eigenvalues for general matrix */
                *vec, /* eigenvectors for real symmetric matrix */ 
                *d, *e,
                *tau;
    int k, l;
    int i, j;
    int matrix_type; /* case 1: symmetric real matrix or Hermite matrix
                        case 2: general real matrix or general matrix */
    Complex     *w_complex; /* eigenvalues for general complex matrix */

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = operand_val.u.int_value;
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = operand_val.u.double_value;
            return result;
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = operand_val.u.complex_value.r; 
	    result.u.complex_value.i = operand_val.u.complex_value.i; 
            return result;
        } else if (operand_val.type == MATRIX_VALUE) {
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                matrix_type = 1;
                for (i=0; i<l; ++i) {
                    for(j=i+1; j<k; ++j) {
                        if (operand_val.u.matrix_value.top[j+i*l] !=
                            operand_val.u.matrix_value.top[i+j*k]) {
                            matrix_type = 2;
                            break;
                        }
                    }
                    if (matrix_type == 2) break;
                } 
                if (matrix_type == 1) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = MATRIX_VALUE;
                    result.u.couple_value.type_second = MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w = (double *)GC_malloc_atomic((k-1)*sizeof(double));
                    vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
                    for (i=0; i<k*l; ++i) {
                        vec[i] = operand_val.u.matrix_value.top[i];
                    }
                    result.u.couple_value.u_first.matrix_value.top  = 
                        (double *)GC_malloc_atomic((k-1)*sizeof(double));
                    slb_matrix_tridiagonalize(vec, w,k);
                    for (i=0; i<k-1; ++i) {
                       result.u.couple_value.u_first.matrix_value.top[i] = w[i];
                    }
                    result.u.couple_value.u_second.matrix_value.top = vec;
                    result.u.couple_value.u_first.matrix_value.column = 1;
                    result.u.couple_value.u_first.matrix_value.row = k-1;
                    result.u.couple_value.u_second.matrix_value.column = k;
                    result.u.couple_value.u_second.matrix_value.row = l;
                    GC_free(w);
                    return result;
                } else if (matrix_type == 2) {
                    clc_runtime_error(NOT_SYMMETRIC_MATRIX_ERR, NULL);
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
                    for(i=0; i<k*l; i++) {
                        vec[i] = 0.;
                    } 
                    w_real = (double *)GC_malloc_atomic(k*sizeof(double));
                    w_imag = (double *)GC_malloc_atomic(k*sizeof(double));
                    result.u.couple_value.u_first.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i].r = 0.; 
                        result.u.couple_value.u_first.complex_matrix_value.top[i].i = 0.; 
                    } 
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i].r = 0.; 
                        result.u.couple_value.u_second.complex_matrix_value.top[i].i = 0.; 
                    } 
                    slb_ger_matrix_eigenvalue(operand_val.u.matrix_value.top, w_real, w_imag, vec, k);
                    for (i=0; i<k; ++i) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i].r = w_real[i];
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i].i = w_imag[i];
                    }
                    for (j=0; j<k; ++j) {
                        if(w_imag[j] == 0.0) { 
                            for (i=0; i<l; ++i) {
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].r = vec[i+j*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].i = 0.0;
                            }
                        } else if (j<l-1 && w_imag[j] == -w_imag[j+1] && w_real[j] == w_real[j+1]) {
                            for (i=0; i<l; ++i) {
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].r = vec[i+j*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].i = vec[i+(j+1)*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+(j+1)*l].r = vec[i+j*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+(j+1)*l].i = - vec[i+(j+1)*l];
                            }
                        } 
                    }
                    result.u.couple_value.u_first.complex_matrix_value.column = k;
                    result.u.couple_value.u_first.complex_matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w_real);
                    GC_free(w_imag);
                    GC_free(vec);
                    return result;
                }
            }
        } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
            k=operand_val.u.complex_matrix_value.column;
            l=operand_val.u.complex_matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                matrix_type = 1;
                for (i=0; i<l; ++i) {
                    for(j=i+1; j<k; ++j) {
                        if (operand_val.u.complex_matrix_value.top[j+i*l].r !=
                            operand_val.u.complex_matrix_value.top[i+j*k].r) {
                            matrix_type = 2;
                            break;
                        }
                        if (operand_val.u.complex_matrix_value.top[j+i*l].i !=
                            -operand_val.u.complex_matrix_value.top[i+j*k].i) {
                            matrix_type = 2;
                            break;
                        }
                    }
                    if (matrix_type == 2) break;
                } 
                if (matrix_type == 1) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w = (double *)GC_malloc_atomic(k*sizeof(double));
                    result.u.couple_value.u_first.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for (i=0; i<k*l; ++i) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i] =
                            operand_val.u.complex_matrix_value.top[i];
                    }
                    slb_he_matrix_eigenvalue(result.u.couple_value.u_second.complex_matrix_value.top, w,k);
                    for (i=0; i<k; ++i) {
                       result.u.couple_value.u_first.matrix_value.top[i*k+i] = w[i];
                    }
                    result.u.couple_value.u_first.matrix_value.column = k;
                    result.u.couple_value.u_first.matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w);
                    return result;
                } else if (matrix_type == 2) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w_complex = (Complex *)GC_malloc(k*sizeof(Complex));
                    result.u.couple_value.u_first.complex_matrix_value.top  = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i].r  = 0.; 
                        result.u.couple_value.u_first.complex_matrix_value.top[i].i  = 0.; 
                    }
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i].r  = 0.; 
                        result.u.couple_value.u_second.complex_matrix_value.top[i].i  = 0.; 
                    }
                    slb_gec_matrix_eigenvalue(operand_val.u.complex_matrix_value.top, w_complex, 
                        result.u.couple_value.u_second.complex_matrix_value.top, k);
                    for (i=0; i<k; ++i) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i] = w_complex[i];
                    }
                    result.u.couple_value.u_first.complex_matrix_value.column = k;
                    result.u.couple_value.u_first.complex_matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w_complex);
                    return result;
                }
            }
        }
    }
    return result; /* dummy */
}

Value
clc_eval_eig_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      *w, /* eigenvalues for real symmetric matrix */
                *w_real, *w_imag, /* eigenvalues for general matrix */
                *vec; /* eigenvectors for real symmetric matrix */ 
    int k, l;
    int i, j;
    int matrix_type; /* case 1: symmetric real matrix or Hermite matrix
                        case 2: general real matrix or general matrix */
    Complex     *w_complex; /* eigenvalues for general complex matrix */

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = operand_val.u.int_value;
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = operand_val.u.double_value;
            return result;
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = operand_val.u.complex_value.r; 
	    result.u.complex_value.i = operand_val.u.complex_value.i; 
            return result;
        } else if (operand_val.type == MATRIX_VALUE) {
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                matrix_type = 1;
                for (i=0; i<l; ++i) {
                    for(j=i+1; j<k; ++j) {
                        if (operand_val.u.matrix_value.top[j+i*l] !=
                            operand_val.u.matrix_value.top[i+j*k]) {
                            matrix_type = 2;
                            break;
                        }
                    }
                    if (matrix_type == 2) break;
                } 
                if (matrix_type == 1) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = MATRIX_VALUE;
                    result.u.couple_value.type_second = MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w = (double *)GC_malloc_atomic(k*sizeof(double));
                    vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
                    for (i=0; i<k*l; ++i) {
                        vec[i] = operand_val.u.matrix_value.top[i];
                    }
                    result.u.couple_value.u_first.matrix_value.top  = 
                        (double *)GC_malloc_atomic(k*l*sizeof(double));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_first.matrix_value.top[i]=0.; 
                    }
                    slb_matrix_eigenvalue(vec, w,k);
                    for (i=0; i<k; ++i) {
                       result.u.couple_value.u_first.matrix_value.top[i*k+i] = w[i];
                    }
                    result.u.couple_value.u_second.matrix_value.top = vec;
                    result.u.couple_value.u_first.matrix_value.column = k;
                    result.u.couple_value.u_first.matrix_value.row = l;
                    result.u.couple_value.u_second.matrix_value.column = k;
                    result.u.couple_value.u_second.matrix_value.row = l;
                    GC_free(w);
                    return result;
                } else if (matrix_type == 2) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    vec = (double *)GC_malloc_atomic(k*l*sizeof(double));
                    for(i=0; i<k*l; i++) {
                        vec[i] = 0.;
                    } 
                    w_real = (double *)GC_malloc_atomic(k*sizeof(double));
                    w_imag = (double *)GC_malloc_atomic(k*sizeof(double));
                    result.u.couple_value.u_first.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i].r = 0.; 
                        result.u.couple_value.u_first.complex_matrix_value.top[i].i = 0.; 
                    } 
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i].r = 0.; 
                        result.u.couple_value.u_second.complex_matrix_value.top[i].i = 0.; 
                    } 
                    slb_ger_matrix_eigenvalue(operand_val.u.matrix_value.top, w_real, w_imag, vec, k);
                    for (i=0; i<k; ++i) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i].r = w_real[i];
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i].i = w_imag[i];
                    }
                    for (j=0; j<k; ++j) {
                        if(w_imag[j] == 0.0) { 
                            for (i=0; i<l; ++i) {
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].r = vec[i+j*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].i = 0.0;
                            }
                        } else if (j<l-1 && w_imag[j] == -w_imag[j+1] && w_real[j] == w_real[j+1]) {
                            for (i=0; i<l; ++i) {
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].r = vec[i+j*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+j*l].i = vec[i+(j+1)*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+(j+1)*l].r = vec[i+j*l];
                                result.u.couple_value.u_second.complex_matrix_value.top[i+(j+1)*l].i = - vec[i+(j+1)*l];
                            }
                        } 
                    }
                    result.u.couple_value.u_first.complex_matrix_value.column = k;
                    result.u.couple_value.u_first.complex_matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w_real);
                    GC_free(w_imag);
                    GC_free(vec);
                    return result;
                }
            }
        } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
            k=operand_val.u.complex_matrix_value.column;
            l=operand_val.u.complex_matrix_value.row;
            if(k != l) {
                clc_runtime_error(NOT_SQUARE_MATRIX_ERR, NULL);
            } else {
                matrix_type = 1;
                for (i=0; i<l; ++i) {
                    for(j=i+1; j<k; ++j) {
                        if (operand_val.u.complex_matrix_value.top[j+i*l].r !=
                            operand_val.u.complex_matrix_value.top[i+j*k].r) {
                            matrix_type = 2;
                            break;
                        }
                        if (operand_val.u.complex_matrix_value.top[j+i*l].i !=
                            -operand_val.u.complex_matrix_value.top[i+j*k].i) {
                            matrix_type = 2;
                            break;
                        }
                    }
                    if (matrix_type == 2) break;
                } 
                if (matrix_type == 1) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w = (double *)GC_malloc_atomic(k*sizeof(double));
                    result.u.couple_value.u_first.matrix_value.top = (double *)GC_malloc(k*l*sizeof(double));
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for (i=0; i<k*l; ++i) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i] =
                            operand_val.u.complex_matrix_value.top[i];
                    }
                    slb_he_matrix_eigenvalue(result.u.couple_value.u_second.complex_matrix_value.top, w,k);
                    for (i=0; i<k; ++i) {
                       result.u.couple_value.u_first.matrix_value.top[i*k+i] = w[i];
                    }
                    result.u.couple_value.u_first.matrix_value.column = k;
                    result.u.couple_value.u_first.matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w);
                    return result;
                } else if (matrix_type == 2) {
                    result.type = COUPLE_VALUE;
                    result.u.couple_value.type_first  = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.type_second = COMPLEX_MATRIX_VALUE;
                    result.u.couple_value.name_first = "val";
                    result.u.couple_value.name_second = "vec";
                    w_complex = (Complex *)GC_malloc(k*sizeof(Complex));
                    result.u.couple_value.u_first.complex_matrix_value.top  = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i].r  = 0.; 
                        result.u.couple_value.u_first.complex_matrix_value.top[i].i  = 0.; 
                    }
                    result.u.couple_value.u_second.complex_matrix_value.top = 
                        (Complex *)GC_malloc(k*l*sizeof(Complex));
                    for(i=0; i<k*l; i++) {
                        result.u.couple_value.u_second.complex_matrix_value.top[i].r  = 0.; 
                        result.u.couple_value.u_second.complex_matrix_value.top[i].i  = 0.; 
                    }
                    slb_gec_matrix_eigenvalue(operand_val.u.complex_matrix_value.top, w_complex, 
                        result.u.couple_value.u_second.complex_matrix_value.top, k);
                    for (i=0; i<k; ++i) {
                        result.u.couple_value.u_first.complex_matrix_value.top[i*k+i] = w_complex[i];
                    }
                    result.u.couple_value.u_first.complex_matrix_value.column = k;
                    result.u.couple_value.u_first.complex_matrix_value.row = l;
                    result.u.couple_value.u_second.complex_matrix_value.column = k;
                    result.u.couple_value.u_second.complex_matrix_value.row = l;
                    GC_free(w_complex);
                    return result;
                }
            }
        }
    }
    return result; /* dummy */
}

Value
clc_eval_transpose_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int k, l;
    int i, j;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = operand_val.u.int_value;
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = operand_val.u.double_value;
            return result;
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = operand_val.u.complex_value.r; 
	    result.u.complex_value.i = -operand_val.u.complex_value.i; 
            return result;
        } else if (operand_val.type == MATRIX_VALUE) {
	    result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = 
                (double *)calloc(k*l,sizeof(double));
            for (i=0; i<k; ++i) {
               for(j=0; j<l; ++j) {
                   result.u.matrix_value.top[i+j*k] = 
                     operand_val.u.matrix_value.top[j+i*l];
               }
            }
            result.u.matrix_value.column = l;
            result.u.matrix_value.row = k;
            return result;
        } else if (operand_val.type == COMPLEX_MATRIX_VALUE) {
	    result.type = COMPLEX_MATRIX_VALUE; 
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.complex_matrix_value.top = 
                (Complex *)GC_malloc(k*l*sizeof(Complex));
            for (i=0; i<k; ++i) {
               for(j=0; j<l; ++j) {
                   result.u.complex_matrix_value.top[i+j*k].r = 
                     operand_val.u.complex_matrix_value.top[j+i*l].r;
                   result.u.complex_matrix_value.top[i+j*k].i = 
                     -operand_val.u.complex_matrix_value.top[j+i*l].i;
               }
            }
            result.u.complex_matrix_value.column = l;
            result.u.complex_matrix_value.row = k;
            return result;
        }
    }
    return result; /* dummy */
}

Value
clc_eval_vsin_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    MPA         a, b;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = sin(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
            a = m_set(operand_val.u.double_value);
            b = m_sin(a);     
	    result.u.double_value = m_iset(b);
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_vsin(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_vsin(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    a = m_set(operand_val.u.matrix_value.top[i*k+j]);
                    b = m_sin(a);     
                    result.u.matrix_value.top[i*k+j] = m_iset(b);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_sin_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double s;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = sin(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = sin(operand_val.u.double_value);
        } else if(operand_val.type == DIF_VALUE) {
	    result.type = DIF_VALUE;
	    result.u.dif_value.n = operand_val.u.dif_value.n;
            result.u.dif_value.value = sin(operand_val.u.dif_value.value);
            result.u.dif_value.jacobi 
                = (double *)calloc(operand_val.u.dif_value.n, sizeof(double));
            s = cos(operand_val.u.dif_value.value);
            for(i = 0; i<operand_val.u.dif_value.n; i++) 
            {
                result.u.dif_value.jacobi[i]=s * operand_val.u.dif_value.jacobi[i];
            }
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_sin(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_sin(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 sin(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

double real_part_of_complex_vsin(double r,double i) {
    double result;
    MPA a,b,c,d,e;
    a = m_set(r);
    b = m_set(i);
    c = m_sin(a);
    d = m_hcos(b);
    e = m_mul(c,d); 
    result = m_iset(e);
    return result;
}

double imaginary_part_of_complex_vsin(double r, double i) {
    double result;
    MPA a,b,c,d,e;
    a = m_set(r);
    b = m_set(i);
    c = m_cos(a);
    d = m_hsin(b);
    e = m_mul(c,d); 
    result = m_iset(e);
    return result;
}

double real_part_of_complex_sin(double r,double i) {
    double result;
    result = sin(r) * cosh(i);
    return result;
}

double imaginary_part_of_complex_sin(double r, double i) {
   double result;
   result = cos(r) * sinh(i);
   return result;
}

Value
clc_eval_vcos_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    MPA         a, b;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = cos(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
            a = m_set(operand_val.u.double_value);
            b = m_cos(a);     
	    result.u.double_value = m_iset(b);
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_cos(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_cos(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    a = m_set(operand_val.u.matrix_value.top[i*k+j]);
                    b = m_cos(a);     
                    result.u.matrix_value.top[i*k+j] = m_iset(b);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_cos_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double s;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = cos(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = cos(operand_val.u.double_value);
        } else if(operand_val.type == DIF_VALUE) {
	    result.type = DIF_VALUE;
	    result.u.dif_value.n = operand_val.u.dif_value.n;
            result.u.dif_value.value = cos(operand_val.u.dif_value.value);
            result.u.dif_value.jacobi 
                = (double *)calloc(operand_val.u.dif_value.n, sizeof(double));
            s = -sin(operand_val.u.dif_value.value);
            for(i = 0; i<operand_val.u.dif_value.n; i++) 
            {
                result.u.dif_value.jacobi[i]=s * operand_val.u.dif_value.jacobi[i];
            }
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_cos(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_cos(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 cos(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

double real_part_of_complex_cos(double r,double i) {
    double result;
    result = cos(r) * cosh(i);
    return result;
}

double imaginary_part_of_complex_cos(double r, double i) {
   double result;
   result = -sin(r) * sinh(i);
   return result;
}

double real_part_of_complex_vcos(double r,double i) {
    double result;
    MPA a,b,c,d,e;
    a = m_set(r);
    b = m_set(i);
    c = m_cos(a);
    d = m_hcos(b);
    e = m_mul(c,d); 
    result = m_iset(e);
    return result;
}

double imaginary_part_of_complex_vcos(double r, double i) {
    double result;
    MPA a,b,c,cc,d,e;
    a = m_set(r);
    b = m_set(i);
    cc = m_sin(a);
    c  = m_mul_s(cc,-1);    
    d = m_hsin(b);
    e = m_mul(c,d); 
    result = m_iset(e);
    return result;
}

Value
clc_eval_vlog_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    MPA         a, b;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = log(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
            a = m_set(operand_val.u.double_value);
            b = m_log(a);     
	    result.u.double_value = m_iset(b);
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_vlog(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_vlog(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    a = m_set(operand_val.u.matrix_value.top[i*k+j]);
                    b = m_log(a);     
                    result.u.matrix_value.top[i*k+j] = m_iset(b);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_log_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double s;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = log(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = log(operand_val.u.double_value);
        } else if(operand_val.type == DIF_VALUE) {
	    result.type = DIF_VALUE;
	    result.u.dif_value.n = operand_val.u.dif_value.n;
            result.u.dif_value.value = log(operand_val.u.dif_value.value);
            result.u.dif_value.jacobi 
                = (double *)calloc(operand_val.u.dif_value.n, sizeof(double));
            s = 1/operand_val.u.dif_value.value;
            for(i = 0; i<operand_val.u.dif_value.n; i++) 
            {
                result.u.dif_value.jacobi[i]=s * operand_val.u.dif_value.jacobi[i];
            }
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_log(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_log(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 log(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

double real_part_of_complex_log(double r,double i) {
    double result;
    result = 0.5*log(r*r + i*i);
    return result;
}

double imaginary_part_of_complex_log(double r, double i) {
   double result;
   result = atan2 (i, r);
   return result;
}

double real_part_of_complex_vlog(double r,double i) {
    double result;
    MPA a,b,c,d,e,f,g;
    a = m_set(r);
    b = m_set(i);
    c = m_mul(a,a);
    d = m_mul(b,b);
    e = m_add(c,d); 
    f = m_log(e);
    g = m_div_s(f,2);
    result = m_iset(g);
    return result;
}

double imaginary_part_of_complex_vlog(double r, double i) {
    double result;
/*    MPA a,b,c,d;
    a = m_set(r);
    b = m_set(i);
    c = m_div(a,b);    
    d = m_atan(c);
    result = m_iset(d); */

    result = atan2 (i, r);

    return result;
}

Value
clc_eval_vexp_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    MPA         a, b;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = exp(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
            a = m_set(operand_val.u.double_value);
            b = m_exp(a);     
	    result.u.double_value = m_iset(b);
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_vexp(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_vexp(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    a = m_set(operand_val.u.matrix_value.top[i*k+j]);
                    b = m_exp(a);     
                    result.u.matrix_value.top[i*k+j] = m_iset(b);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_exp_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double s;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = exp(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = exp(operand_val.u.double_value);
        } else if(operand_val.type == DIF_VALUE) {
	    result.type = DIF_VALUE;
	    result.u.dif_value.n = operand_val.u.dif_value.n;
            result.u.dif_value.value = exp(operand_val.u.dif_value.value);
            result.u.dif_value.jacobi 
                = (double *)calloc(operand_val.u.dif_value.n, sizeof(double));
            s = exp(operand_val.u.dif_value.value);
            for(i = 0; i<operand_val.u.dif_value.n; i++) 
            {
                result.u.dif_value.jacobi[i]=s * operand_val.u.dif_value.jacobi[i];
            }
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_exp(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_exp(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 exp(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

double real_part_of_complex_exp(double r,double i) {
    double result;
    result = exp(r) * cos(i);
    return result;
}

double imaginary_part_of_complex_exp(double r, double i) {
   double result;
   result = exp(r) * sin(i);
   return result;
}

double real_part_of_complex_vexp(double r,double i) {
    double result;
    MPA a,b,c,d,e;
    a = m_set(r);
    b = m_set(i);
    c = m_exp(a);
    d = m_cos(b);
    e = m_mul(c,d); 
    result = m_iset(e);
    return result;
}

double imaginary_part_of_complex_vexp(double r, double i) {
    double result;
    MPA a,b,c,d,e;
    a = m_set(r);
    b = m_set(i);
    c = m_exp(a);
    d = m_sin(b);
    e = m_mul(c,d); 
    result = m_iset(e);
    return result;
}

Value
clc_eval_norm_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    double      t, tt;
    int         i, j;
    long int row, column;

    operand_val = eval_expression(env, operand);
    if(operand_val.type == DOUBLE_VALUE) {
        result.type = DOUBLE_VALUE;
        result.u.double_value = fabs(operand_val.u.double_value);
    } else if (operand_val.type == COMPLEX_VALUE) {
        result.type = DOUBLE_VALUE;
        result.u.double_value = 
           sqrt(operand_val.u.complex_value.r*operand_val.u.complex_value.r 
          +operand_val.u.complex_value.i*operand_val.u.complex_value.i);
    } else if(operand_val.type == MATRIX_VALUE) {
        result.type = DOUBLE_VALUE;
        column=operand_val.u.matrix_value.column;
        row=operand_val.u.matrix_value.row;
        result.u.double_value = 
        slb_matrix_norm(row,column,operand_val.u.matrix_value.top);
    } else if(operand_val.type == COMPLEX_MATRIX_VALUE) {
        result.type = DOUBLE_VALUE;
        column=operand_val.u.complex_matrix_value.column;
        row=operand_val.u.complex_matrix_value.row;
        t=0.;
        for(i=0; i<row; ++i) { 
            tt=0.;
            for (j=0; j<column; ++j) {
                tt += sqrt(operand_val.u.complex_matrix_value.top[i].r*operand_val.u.complex_matrix_value.top[i].r+operand_val.u.complex_matrix_value.top[i].i*operand_val.u.complex_matrix_value.top[i].i);
            }
            if(tt>t) t=tt; 
        } 
        result.u.double_value = t;
    } 
    return result;
}

Value
clc_eval_minus_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = -(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = -(operand_val.u.double_value);
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                           -operand_val.u.complex_value.r, 
	    result.u.complex_value.i = 
                           -operand_val.u.complex_value.i;
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 -(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_log10_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = log10(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = log10(operand_val.u.double_value);
        }
        else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 log10(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_fabs_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;
    double temp_1, temp_2;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = fabs(operand_val.u.int_value);
        return result;
    } else if(operand_val.type == DOUBLE_VALUE) {
        result.type = DOUBLE_VALUE;
        result.u.double_value = fabs(operand_val.u.double_value);
        return result;
    } else if(operand_val.type == MATRIX_VALUE) {
        result.type = MATRIX_VALUE;
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
        for(i=0; i < l; i++) {
            for(j=0; j < k; j++) {
                result.u.matrix_value.top[i*k+j] = 
                             fabs(operand_val.u.matrix_value.top[i*k+j]);
            }
        }        
        result.u.matrix_value.column = k;
        result.u.matrix_value.row = l;
        return result;
    } else if(operand_val.type == INTERVAL_MATRIX_VALUE) {  
        result.type = MATRIX_VALUE;
        k=operand_val.u.matrix_value.column;
        l=operand_val.u.matrix_value.row;
        result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
        for(i=0; i < l; i++) {
            for(j=0; j < k; j++) {
                temp_1 = fabs(operand_val.u.interval_matrix_value.lower.top[i*k+j]);
                temp_2 = fabs(operand_val.u.interval_matrix_value.upper.top[i*k+j]);
                if(temp_1 >= temp_2) {
                    result.u.matrix_value.top[i*k+j] = temp_1;
                } else {
                    result.u.matrix_value.top[i*k+j] = temp_2;
                }
            }
        }        
        result.u.matrix_value.column = k;
        result.u.matrix_value.row = l;
        return result;
     }
     return result; /* dummy */
}

Value
clc_eval_vtan_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    MPA         a, b, c, d;
    int i, j, k, l;
    double *cc;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = tan(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
            a = m_set(operand_val.u.double_value);
            b = m_cos(a);     
            c = m_sin(a);
            d = m_div(c,b);
	    result.u.double_value = m_iset(d);
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
            cc = complex_vtan(
                    operand_val.u.complex_value.r, 
                    operand_val.u.complex_value.i);
	    result.u.complex_value.r = cc[0];
	    result.u.complex_value.i = cc[1];
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 tan(operand_val.u.matrix_value.top[i*k+j]);
                    a = m_set(operand_val.u.matrix_value.top[i*k+j]);
                    b = m_cos(a);     
                    c = m_sin(a);
                    d = m_div(c,b);
                    result.u.matrix_value.top[i*k+j] = m_iset(d);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_tan_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = tan(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = tan(operand_val.u.double_value);
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_tan(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_tan(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 tan(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

double real_part_of_complex_tan(double r,double i) {
    double result;
    result = sin(2*r) / (cos(2*r)+cosh(2*i));
    return result;
}

double imaginary_part_of_complex_tan(double r, double i) {
   double result;
   result = sinh(2*i) / (cos(2*r)+cosh(2*i));
   return result;
}

double *complex_vtan(double r,double i) {
    double *result;
    MPA a,b,c,d,e,g,h,k,l;

    result = (double*)GC_malloc_atomic(2*sizeof(double));
    a = m_set(r);
    g = m_mul_s(a,2);
    b = m_set(i);
    h = m_mul_s(b,2);
    c = m_sin(g);
    k = m_cos(g);
    d = m_hcos(h);
    e = m_add(k,d);
    a = m_div(c,e);
    result[0] = m_iset(a);
    l = m_hsin(h);
    a = m_div(l,e);
    result[1] = m_iset(a);
    return result;
}

Value
clc_eval_sqrt_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = sqrt(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
            if(operand_val.u.double_value < 0) {
                result.type = COMPLEX_VALUE;
                result.u.complex_value.r = 0;
                result.u.complex_value.i = sqrt(-operand_val.u.double_value);
            }
            else {
	        result.type = DOUBLE_VALUE;
	        result.u.double_value = sqrt(operand_val.u.double_value);
            }
        } else if (operand_val.type == COMPLEX_VALUE) {
	    result.type = COMPLEX_VALUE;
	    result.u.complex_value.r = 
                      real_part_of_complex_sqrt(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
	    result.u.complex_value.i = 
                      imaginary_part_of_complex_sqrt(
                           operand_val.u.complex_value.r, 
                           operand_val.u.complex_value.i);
        } else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 sqrt(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

double real_part_of_complex_sqrt(double r,double i) {
    double result;
    double abs_z, t;

    abs_z = sqrt(r*r + i*i);
    if(abs_z == 0.) result = 0.;
    else if (r > 0.) {
        result = sqrt(0.5 * (abs_z + r));
    } else {
        t = sqrt(0.5 * (abs_z - r));
        if(i < 0.) t = -t;
        t = i/t;
        result = 0.5 * t;
    }
    return result;
}

double imaginary_part_of_complex_sqrt(double r, double i) {
    double result;
    double abs_z, t;

    abs_z = sqrt(r*r + i*i);
    if(abs_z == 0.) result = 0.;
    else if (r > 0.) {
        t = sqrt(0.5 * (abs_z + r));
        t = i/t;
        result = 0.5 * t;
    } else {
        t = sqrt(0.5 * (abs_z - r));
        if(i < 0.) t = -t;
        result = t;
    }
    return result;
}

Value
clc_eval_atan_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = atan(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = atan(operand_val.u.double_value);
        }
        else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 atan(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_acos_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = acos(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = acos(operand_val.u.double_value);
        }
        else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 acos(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_asin_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = asin(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = asin(operand_val.u.double_value);
        }
        else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 asin(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_cosh_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = cosh(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = cosh(operand_val.u.double_value);
        }
        else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 cosh(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_sinh_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = sinh(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = sinh(operand_val.u.double_value);
        }
        else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 sinh(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_tanh_expression(LocalEnvironment *env, Expression *operand)
{
    Value	operand_val;
    Value	result;
    int i, j, k, l;

    operand_val = eval_expression(env, operand);
    if (operand_val.type == INT_VALUE) {
	result.type = INT_VALUE;
	result.u.int_value = tanh(operand_val.u.int_value);
    } 
    else 
    {
        if(operand_val.type == DOUBLE_VALUE) {
	    result.type = DOUBLE_VALUE;
	    result.u.double_value = tanh(operand_val.u.double_value);
        }
        else {
            result.type = MATRIX_VALUE;
            k=operand_val.u.matrix_value.column;
            l=operand_val.u.matrix_value.row;
            result.u.matrix_value.top = (double *)calloc(k*l,sizeof(double));
            for(i=0; i < l; i++) {
                for(j=0; j < k; j++) {
                    result.u.matrix_value.top[i*k+j] = 
                                 tanh(operand_val.u.matrix_value.top[i*k+j]);
                }
            }        
            result.u.matrix_value.column = k;
            result.u.matrix_value.row = l;
        }  
    }
    return result;
}

Value
clc_eval_matrix_expression(LocalEnvironment *env, Expression *mat, 
                                         long int column, long int row)
{
    Value	result;
    Value       temp;
    int i, j; 
    long int k, l;
    double *A;
    k=column;
    l=row;
    A=(double *)calloc(k*l,sizeof(double));
    for(i=0; i < l; ++i) {
        for(j=0; j < k; ++j) {
            temp = eval_expression(env, &mat[i*k+j]); 
            A[i*k+j] = temp.u.double_value;
        }
    }

    result.u.matrix_value.top = A;
    result.u.matrix_value.column = k;
    result.u.matrix_value.row = l;
    result.type = MATRIX_VALUE; 

    return result;
}

Value
clc_eval_matrix_value_expression(LocalEnvironment *env, double *mat, 
                                         long int column, long int row)
{
    Value	result;
    long int k, l;

    k=column;
    l=row;
    result.u.matrix_value.top = mat;
    result.u.matrix_value.column = column;
    result.u.matrix_value.row = row;
    result.type = MATRIX_VALUE; 

    return result;
}

Value
clc_eval_complex_matrix_value_expression(LocalEnvironment *env, Complex *mat, 
                                         long int column, long int row)
{
    Value	result;
    long int k, l;
    k=column;
    l=row;
    result.u.complex_matrix_value.top = mat;
    result.u.complex_matrix_value.column = column;
    result.u.complex_matrix_value.row = row;
    result.type = COMPLEX_MATRIX_VALUE; 

    return result;
}

Value
clc_eval_matrix_element_expression(LocalEnvironment *env, 
                                   Expression *matrix_element_expression)
{
    Value	result;
    result = eval_expression(env, matrix_element_expression);

    return result;
}

static Value
eval_if_expression(LocalEnvironment *env,
		   Expression *condition, Expression *then_expression,
		   Expression *else_expression)
{
    Value	condition_val;
    Value	result;

    condition_val = eval_expression(env, condition);
    if (condition_val.type != INT_VALUE) {
	clc_runtime_error(BOOLEAN_EXPECTED_ERR, NULL);
    }
    if (condition_val.u.int_value) {
	result = eval_expression(env, then_expression);
    } else {
        if(else_expression == NULL) 
        {
            result.type = DOUBLE_VALUE;
            result.u.double_value = 1.0;
        }
        else
        {
	    result = eval_expression(env, else_expression);
        }
    }
    return result;
}

static Value
eval_if_expression2(LocalEnvironment *env,
		   Expression *condition, Expression *then_expression)
{
    Value	condition_val;
    Value	result;

    condition_val = eval_expression(env, condition);
    if (condition_val.type != INT_VALUE) {
	clc_runtime_error(BOOLEAN_EXPECTED_ERR, NULL);
    }
    if (condition_val.u.int_value) {
	result = eval_expression(env, then_expression);
    } else {
        result.type = DOUBLE_VALUE;
        result.u.double_value = 1.0;
    }
    return result;
}

static Value
eval_for_expression(LocalEnvironment *env,
		      Expression *first_condition, 
		      Expression *second_condition, 
		      Expression *third_condition, 
                      Expression *expression_list)
{
    Value	first_condition_val;
    Value	second_condition_val;
    Value	third_condition_val;
    Value	result;

    first_condition_val = eval_expression(env, first_condition);
    for (;;) {
	second_condition_val = eval_expression(env, second_condition);
	if (second_condition_val.type != INT_VALUE) {
	    clc_runtime_error(BOOLEAN_EXPECTED_ERR, NULL);
	}
	if (!second_condition_val.u.int_value)
	    break;

	result = eval_expression(env, expression_list);
        third_condition_val = eval_expression(env, third_condition);
    }
    return result;
}

static Value
eval_while_expression(LocalEnvironment *env,
		      Expression *condition, Expression *expression_list)
{
    Value	condition_val;
    Value	result;

    for (;;) {
	condition_val = eval_expression(env, condition);
	if (condition_val.type != INT_VALUE) {
	    clc_runtime_error(BOOLEAN_EXPECTED_ERR, NULL);
	}
	if (!condition_val.u.int_value)
	    break;

	result = eval_expression(env, expression_list);
    }
    return result;
}

static Value
eval_do_while_expression(LocalEnvironment *env,
                      Expression *expression_list, Expression *condition) 
{
    Value	condition_val;
    Value	result;

    for (;;) {
	result = eval_expression(env, expression_list);

	condition_val = eval_expression(env, condition);
	if (condition_val.type != INT_VALUE) {
	    clc_runtime_error(BOOLEAN_EXPECTED_ERR, NULL);
	}
	if (!condition_val.u.int_value)
	    break;
    }
    return result;
}

static LocalEnvironment *
alloc_local_environment()
{
    LocalEnvironment *ret;

    ret = SLAB_MEMORY_malloc(sizeof(LocalEnvironment));
    ret->variable = NULL;

    return ret;
}

static void
dispose_local_environment(LocalEnvironment *env)
{
    while (env->variable) {
	Variable	*temp;
	temp = env->variable->next;
	SLAB_MEMORY_free(env->variable);
	env->variable = temp;
    }
    SLAB_MEMORY_free(env);
}

static Value
eval_function_call_expression(LocalEnvironment *env,
			      char *identifier, Expression *argument)
{
    Value	result;
    Expression		*arg_p;
    ParameterList	*param_p;
    LocalEnvironment	*local_env;
    FunctionDefinition	*func;
    
    func = clc_search_function(identifier);
    if (func == NULL) {
	clc_runtime_error(FUNCTION_NOT_FOUND_ERR, "name..%s\n", identifier);
    }
    local_env = alloc_local_environment();

    for (arg_p = argument, param_p = func->parameter;
        arg_p;
        arg_p = arg_p->u.expression_list.next,
        param_p = param_p->next) {
	Value arg_val;

	if (param_p == NULL) {
	    clc_runtime_error(ARGUMENT_TOO_MANY_ERR, NULL);
	}
	arg_val = eval_expression(env,
				  arg_p->u.expression_list.expression);
	clc_add_local_variable(local_env, param_p->name, &arg_val);
    }
    if (param_p) {
	clc_runtime_error(ARGUMENT_TOO_FEW_ERR, NULL);
    }
    result = eval_expression(local_env, func->expression_list);

    dispose_local_environment(local_env);

    return result;
}

Value
eval_expression(LocalEnvironment *env, Expression *expr)
{
    Value	v;
    switch (expr->type) {

      case INT_EXPRESSION:
	v = eval_int_expression(expr->u.int_value);
	break;
      case DOUBLE_EXPRESSION:
	v = eval_double_expression(expr->u.double_value);
	break;
      case COMPLEX_EXPRESSION:
	v = eval_complex_expression(expr->u.complex_value.r, 
                                    expr->u.complex_value.i);
	break;
      case INTERVAL_EXPRESSION:
        v = eval_interval_expression(env,
              expr->u.interval_expression.lower,
              expr->u.interval_expression.upper);
        break;
      case INTERVAL_VALUE_EXPRESSION:
        v = eval_interval_value_expression(
              expr->u.interval_value.lower,
              expr->u.interval_value.upper);
        break;
      case INTERVAL_MATRIX_VALUE_EXPRESSION:
        v = eval_interval_matrix_value_expression(
              expr->u.interval_matrix_value.lower,
              expr->u.interval_matrix_value.upper);
        break;
      case IDENTIFIER_EXPRESSION:
	v = eval_identifier_expression(env, expr->u.identifier);
	break;
      case COUPLE_IDENTIFIER_EXPRESSION:
	v = eval_couple_identifier_expression(env, expr->u.couple_identifier_expression.identifier, 
                                                   expr->u.couple_identifier_expression.name);
	break;
      case COUPLE_EXPRESSION:
	v = eval_couple_expression(env, expr->u.couple_expression.name_first,
                                        expr->u.couple_expression.name_second);
	break;
      case TRIPLE_EXPRESSION:
	v = eval_triple_expression(env, expr->u.triple_expression.name_first,
                                        expr->u.triple_expression.name_second,
                                        expr->u.triple_expression.name_third);
	break;
      case PRINT_EXPRESSION:
	eval_print_expression(env, expr->u.identifier);
	break;
      case PLOT_EXPRESSION:
	eval_plot_expression(env, expr->u.plot_expression.name_first, 
                                   expr->u.plot_expression.name_second);
	break;
      case PLOT2_EXPRESSION:
	eval_plot2_expression(env, expr->u.plot2_expression.name_first, 
                                   expr->u.plot2_expression.name_second);
	break;
      case PLOT3_EXPRESSION:
	eval_plot3_expression(env, expr->u.plot3_expression.name_first, 
                                   expr->u.plot3_expression.name_second,
                                   expr->u.plot3_expression.name_third);
	break;
      case LINPRO_EXPRESSION:
	eval_linpro_expression(env, expr->u.linpro_expression.name_first, 
                                   expr->u.linpro_expression.name_second,
                                   expr->u.linpro_expression.name_third);
	break;
      case WHO_EXPRESSION:
	eval_who_expression(env);
	break;
      case MATRIX_ONES_EXPRESSION:
	v = eval_matrix_ones_expression(env, 
               expr->u.matrix_ones_expression.row,
               expr->u.matrix_ones_expression.column);
	break;
      case MATRIX_EYE_EXPRESSION:
	v = eval_matrix_eye_expression(env, 
               expr->u.matrix_eye_expression.row,
               expr->u.matrix_eye_expression.column);
	break;
      case MATRIX_RAND_EXPRESSION:
	v = eval_matrix_rand_expression(env, 
               expr->u.matrix_rand_expression.row,
               expr->u.matrix_rand_expression.column);
	break;
      case FFT_EXPRESSION:
	v = eval_fft_expression(env, 
               expr->u.fft_expression.data,
               expr->u.fft_expression.number);
	break;
      case IFFT_EXPRESSION:
	v = eval_ifft_expression(env, 
               expr->u.ifft_expression.data,
               expr->u.ifft_expression.number);
	break;
      case LREAD_EXPRESSION:
	v = eval_lread_expression();
	break;
      case MATRIX_ZEROS_EXPRESSION:
	v = eval_matrix_zeros_expression(env, 
               expr->u.matrix_zeros_expression.row,
               expr->u.matrix_zeros_expression.column);
	break;
      case LINSPACE_EXPRESSION:
	v = eval_linspace_expression(env, 
               expr->u.linspace_expression.start,
               expr->u.linspace_expression.end,
               expr->u.linspace_expression.number);
	break;
      case MATRIX_ELEMENT_ASSIGN_EXPRESSION:
	v = eval_matrix_element_assign_expression(env, 
               expr->u.matrix_element_assign_expression.variable,
               expr->u.matrix_element_assign_expression.row,
               expr->u.matrix_element_assign_expression.column,
               expr->u.matrix_element_assign_expression.operand);
	break;
      case MATRIX_ELEMENT_ASSIGN_EXPRESSION2:
	v = eval_matrix_element_assign_expression2(env, 
               expr->u.matrix_element_assign_expression.variable,
               expr->u.matrix_element_assign_expression.row,
               expr->u.matrix_element_assign_expression.operand);
	break;
      case MATRIX_ELEMENT_CALL_EXPRESSION:
	v = eval_matrix_element_call_expression(env, 
               expr->u.matrix_element_call_expression.identifier,
               expr->u.matrix_element_call_expression.row,
               expr->u.matrix_element_call_expression.column);
	break;
      case MATRIX_ELEMENT_CALL_EXPRESSION2:
	v = eval_matrix_element_call_expression2(env, 
               expr->u.matrix_element_call_expression.identifier,
               expr->u.matrix_element_call_expression.row);
	break;
      case EXPRESSION_LIST_EXPRESSION:
	v = eval_expression_list_expression
	    (env,
	     expr->u.expression_list.expression,
	     expr->u.expression_list.next);
	break;
      case ASSIGN_EXPRESSION:
	v = eval_assign_expression(env,
				   expr->u.assign_expression.variable,
				   expr->u.assign_expression.operand);
	break;
      case ADD_EXPRESSION:	/* FALLTHRU */
      case SUB_EXPRESSION:	/* FALLTHRU */
      case MUL_EXPRESSION:	/* FALLTHRU */
      case DIV_EXPRESSION:	/* FALLTHRU */
      case MOD_EXPRESSION:	/* FALLTHRU */
      case EQ_EXPRESSION:	/* FALLTHRU */
      case NE_EXPRESSION:	/* FALLTHRU */
      case OR_EXPRESSION:	/* FALLTHRU */
      case AND_EXPRESSION:	/* FALLTHRU */
      case GT_EXPRESSION:	/* FALLTHRU */
      case GE_EXPRESSION:	/* FALLTHRU */
      case LT_EXPRESSION:	/* FALLTHRU */
      case LE_EXPRESSION:	/* FALLTHRU */
      case SOLVE_EXPRESSION:   /* FALLTHRU */
      case EXACT_SOLVE_EXPRESSION:   /* FALLTHRU */
	v = clc_eval_binary_expression(env,
				       expr->type,
				       expr->u.binary_expression.left,
				       expr->u.binary_expression.right);
	break;
      case MINUS_EXPRESSION:
	v = clc_eval_minus_expression(env, expr->u.minus_expression);
	break;
      case DIF_EXPRESSION:
	v = clc_eval_dif_expression(env, expr->u.dif_expression.operand,
                                    expr->u.dif_expression.n, expr->u.dif_expression.i);
	break;
      case NORM_EXPRESSION:
	v = clc_eval_norm_expression(env, expr->u.norm_expression);
	break;
      case INV_EXPRESSION:
	v = clc_eval_inv_expression(env, expr->u.inv_expression);
	break;
      case REAL_EXPRESSION:
	v = clc_eval_real_expression(env, expr->u.real_expression);
	break;
      case IMAG_EXPRESSION:
	v = clc_eval_imag_expression(env, expr->u.imag_expression);
	break;
      case UPPER_EXPRESSION:
	v = clc_eval_upper_expression(env, expr->u.upper_expression);
	break;
      case LOWER_EXPRESSION:
	v = clc_eval_lower_expression(env, expr->u.lower_expression);
	break;
      case TRANSPOSE_EXPRESSION:
	v = clc_eval_transpose_expression(env, expr->u.transpose_expression);
	break;
      case EIG_EXPRESSION:
	v = clc_eval_eig_expression(env, expr->u.eig_expression);
	break;
      case TRI_EXPRESSION:
	v = clc_eval_tri_expression(env, expr->u.eig_expression);
	break;
      case VEIG_EXPRESSION:
	v = clc_eval_veig_expression(env, expr->u.veig_expression);
	break;
      case SCHUR_EXPRESSION:
	v = clc_eval_schur_expression(env, expr->u.schur_expression);
	break;
      case LU_EXPRESSION:
	v = clc_eval_lu_expression(env, expr->u.lu_expression);
	break;
      case QR_EXPRESSION:
	v = clc_eval_qr_expression(env, expr->u.qr_expression);
	break;
      case CHOL_EXPRESSION:
	v = clc_eval_chol_expression(env, expr->u.chol_expression);
	break;
      case SVD_EXPRESSION:
	v = clc_eval_svd_expression(env, expr->u.svd_expression);
	break;
      case VSIN_EXPRESSION:
	v = clc_eval_vsin_expression(env, expr->u.vsin_expression);
	break;
      case SIN_EXPRESSION:
	v = clc_eval_sin_expression(env, expr->u.sin_expression);
	break;
      case POW_EXPRESSION:
	v = clc_eval_pow_expression(env, expr->u.pow_expression.left,
                                         expr->u.pow_expression.right);
	break;
      case COS_EXPRESSION:
	v = clc_eval_cos_expression(env, expr->u.cos_expression);
	break;
      case VCOS_EXPRESSION:
	v = clc_eval_vcos_expression(env, expr->u.cos_expression);
	break;
      case LOG_EXPRESSION:
	v = clc_eval_log_expression(env, expr->u.log_expression);
	break;
      case VLOG_EXPRESSION:
	v = clc_eval_vlog_expression(env, expr->u.log_expression);
	break;
      case LOGTEN_EXPRESSION:
	v = clc_eval_log10_expression(env, expr->u.log10_expression);
	break;
      case EXP_EXPRESSION:
	v = clc_eval_exp_expression(env, expr->u.exp_expression);
	break;
      case VEXP_EXPRESSION:
	v = clc_eval_vexp_expression(env, expr->u.exp_expression);
	break;
      case FABS_EXPRESSION:
	v = clc_eval_fabs_expression(env, expr->u.fabs_expression);
	break;
      case TAN_EXPRESSION:
	v = clc_eval_tan_expression(env, expr->u.tan_expression);
	break;
      case VTAN_EXPRESSION:
	v = clc_eval_vtan_expression(env, expr->u.vtan_expression);
	break;
      case ATAN_EXPRESSION:
	v = clc_eval_atan_expression(env, expr->u.atan_expression);
	break;
      case ACOS_EXPRESSION:
	v = clc_eval_acos_expression(env, expr->u.acos_expression);
	break;
      case ASIN_EXPRESSION:
	v = clc_eval_asin_expression(env, expr->u.asin_expression);
	break;
      case COSH_EXPRESSION:
	v = clc_eval_cosh_expression(env, expr->u.cosh_expression);
	break;
      case SINH_EXPRESSION:
	v = clc_eval_sinh_expression(env, expr->u.sinh_expression);
	break;
      case TANH_EXPRESSION:
	v = clc_eval_tanh_expression(env, expr->u.tanh_expression);
	break;
      case SQRT_EXPRESSION:
	v = clc_eval_sqrt_expression(env, expr->u.sqrt_expression);
	break;
      case SPLIT_EXPRESSION:
	v = clc_eval_split_expression(env, expr->u.split_expression);
	break;
      case SIZE_EXPRESSION:
	v = clc_eval_size_expression(env, expr->u.size_expression);
	break;
      case NAME_EXPRESSION:
	v.type = NAME_VALUE;
        v.u.name_value.name 
            = (char *)clc_global_malloc(strlen(expr->u.name_expression.name)+1);
        strcpy(v.u.name_value.name, expr->u.name_expression.name);
	break;
      case UP_EXPRESSION:
	v = clc_eval_up_expression(env, expr->u.up_expression);
	break;
      case DOWN_EXPRESSION:
	v = clc_eval_down_expression(env, expr->u.down_expression);
	break;
      case NEAR_EXPRESSION:
	v = clc_eval_near_expression(env, expr->u.near_expression);
	break;
      case TIC_EXPRESSION:
	v = clc_eval_tic_expression(env, expr->u.tic_expression);
	break;
      case TOC_EXPRESSION:
	v = clc_eval_toc_expression(env, expr->u.toc_expression);
	break;
      case MATRIX_EXPRESSION:
        v = clc_eval_matrix_expression(env, 
                                       expr->u.matrix_expression.top,
                                       expr->u.matrix_expression.column,
                                       expr->u.matrix_expression.row);
        break;
      case MATRIX:
        v = clc_eval_matrix(env, expr->u.matrix.matrix_element);
        break;
      case MATRIX_VALUE_EXPRESSION:
        v = clc_eval_matrix_value_expression(env, 
                                       expr->u.matrix_value.top,
                                       expr->u.matrix_value.column,
                                       expr->u.matrix_value.row);
        break;
      case COMPLEX_MATRIX_VALUE_EXPRESSION:
        v = clc_eval_complex_matrix_value_expression(env, 
                                       expr->u.complex_matrix_value.top,
                                       expr->u.complex_matrix_value.column,
                                       expr->u.complex_matrix_value.row);
        break;
      case MATRIX_ELEMENT_EXPRESSION:
        v = clc_eval_matrix_element_expression(env, 
                     expr->u.matrix_element_expression.element);
        break;
      case IF_EXPRESSION:
	v = eval_if_expression(env,
			       expr->u.if_expression.condition,
			       expr->u.if_expression.then_expression,
			       expr->u.if_expression.else_expression);
	break;
      case FOR_EXPRESSION:
	v = eval_for_expression(env,
				  expr->u.for_expression.first_condition,
				  expr->u.for_expression.second_condition,
				  expr->u.for_expression.third_condition,
				  expr->u.for_expression.expression_list);
	break;
      case WHILE_EXPRESSION:
	v = eval_while_expression(env,
				  expr->u.while_expression.condition,
				  expr->u.while_expression.expression_list);
	break;
      case DO_WHILE_EXPRESSION:
	v = eval_do_while_expression(env,
				  expr->u.do_while_expression.expression_list,
				  expr->u.do_while_expression.condition);
	break;
      case FUNCTION_CALL_EXPRESSION:
	v = eval_function_call_expression
	    (env,
	     expr->u.function_call_expression.identifier,
	     expr->u.function_call_expression.argument);
	break;
      case EVAL_EXPRESSION:
	v = eval_eval_expression
	    (env, expr->u.eval_expression.name,
	     expr->u.eval_expression.argument);
	break;
      case CON_EXPRESSION:
	v = clc_eval_con_expression(env,                                                          expr->u.con_expression.exp, 
              expr->u.con_expression.next);
	break;
      case MAX_EXPRESSION:
	v = clc_eval_max_expression(env,                                                          expr->u.max_expression.left, 
              expr->u.max_expression.right);
	break;
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }

    return v;
}

void
clc_eval_expression_with_print(Expression *expression)
{
    Value		v;
    RuntimeError	error_id;
    long int i, j;
    if ((error_id = 
           (RuntimeError)setjmp(clc_current_interpreter
		 ->error_recovery_environment)) == 0) {
        v = eval_expression(NULL, expression);
        switch(v.type) {
            case INT_VALUE:
                printf("ans = \n     %d\n", v.u.int_value);
                break;
            case NULL_VALUE:
            case DOUBLE_VALUE:
                printf("ans = \n     %10.5f\n", v.u.double_value);
                break;
            case COMPLEX_VALUE:
                if (v.u.complex_value.i > 0.0) {
                    printf("ans = \n     %10.5f + %10.5f i\n",v.u.complex_value.r,
                                             v.u.complex_value.i);
                } else {
                    printf("ans = \n     %10.5f - %10.5f i\n",v.u.complex_value.r,
                                             -v.u.complex_value.i);
                }
                break;
            case INTERVAL_VALUE:
                printf("ans =\n    [%10.5f , %10.5f ]\n",
                    v.u.interval_value.lower,
                    v.u.interval_value.upper);
                break;
            case INTERVAL_MATRIX_VALUE:
                printf("ans =\n");
                for(i = 0; i < v.u.interval_matrix_value.lower.row; i++) {
                    printf("      |");
                    for(j = 0; j < v.u.interval_matrix_value.lower.column; j++) {
                        printf(" [%10.5f,%10.5f]", 
                            v.u.interval_matrix_value.lower.top[i + 
                                        j * v.u.interval_matrix_value.lower.row], 
                            v.u.interval_matrix_value.upper.top[i + 
                                        j * v.u.interval_matrix_value.lower.row]); 
                    }
                    printf("   |\n");
                } 
                break;
            case MATRIX_VALUE:
                printf("ans = \n");
                for(i = 0; i < v.u.matrix_value.row; i++) {
                    printf("      |");
                    for(j = 0; j < v.u.matrix_value.column; j++) {
                        printf(" %10.5f", 
                            v.u.matrix_value.top[i + j * v.u.matrix_value.row]);
                    }
                    printf("   |\n");
                }
                break; 
            case COMPLEX_MATRIX_VALUE:
                printf("ans = \n");
                for(i = 0; i < v.u.complex_matrix_value.row; i++) {
                    printf("      |");
                    for(j = 0; j < v.u.complex_matrix_value.column; j++) {
                        if( v.u.complex_matrix_value.top[i + 
                                j * v.u.matrix_value.row].i == 0.0) {
                            printf(" %10.5f           ",
                                v.u.complex_matrix_value.top[i + 
                                j * v.u.matrix_value.row].r);
                        } else {
                            if (v.u.complex_matrix_value.top[i +
                                j * v.u.matrix_value.row].i >= 0.0) {
                                printf(" %10.5f + %10.5f i", 
                                    v.u.complex_matrix_value.top[i + 
                                    j * v.u.matrix_value.row].r,
                                    v.u.complex_matrix_value.top[i + 
                                    j * v.u.matrix_value.row].i);
                            } else {
                                printf(" %10.5f - %10.5f i", 
                                    v.u.complex_matrix_value.top[i + 
                                    j * v.u.matrix_value.row].r,
                                    -v.u.complex_matrix_value.top[i + 
                                    j * v.u.matrix_value.row].i);
                            }
                        }
                    }
                    printf("   |\n");
                } 
                break;
            case DIF_VALUE:
                printf("ans = \n");
                printf("      %10.5f <",v.u.dif_value.value);
                for(i = 0; i < v.u.dif_value.n; i++) {
                    printf(" %10.5f ", 
                        v.u.dif_value.jacobi[i]);
                }
                printf(">\n");
                break; 
            case DIF_MATRIX_VALUE:
                printf("ans = \n");
                for(i = 0; i < v.u.difmatrix_value.n; i++) {
                    printf("      %10.5f <", v.u.difmatrix_value.value[i]);
                    for(j = 0; j < v.u.difmatrix_value.column; j++) {
                        printf(" %10.5f", 
                            v.u.difmatrix_value.jacobi[i + j * v.u.difmatrix_value.n]);
                    }
                    printf(" > \n");
                }
                break; 
            case NAME_VALUE:
                printf("ans = \n");
                printf("      %s \n", v.u.name_value.name);
                break; 
            case COUPLE_VALUE:
                printf("ans.%s =\n", v.u.couple_value.name_first);
                switch(v.u.couple_value.type_first) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        printf("    \n     %10.5f\n", 
                           v.u.couple_value.u_first.double_value);
                        break;
                    case COMPLEX_VALUE:
                        if (v.u.couple_value.u_first.complex_value.i > 0.0) {
                            printf("    \n     %10.5f + %10.5f i\n",
                               v.u.couple_value.u_first.complex_value.r,
                                         v.u.couple_value.u_first.complex_value.i);
                        } else {
                            printf("    \n     %10.5f - %10.5f i\n",
                               v.u.couple_value.u_first.complex_value.r,
                                         -v.u.couple_value.u_first.complex_value.i);
                        }
                        break;
                    case INTERVAL_VALUE:
                        printf("    \n    [%10.5f , %10.5f ]\n",
                            v.u.couple_value.u_first.interval_value.lower,
                            v.u.couple_value.u_first.interval_value.upper);
                    break;
                    case INTERVAL_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.couple_value.u_first.interval_matrix_value.lower.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.couple_value.u_first.interval_matrix_value.lower.column; j++) {
                                printf(" [%10.5f,%10.5f]", 
                                    v.u.couple_value.u_first.interval_matrix_value.lower.top[i + 
                                    j * v.u.couple_value.u_first.interval_matrix_value.lower.row], 
                                    v.u.interval_matrix_value.upper.top[i + 
                                         j * v.u.couple_value.u_first.interval_matrix_value.lower.row]); 
                            }
                            printf("   |\n");
                        } 
                        break;
                    case MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.couple_value.u_first.matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.couple_value.u_first.matrix_value.column; j++) {
                                printf(" %10.5f", 
                                    v.u.couple_value.u_first.matrix_value.top[i + j * v.u.couple_value.u_first.matrix_value.row]);
                            }
                            printf("   |\n");
                        } 
                        break; 
                    case COMPLEX_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.couple_value.u_first.complex_matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.couple_value.u_first.complex_matrix_value.column; j++) {
                                if( v.u.couple_value.u_first.complex_matrix_value.top[i + 
                                        j * v.u.couple_value.u_first.matrix_value.row].i == 0.0) {
                                    printf(" %6.3f           ",
                                        v.u.couple_value.u_first.complex_matrix_value.top[i + 
                                        j * v.u.couple_value.u_first.matrix_value.row].r);
                                 } else {
                                     if (v.u.couple_value.u_first.complex_matrix_value.top[i +
                                     j * v.u.couple_value.u_first.matrix_value.row].i >= 0.0) {
                                 printf(" %6.3f + %6.3f i", 
                                         v.u.couple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_first.matrix_value.row].r,
                                         v.u.couple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_first.matrix_value.row].i);
                                 } else {
                                     printf(" %6.3f - %6.3f i", 
                                         v.u.couple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_first.matrix_value.row].r,
                                         -v.u.couple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_first.matrix_value.row].i);
                                 }
                             }
                        }
                        printf("   |\n");
                    } 
                    break;
                }
                printf("\n");
                printf("ans.%s =\n", v.u.couple_value.name_second);
                switch(v.u.couple_value.type_second) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        printf("    \n     %10.5f\n", 
                           v.u.couple_value.u_second.double_value);
                        break;
                    case COMPLEX_VALUE:
                        if (v.u.couple_value.u_second.complex_value.i > 0.0) {
                            printf("    \n     %10.5f + %10.5f i\n",
                               v.u.couple_value.u_second.complex_value.r,
                                         v.u.couple_value.u_second.complex_value.i);
                        } else {
                            printf("    \n     %10.5f - %10.5f i\n",
                               v.u.couple_value.u_second.complex_value.r,
                                         -v.u.couple_value.u_second.complex_value.i);
                        }
                        break;
                    case INTERVAL_VALUE:
                        printf("    \n    [%10.5f , %10.5f ]\n",
                            v.u.couple_value.u_second.interval_value.lower,
                            v.u.couple_value.u_second.interval_value.upper);
                    break;
                    case INTERVAL_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.couple_value.u_second.interval_matrix_value.lower.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.couple_value.u_second.interval_matrix_value.lower.column; j++) {
                                printf(" [%10.5f,%10.5f]", 
                                    v.u.couple_value.u_second.interval_matrix_value.lower.top[i + 
                                    j * v.u.couple_value.u_second.interval_matrix_value.lower.row], 
                                    v.u.interval_matrix_value.upper.top[i + 
                                         j * v.u.couple_value.u_second.interval_matrix_value.lower.row]); 
                            }
                            printf("   |\n");
                        } 
                        break;
                    case MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.couple_value.u_second.matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.couple_value.u_second.matrix_value.column; j++) {
                                printf(" %10.5f", 
                                    v.u.couple_value.u_second.matrix_value.top[i + j * v.u.couple_value.u_second.matrix_value.row]);
                            }
                            printf("   |\n");
                        } 
                        break; 
                    case COMPLEX_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.couple_value.u_second.complex_matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.couple_value.u_second.complex_matrix_value.column; j++) {
                                if( v.u.couple_value.u_second.complex_matrix_value.top[i + 
                                        j * v.u.couple_value.u_second.matrix_value.row].i == 0.0) {
                                    printf(" %6.3f           ",
                                        v.u.couple_value.u_second.complex_matrix_value.top[i + 
                                        j * v.u.couple_value.u_second.matrix_value.row].r);
                                 } else {
                                     if (v.u.couple_value.u_second.complex_matrix_value.top[i +
                                     j * v.u.couple_value.u_second.matrix_value.row].i >= 0.0) {
                                 printf(" %6.3f + %6.3f i", 
                                         v.u.couple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_second.matrix_value.row].r,
                                         v.u.couple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_second.matrix_value.row].i);
                                 } else {
                                     printf(" %6.3f - %6.3f i", 
                                         v.u.couple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_second.matrix_value.row].r,
                                         -v.u.couple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.couple_value.u_second.matrix_value.row].i);
                                 }
                             }
                        }
                        printf("   |\n");
                    }
                } 
                break; 
            case TRIPLE_VALUE:
                printf("ans.%s =\n", v.u.triple_value.name_first);
                switch(v.u.triple_value.type_first) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        printf("    \n     %10.5f\n", 
                           v.u.triple_value.u_first.double_value);
                        break;
                    case COMPLEX_VALUE:
                        if (v.u.triple_value.u_first.complex_value.i > 0.0) {
                            printf("    \n     %10.5f + %10.5f i\n",
                               v.u.triple_value.u_first.complex_value.r,
                                         v.u.triple_value.u_first.complex_value.i);
                        } else {
                            printf("    \n     %10.5f - %10.5f i\n",
                               v.u.triple_value.u_first.complex_value.r,
                                         -v.u.triple_value.u_first.complex_value.i);
                        }
                        break;
                    case INTERVAL_VALUE:
                        printf("    \n    [%10.5f , %10.5f ]\n",
                            v.u.triple_value.u_first.interval_value.lower,
                            v.u.triple_value.u_first.interval_value.upper);
                    break;
                    case INTERVAL_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_first.interval_matrix_value.lower.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_first.interval_matrix_value.lower.column; j++) {
                                printf(" [%10.5f,%10.5f]", 
                                    v.u.triple_value.u_first.interval_matrix_value.lower.top[i + 
                                    j * v.u.triple_value.u_first.interval_matrix_value.lower.row], 
                                    v.u.interval_matrix_value.upper.top[i + 
                                         j * v.u.triple_value.u_first.interval_matrix_value.lower.row]); 
                            }
                            printf("   |\n");
                        } 
                        break;
                    case MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_first.matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_first.matrix_value.column; j++) {
                                printf(" %10.5f", 
                                    v.u.triple_value.u_first.matrix_value.top[i + j * v.u.triple_value.u_first.matrix_value.row]);
                            }
                            printf("   |\n");
                        } 
                        break; 
                    case COMPLEX_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_first.complex_matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_first.complex_matrix_value.column; j++) {
                                if( v.u.triple_value.u_first.complex_matrix_value.top[i + 
                                        j * v.u.triple_value.u_first.matrix_value.row].i == 0.0) {
                                    printf(" %6.3f           ",
                                        v.u.triple_value.u_first.complex_matrix_value.top[i + 
                                        j * v.u.triple_value.u_first.matrix_value.row].r);
                                 } else {
                                     if (v.u.triple_value.u_first.complex_matrix_value.top[i +
                                     j * v.u.triple_value.u_first.matrix_value.row].i >= 0.0) {
                                 printf(" %6.3f + %6.3f i", 
                                         v.u.triple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_first.matrix_value.row].r,
                                         v.u.triple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_first.matrix_value.row].i);
                                 } else {
                                     printf(" %6.3f - %6.3f i", 
                                         v.u.triple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_first.matrix_value.row].r,
                                         -v.u.triple_value.u_first.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_first.matrix_value.row].i);
                                 }
                             }
                        }
                        printf("   |\n");
                    } 
                    break;
                }
                printf("\n");
                printf("ans.%s =\n", v.u.triple_value.name_second);
                switch(v.u.triple_value.type_second) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        printf("    \n     %10.5f\n", 
                           v.u.triple_value.u_second.double_value);
                        break;
                    case COMPLEX_VALUE:
                        if (v.u.triple_value.u_second.complex_value.i > 0.0) {
                            printf("    \n     %10.5f + %10.5f i\n",
                               v.u.triple_value.u_second.complex_value.r,
                                         v.u.couple_value.u_second.complex_value.i);
                        } else {
                            printf("    \n     %10.5f - %10.5f i\n",
                               v.u.triple_value.u_second.complex_value.r,
                                         -v.u.couple_value.u_second.complex_value.i);
                        }
                        break;
                    case INTERVAL_VALUE:
                        printf("    \n    [%10.5f , %10.5f ]\n",
                            v.u.triple_value.u_second.interval_value.lower,
                            v.u.triple_value.u_second.interval_value.upper);
                    break;
                    case INTERVAL_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_second.interval_matrix_value.lower.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_second.interval_matrix_value.lower.column; j++) {
                                printf(" [%10.5f,%10.5f]", 
                                    v.u.triple_value.u_second.interval_matrix_value.lower.top[i + 
                                    j * v.u.triple_value.u_second.interval_matrix_value.lower.row], 
                                    v.u.interval_matrix_value.upper.top[i + 
                                         j * v.u.triple_value.u_second.interval_matrix_value.lower.row]); 
                            }
                            printf("   |\n");
                        } 
                        break;
                    case MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_second.matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_second.matrix_value.column; j++) {
                                printf(" %10.5f", 
                                    v.u.triple_value.u_second.matrix_value.top[i + j * v.u.triple_value.u_second.matrix_value.row]);
                            }
                            printf("   |\n");
                        } 
                        break; 
                    case COMPLEX_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_second.complex_matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_second.complex_matrix_value.column; j++) {
                                if( v.u.triple_value.u_second.complex_matrix_value.top[i + 
                                        j * v.u.triple_value.u_second.matrix_value.row].i == 0.0) {
                                    printf(" %6.3f           ",
                                        v.u.triple_value.u_second.complex_matrix_value.top[i + 
                                        j * v.u.triple_value.u_second.matrix_value.row].r);
                                 } else {
                                     if (v.u.triple_value.u_second.complex_matrix_value.top[i +
                                     j * v.u.triple_value.u_second.matrix_value.row].i >= 0.0) {
                                 printf(" %6.3f + %6.3f i", 
                                         v.u.triple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_second.matrix_value.row].r,
                                         v.u.triple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_second.matrix_value.row].i);
                                 } else {
                                     printf(" %6.3f - %6.3f i", 
                                         v.u.triple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_second.matrix_value.row].r,
                                         -v.u.triple_value.u_second.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_second.matrix_value.row].i);
                                 }
                             }
                        }
                        printf("   |\n");
                    }
                } 
                printf("\n");
                printf("ans.%s =\n", v.u.triple_value.name_third);
                switch(v.u.triple_value.type_third) {
                    case INT_VALUE:
                    case COUPLE_VALUE:
                    case TRIPLE_VALUE:
                    case NULL_VALUE:
                    case DOUBLE_VALUE:
                        printf("    \n     %10.5f\n", 
                           v.u.triple_value.u_third.double_value);
                        break;
                    case COMPLEX_VALUE:
                        if (v.u.triple_value.u_third.complex_value.i > 0.0) {
                            printf("    \n     %10.5f + %10.5f i\n",
                               v.u.triple_value.u_third.complex_value.r,
                                         v.u.triple_value.u_third.complex_value.i);
                        } else {
                            printf("    \n     %10.5f - %10.5f i\n",
                               v.u.triple_value.u_third.complex_value.r,
                                         -v.u.triple_value.u_third.complex_value.i);
                        }
                        break;
                    case INTERVAL_VALUE:
                        printf("    \n    [%10.5f , %10.5f ]\n",
                            v.u.triple_value.u_third.interval_value.lower,
                            v.u.triple_value.u_third.interval_value.upper);
                    break;
                    case INTERVAL_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_third.interval_matrix_value.lower.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_third.interval_matrix_value.lower.column; j++) {
                                printf(" [%10.5f,%10.5f]", 
                                    v.u.triple_value.u_third.interval_matrix_value.lower.top[i + 
                                    j * v.u.triple_value.u_third.interval_matrix_value.lower.row], 
                                    v.u.interval_matrix_value.upper.top[i + 
                                         j * v.u.triple_value.u_third.interval_matrix_value.lower.row]); 
                            }
                            printf("   |\n");
                        } 
                        break;
                    case MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_third.matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_third.matrix_value.column; j++) {
                                printf(" %10.5f", 
                                    v.u.triple_value.u_third.matrix_value.top[i + j * v.u.triple_value.u_third.matrix_value.row]);
                            }
                            printf("   |\n");
                        } 
                        break; 
                    case COMPLEX_MATRIX_VALUE:
                        printf("    \n");
                        for(i = 0; i < v.u.triple_value.u_third.complex_matrix_value.row; i++) {
                            printf("      |");
                            for(j = 0; j < v.u.triple_value.u_third.complex_matrix_value.column; j++) {
                                if( v.u.triple_value.u_third.complex_matrix_value.top[i + 
                                        j * v.u.triple_value.u_third.matrix_value.row].i == 0.0) {
                                    printf(" %6.3f           ",
                                        v.u.triple_value.u_third.complex_matrix_value.top[i + 
                                        j * v.u.triple_value.u_third.matrix_value.row].r);
                                 } else {
                                     if (v.u.triple_value.u_third.complex_matrix_value.top[i +
                                     j * v.u.triple_value.u_third.matrix_value.row].i >= 0.0) {
                                 printf(" %6.3f + %6.3f i", 
                                         v.u.triple_value.u_third.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_third.matrix_value.row].r,
                                         v.u.triple_value.u_third.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_third.matrix_value.row].i);
                                 } else {
                                     printf(" %6.3f - %6.3f i", 
                                         v.u.triple_value.u_third.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_third.matrix_value.row].r,
                                         -v.u.triple_value.u_third.complex_matrix_value.top[i + 
                                         j * v.u.triple_value.u_third.matrix_value.row].i);
                                 }
                             }
                        }
                        printf("   |\n");
                    }
                } 
                break;
        }
        clc_reopen_current_storage();
    }
}

static void 
clc_print_value(Value v) 
{
    int i,j;
    if (v.type == DOUBLE_VALUE) {
        printf("ans = \n     %22.17f\n", v.u.double_value);
    } else if (v.type == COMPLEX_VALUE) {
        printf("ans = \n     %10.5f + %10.5f i\n",v.u.complex_value.r,
                    v.u.complex_value.i);
    } else if (v.type == INTERVAL_VALUE) {
        printf("ans =\n    [%10.5f , %10.5f ]\n",
            v.u.interval_value.lower,
            v.u.interval_value.upper);
    } else if (v.type == INTERVAL_MATRIX_VALUE) {
        printf("ans =\n");
        for(i = 0; i < v.u.interval_matrix_value.lower.row; i++) {
            printf("      |");
            for(j = 0; j < v.u.interval_matrix_value.lower.column; j++) {
                printf(" [%10.5f,%10.5f]", 
                  v.u.interval_matrix_value.lower.top
                     [i + j * v.u.interval_matrix_value.lower.row], 
                  v.u.interval_matrix_value.upper.top
                     [i + j * v.u.interval_matrix_value.lower.row]); 
            }
            printf("   |\n");
         } 
     } else if (v.type == MATRIX_VALUE) {
         printf("ans = \n");
         for(i = 0; i < v.u.matrix_value.row; i++) {
             printf("      |");
             for(j = 0; j < v.u.matrix_value.column; j++) {
                 printf(" %10.5f", 
                    v.u.matrix_value.top[i + j * v.u.matrix_value.row]);
             }
             printf("   |\n");
         } 
     } 
} 

void 
clc_fprint_value(Value v, FILE *fp) 
{
    int i,j;
    char *p;
    if (v.type == DOUBLE_VALUE) {
        p = "DOUBLE_VALUE";
        fprintf(fp, "%s\n", p);
        fprintf(fp, "%22.17f\n", v.u.double_value);
    } else if (v.type == COMPLEX_VALUE) {
        fprintf(fp, "%22.17f,   %22.17f",v.u.complex_value.r,
                    v.u.complex_value.i);
    } else if (v.type == INTERVAL_VALUE) {
        fprintf(fp,"%22.17f,   %22.17f",
            v.u.interval_value.lower,
            v.u.interval_value.upper);
    } else if (v.type == INTERVAL_MATRIX_VALUE) {
        for(i = 0; i < v.u.interval_matrix_value.lower.row; i++) {
            for(j = 0; j < v.u.interval_matrix_value.lower.column; j++) {
                fprintf(fp," %22.17f,   %22.17f   ", 
                  v.u.interval_matrix_value.lower.top
                     [i + j * v.u.interval_matrix_value.lower.row], 
                  v.u.interval_matrix_value.upper.top
                     [i + j * v.u.interval_matrix_value.lower.row]); 
            }
            fprintf(fp,"\n");
         } 
     } else if (v.type == MATRIX_VALUE) {
         for(i = 0; i < v.u.matrix_value.row; i++) {
             for(j = 0; j < v.u.matrix_value.column; j++) {
                 fprintf(fp," %22.17f", 
                    v.u.matrix_value.top[i + j * v.u.matrix_value.row]);
             }
             fprintf(fp,"\n");
         } 
     } 
} 


void
clc_eval_expression(Expression *expression)
{
    Value		v;
    RuntimeError	error_id;

    if ((error_id = (RuntimeError)setjmp(clc_current_interpreter
					 ->error_recovery_environment)) == 0) 
    {
	v = eval_expression(NULL, expression);
    } 
    else 
    {
    }
    clc_reopen_current_storage();
}


void 
eval_print_expression(LocalEnvironment *env, char *identifier)
{
    Variable    *pos;
    
    if (env != NULL) {
       for(pos = env->variable; pos; pos = pos->next) {
          if (!strcmp(pos->name, identifier))
             break;
       }
    }
    if (pos != NULL) {
        clc_print_value(pos->value);
    }    
    for(pos = clc_current_interpreter->variable; pos; pos = pos->next) {
       if (!strcmp(pos->name, identifier))
          break;
    }
    if (pos != NULL) {
        clc_print_value(pos->value);
    }    
}

/*Value
eval_plot_expression(LocalEnvironment *env, char *first, char *second)
{
    Variable    *pos;
    Variable    *pos1;
    FILE *fp;
    int m1, n1, m2, n2;
    int i, j;
    Value v;
printf("plot\n");
    fp = fopen("_temp", "w");
    if (fp == NULL) 
    {
         fprintf(stderr, "   Error:canonot open file!\n");
    } 
    else 
    {
        if (env != NULL) {
            for(pos = env->variable; pos; pos = pos->next) {
                if (!strcmp(pos->name, first))
                     break;
            }
        }
        if (pos == NULL) 
        {
            for(pos = clc_current_interpreter->variable; pos; pos = pos->next) 
            {
                if (!strcmp(pos->name, first))
                break;
            }
            if (pos == NULL) {
                clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", first);
            }
        }    
        if (env != NULL) {
            for(pos1 = env->variable; pos1; pos1 = pos1->next) {
                if (!strcmp(pos1->name, second))
                     break;
            }
        }
        if (pos1 == NULL) 
        {
            for(pos1 = clc_current_interpreter->variable; pos1; pos1 = pos->next) 
            {
                if (!strcmp(pos1->name, second))
                break;
            }
            if (pos1 == NULL) {
                clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", second);
            }
        }    
        if(pos->value.type != MATRIX_VALUE) 
        {
            clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", first);
        }
        if(pos1->value.type != MATRIX_VALUE) 
        {
            clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", first);
        } 
        m1=pos->value.u.matrix_value.column;
        n1=pos->value.u.matrix_value.row;
        m2=pos1->value.u.matrix_value.column;
        n2=pos1->value.u.matrix_value.row;
        if (m1*n1 != m2*n2)
        {
            clc_runtime_error(NOT_SAME_DIMENSION_MATRIX_ERR, "(%s)\n", first);
        }
        for(i = 0; i < m1*n1; i++) 
        {
            printf("%22.17lg,     %22.17lg\n", 
                pos->value.u.matrix_value.top[i], 
                pos1->value.u.matrix_value.top[i]); 
        } 
        fclose(fp);
        return pos->value;
    }
} */

void 
eval_plot_expression(LocalEnvironment *env, char *identifier, char *name_second)
{
    Variable    *pos;
    Variable    *pos1;
    FILE *fp;
    int m1, n1, m2, n2;
    int i;
    fp = fopen("./plot/_temp", "w");
    if (fp == NULL) 
    {
         clc_runtime_error(CANNOT_OPEN_FILE_ERR, "(%s)\n", NULL);
    } 
    pos = NULL;
    pos1 = NULL;
    if (env != NULL) {
       for(pos = env->variable; pos; pos = pos->next) {
          if (!strcmp(pos->name, identifier))
             break;
       }
    }
    if (pos == NULL) 
    {
        for(pos = clc_current_interpreter->variable; pos; pos = pos->next) 
        {
            if (!strcmp(pos->name, identifier))
                break;
        }
    }
    if (pos == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);
    }    
    if (env != NULL) 
    {
       for(pos1 = env->variable; pos1; pos1 = pos1->next) 
       {
          if (!strcmp(pos->name, name_second))
             break;
       }
    }
    if (pos1 == NULL) 
    {
        for(pos1 = clc_current_interpreter->variable; pos1; pos1 = pos1->next) 
        {
            if (!strcmp(pos1->name, name_second))
                break;
        }
    }
    if (pos1 == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_second);
    }    
    if(pos->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", identifier);
    }
    if(pos1->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", name_second);
    } 
    m1=pos->value.u.matrix_value.column;
    n1=pos->value.u.matrix_value.row;
    m2=pos1->value.u.matrix_value.column;
    n2=pos1->value.u.matrix_value.row;
    if (m1*n1 != m2*n2)
    {
        clc_runtime_error(NOT_SAME_DIMENSION_MATRIX_ERR, "(%s)\n", identifier);
    }
    for(i = 0; i < m1*n1; i++) 
    {
        fprintf(fp,"%22.17f     %22.17f\n", 
            pos->value.u.matrix_value.top[i], 
            pos1->value.u.matrix_value.top[i]); 
    } 
    fclose(fp);
    fp = popen("gnuplot './plot/_plot'", "r");
    pclose(fp);
}

void 
eval_plot2_expression(LocalEnvironment *env, char *identifier, char *name_second)
{
    Variable    *pos;
    Variable    *pos1;
    FILE *fp;
    int m1, n1, m2, n2;
    int i;
    fp = fopen("./plot/_temp", "w");
    if (fp == NULL) 
    {
         clc_runtime_error(CANNOT_OPEN_FILE_ERR, "(%s)\n", NULL);
    } 
    pos = NULL;
    pos1 = NULL;
    if (env != NULL) {
       for(pos = env->variable; pos; pos = pos->next) {
          if (!strcmp(pos->name, identifier))
             break;
       }
    }
    if (pos == NULL) 
    {
        for(pos = clc_current_interpreter->variable; pos; pos = pos->next) 
        {
            if (!strcmp(pos->name, identifier))
                break;
        }
    }
    if (pos == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);
    }    
    if (env != NULL) 
    {
       for(pos1 = env->variable; pos1; pos1 = pos1->next) 
       {
          if (!strcmp(pos->name, name_second))
             break;
       }
    }
    if (pos1 == NULL) 
    {
        for(pos1 = clc_current_interpreter->variable; pos1; pos1 = pos1->next) 
        {
            if (!strcmp(pos1->name, name_second))
                break;
        }
    }
    if (pos1 == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_second);
    }    
    if(pos->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", identifier);
    }
    if(pos1->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", name_second);
    } 
    m1=pos->value.u.matrix_value.column;
    n1=pos->value.u.matrix_value.row;
    m2=pos1->value.u.matrix_value.column;
    n2=pos1->value.u.matrix_value.row;
    if (m1*n1 != m2*n2)
    {
        clc_runtime_error(NOT_SAME_DIMENSION_MATRIX_ERR, "(%s)\n", identifier);
    }
    for(i = 0; i < m1*n1; i++) 
    {
        fprintf(fp,"%22.17f     %22.17f\n", 
            pos->value.u.matrix_value.top[i], 
            pos1->value.u.matrix_value.top[i]); 
    } 
    fclose(fp);
    fp = popen("gnuplot './plot/_plot2'", "r");
    pclose(fp);
}


void 
eval_plot3_expression(LocalEnvironment *env, char *identifier, char *name_second, char *name_third)
{
    Variable    *pos;
    Variable    *pos1;
    Variable    *pos2;
    FILE *fp;
    int m1, n1, m2, n2, m3, n3;
    int i, j, i2;

    fp = fopen("./plot/_temp", "w");
    if (fp == NULL) 
    {
         clc_runtime_error(CANNOT_OPEN_FILE_ERR, "(%s)\n", NULL);
    } 
    pos = NULL;
    pos1 = NULL;
    pos2 = NULL;
    if (env != NULL) {
       for(pos = env->variable; pos; pos = pos->next) {
          if (!strcmp(pos->name, identifier))
             break;
       }
    }
    if (pos == NULL) 
    {
        for(pos = clc_current_interpreter->variable; pos; pos = pos->next) 
        {
            if (!strcmp(pos->name, identifier))
                break;
        }
    }
    if (pos == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", identifier);
    }    
    if (env != NULL) 
    {
       for(pos1 = env->variable; pos1; pos1 = pos1->next) 
       {
          if (!strcmp(pos->name, name_second))
             break;
       }
    }
    if (pos1 == NULL) 
    {
        for(pos1 = clc_current_interpreter->variable; pos1; pos1 = pos1->next) 
        {
            if (!strcmp(pos1->name, name_second))
                break;
        }
    }
    if (pos1 == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_second);
    }    
    if (env != NULL) {
       for(pos2 = env->variable; pos2; pos2 = pos2->next) {
          if (!strcmp(pos2->name, name_third))
             break;
       }
    }
    if (pos2 == NULL) 
    {
        for(pos2 = clc_current_interpreter->variable; pos2; pos2 = pos2->next) 
        {
            if (!strcmp(pos2->name, name_third))
                break;
        }
    }
    if (pos2 == NULL) 
    {
        clc_runtime_error(VARIABLE_NOT_FOUND_ERR, "(%s)\n", name_third);
    }    
    if(pos->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", identifier);
    }
    if(pos1->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", name_second);
    } 
    if(pos2->value.type != MATRIX_VALUE) 
    {
        clc_runtime_error(NOT_MATRIX_ERR, "(%s)\n", name_third);
    }
    m1=pos->value.u.matrix_value.column;
    n1=pos->value.u.matrix_value.row;
    m2=pos1->value.u.matrix_value.column;
    n2=pos1->value.u.matrix_value.row;
    m3=pos2->value.u.matrix_value.column;
    n3=pos2->value.u.matrix_value.row;
    if (m1*n1*m2*n2 != m3*n3)
    {
        clc_runtime_error(NOT_SAME_DIMENSION_MATRIX_ERR, "(%s)\n", identifier);
    }
    for(i = 0; i < m1*n1; i++) 
    {
        for(j=0; j < m2*n2; j++)
        {
            i2 = i*m1*n1+j;
            fprintf(fp,"%22.17f     %22.17f    %22.17f\n", 
                pos->value.u.matrix_value.top[i], 
                pos1->value.u.matrix_value.top[j], 
                pos2->value.u.matrix_value.top[i2]); 
        }
        fprintf(fp,"\n"); 
    } 
    fclose(fp);
    fp = popen("gnuplot './plot/_plot3'", "r");
    pclose(fp);
}

Value
eval_function_expression(LocalEnvironment *env, Expression *expr)
{
    Value	v;
    switch (expr->type) {
      case INT_EXPRESSION:
	v = eval_int_expression(expr->u.int_value);
	break;
      case DOUBLE_EXPRESSION:
	v = eval_double_expression(expr->u.double_value);
	break;
      case COMPLEX_EXPRESSION:
	v = eval_complex_expression(expr->u.complex_value.r, 
                                    expr->u.complex_value.i);
	break;
      case INTERVAL_EXPRESSION:
        v = eval_interval_expression(env,
              expr->u.interval_expression.lower,
              expr->u.interval_expression.upper);
        break;
      case INTERVAL_VALUE_EXPRESSION:
        v = eval_interval_value_expression(
              expr->u.interval_value.lower,
              expr->u.interval_value.upper);
        break;
      case INTERVAL_MATRIX_VALUE_EXPRESSION:
        v = eval_interval_matrix_value_expression(
              expr->u.interval_matrix_value.lower,
              expr->u.interval_matrix_value.upper);
        break;
      case IDENTIFIER_EXPRESSION:
	v = eval_identifier_expression(env, expr->u.identifier);
	break;
      case PRINT_EXPRESSION:
	eval_print_expression(env, expr->u.identifier);
	break;
      case WHO_EXPRESSION:
	eval_who_expression(env);
	break;
      case MATRIX_ONES_EXPRESSION:
	v = eval_matrix_ones_expression(env, 
               expr->u.matrix_ones_expression.row,
               expr->u.matrix_ones_expression.column);
	break;
      case MATRIX_EYE_EXPRESSION:
	v = eval_matrix_eye_expression(env, 
               expr->u.matrix_eye_expression.row,
               expr->u.matrix_eye_expression.column);
	break;
      case MATRIX_RAND_EXPRESSION:
	v = eval_matrix_rand_expression(env, 
               expr->u.matrix_rand_expression.row,
               expr->u.matrix_rand_expression.column);
	break;
      case MATRIX_ZEROS_EXPRESSION:
	v = eval_matrix_zeros_expression(env, 
               expr->u.matrix_zeros_expression.row,
               expr->u.matrix_zeros_expression.column);
	break;
      case MATRIX_ELEMENT_CALL_EXPRESSION:
	v = eval_matrix_element_call_expression(env, 
               expr->u.matrix_element_call_expression.identifier,
               expr->u.matrix_element_call_expression.row,
               expr->u.matrix_element_call_expression.column);
	break;
      case MATRIX_ELEMENT_ASSIGN_EXPRESSION:
	v = eval_matrix_element_assign_expression(env, 
               expr->u.matrix_element_assign_expression.variable,
               expr->u.matrix_element_assign_expression.row,
               expr->u.matrix_element_assign_expression.column,
               expr->u.matrix_element_assign_expression.operand);
	break;
      case EXPRESSION_LIST_EXPRESSION:
	v = eval_function_expression_list_expression
	    (env,
	     expr->u.expression_list.expression,
	     expr->u.expression_list.next);
	break;
      case ASSIGN_EXPRESSION:
	v = eval_function_assign_expression(env,
				   expr->u.assign_expression.variable,
				   expr->u.assign_expression.operand);
	break;
      case ADD_EXPRESSION:	/* FALLTHRU */
      case SUB_EXPRESSION:	/* FALLTHRU */
      case MUL_EXPRESSION:	/* FALLTHRU */
      case DIV_EXPRESSION:	/* FALLTHRU */
      case MOD_EXPRESSION:	/* FALLTHRU */
      case EQ_EXPRESSION:	/* FALLTHRU */
      case NE_EXPRESSION:	/* FALLTHRU */
      case OR_EXPRESSION:	/* FALLTHRU */
      case AND_EXPRESSION:	/* FALLTHRU */
      case GT_EXPRESSION:	/* FALLTHRU */
      case GE_EXPRESSION:	/* FALLTHRU */
      case LT_EXPRESSION:	/* FALLTHRU */
      case LE_EXPRESSION:	/* FALLTHRU */
	v = clc_eval_binary_expression(env,
				       expr->type,
				       expr->u.binary_expression.left,
				       expr->u.binary_expression.right);
	break;
      case MINUS_EXPRESSION:
	v = clc_eval_minus_expression(env, expr->u.minus_expression);
	break;
      case NORM_EXPRESSION:
	v = clc_eval_norm_expression(env, expr->u.norm_expression);
	break;
      case INV_EXPRESSION:
	v = clc_eval_inv_expression(env, expr->u.inv_expression);
	break;
      case TRANSPOSE_EXPRESSION:
	v = clc_eval_transpose_expression(env, expr->u.transpose_expression);
	break;
      case EIG_EXPRESSION:
	v = clc_eval_eig_expression(env, expr->u.eig_expression);
	break;
      case SIN_EXPRESSION:
	v = clc_eval_sin_expression(env, expr->u.sin_expression);
	break;
      case COS_EXPRESSION:
	v = clc_eval_cos_expression(env, expr->u.cos_expression);
	break;
      case LOG_EXPRESSION:
	v = clc_eval_log_expression(env, expr->u.log_expression);
	break;
      case LOGTEN_EXPRESSION:
	v = clc_eval_log10_expression(env, expr->u.log10_expression);
	break;
      case EXP_EXPRESSION:
	v = clc_eval_exp_expression(env, expr->u.exp_expression);
	break;
      case FABS_EXPRESSION:
	v = clc_eval_fabs_expression(env, expr->u.fabs_expression);
	break;
      case TAN_EXPRESSION:
	v = clc_eval_tan_expression(env, expr->u.tan_expression);
	break;
      case ATAN_EXPRESSION:
	v = clc_eval_atan_expression(env, expr->u.atan_expression);
	break;
      case ACOS_EXPRESSION:
	v = clc_eval_acos_expression(env, expr->u.acos_expression);
	break;
      case ASIN_EXPRESSION:
	v = clc_eval_asin_expression(env, expr->u.asin_expression);
	break;
      case COSH_EXPRESSION:
	v = clc_eval_cosh_expression(env, expr->u.cosh_expression);
	break;
      case SINH_EXPRESSION:
	v = clc_eval_sinh_expression(env, expr->u.sinh_expression);
	break;
      case TANH_EXPRESSION:
	v = clc_eval_tanh_expression(env, expr->u.tanh_expression);
	break;
      case SQRT_EXPRESSION:
	v = clc_eval_sqrt_expression(env, expr->u.sqrt_expression);
	break;
      case UP_EXPRESSION:
	v = clc_eval_up_expression(env, expr->u.up_expression);
	break;
      case DOWN_EXPRESSION:
	v = clc_eval_down_expression(env, expr->u.down_expression);
	break;
      case NEAR_EXPRESSION:
	v = clc_eval_near_expression(env, expr->u.near_expression);
	break;
      case MATRIX_EXPRESSION:
        v = clc_eval_matrix_expression(env, 
                                       expr->u.matrix_expression.top,
                                       expr->u.matrix_expression.column,
                                       expr->u.matrix_expression.row);
        break;
      case MATRIX:
        v = clc_eval_matrix(env, expr->u.matrix.matrix_element);
        break;
      case MATRIX_VALUE_EXPRESSION:
        v = clc_eval_matrix_value_expression(env, 
                                       expr->u.matrix_value.top,
                                       expr->u.matrix_value.column,
                                       expr->u.matrix_value.row);
        break;
      case COMPLEX_MATRIX_VALUE_EXPRESSION:
        v = clc_eval_complex_matrix_value_expression(env, 
                                       expr->u.complex_matrix_value.top,
                                       expr->u.complex_matrix_value.column,
                                       expr->u.complex_matrix_value.row);
        break;
      case MATRIX_ELEMENT_EXPRESSION:
        v = clc_eval_matrix_element_expression(env, 
                     expr->u.matrix_element_expression.element);
        break;
      case IF_EXPRESSION:
	v = eval_if_expression(env,
			       expr->u.if_expression.condition,
			       expr->u.if_expression.then_expression,
			       expr->u.if_expression.else_expression);
	break;
      case FOR_EXPRESSION:
	v = eval_for_expression(env,
				  expr->u.for_expression.first_condition,
				  expr->u.for_expression.second_condition,
				  expr->u.for_expression.third_condition,
				  expr->u.for_expression.expression_list);
	break;
      case WHILE_EXPRESSION:
	v = eval_while_expression(env,
				  expr->u.while_expression.condition,
				  expr->u.while_expression.expression_list);
	break;
      case FUNCTION_CALL_EXPRESSION:
	v = eval_function_call_expression
	    (env,
	     expr->u.function_call_expression.identifier,
	     expr->u.function_call_expression.argument);
	break;
      case EXPRESSION_TYPE_NUM:	/* FALLTHRU */
      default:
    }
    return v;
}

Value eval_interval_value_expression(double lower,double upper)
{
    Value v;
  
    v.type = INTERVAL_VALUE;
    v.u.interval_value.lower = lower;
    v.u.interval_value.upper = upper;

    return v;
}

Value eval_interval_matrix_value_expression(matr lower,matr upper)
{
    Value v;
  
    v.type = INTERVAL_MATRIX_VALUE;
    v.u.interval_matrix_value.lower = lower;
    v.u.interval_matrix_value.upper = upper;

    return v;
}

Value eval_interval_expression(LocalEnvironment *env, Expression *lower, Expression *upper)
{
    Value v;
    Value v1,v2;

    v1 = eval_expression(env,lower);
    v2 = eval_expression(env,upper); 

    if (v1.type == DOUBLE_VALUE
        || v2.type == DOUBLE_VALUE) {
        v.type = INTERVAL_VALUE;
        v.u.interval_value.lower = v1.u.double_value;
        v.u.interval_value.upper = v2.u.double_value;
        if (v.u.interval_value.lower > v.u.interval_value.upper) 
            printf("void interval!\n");
        return v;
    } else if (v1.type == MATRIX_VALUE &&
               v2.type == MATRIX_VALUE) {
        v.type = INTERVAL_MATRIX_VALUE;
        v.u.interval_matrix_value.lower.top = v1.u.matrix_value.top;
        v.u.interval_matrix_value.lower.row = v1.u.matrix_value.row;
        v.u.interval_matrix_value.lower.column = v1.u.matrix_value.column;
        v.u.interval_matrix_value.upper.top = v2.u.matrix_value.top;
        v.u.interval_matrix_value.upper.row = v2.u.matrix_value.row;
        v.u.interval_matrix_value.upper.column = v2.u.matrix_value.column;
        return v;
    } 
    return v; /* dummy */
}
